/*
 * docker_driver.c: core driver functions for managing Docker containers
 *
 * Copyright (C) 2020 Virtuozzo, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
#include <config.h>

#include <curl/curl.h>
#include <wsman-api.h>
#include <wsman-client.h>
#include <u/misc.h>

#include "internal.h"
#include "datatypes.h"
#include "virdomainobjlist.h"
#include "viralloc.h"
#include "virlog.h"
#include "viruuid.h"
#include "virstring.h"
#include "virjson.h"
#include "virfile.h"
#include "network_conf.h"
#include "storage_conf.h"
#include "configmake.h"
#include "base64.h"
#include "docker_driver.h"

#define DOCKER_API_VERSION "1.40"
#define DOCKER_NAME_REGEXP "^/?[a-zA-Z0-9][a-zA-Z0-9_.-]+$"
#define LABEL_LIBVIRT_UUID "libvirt.uuid"

#define VIR_FROM_THIS VIR_FROM_DOCKER

VIR_LOG_INIT("docker.docker_driver");


typedef enum {
    DOCKER_OS_UNKNOWN = 0,
    DOCKER_OS_LINUX = 1,
    DOCKER_OS_WINDOWS = 2
} dockerOS;


typedef struct _dockerPrivate dockerPrivate;
typedef dockerPrivate *dockerPrivatePtr;
struct _dockerPrivate {
    char *transport;
    char *pkipath;
    char *hostUri;
    int noVerify;
    dockerOS os;
    virArch hostArch;
    virCapsPtr hostCaps;
    virDomainXMLOptionPtr xmlOpt;
};


typedef struct _dockerApiParam dockerApiParam;
typedef dockerApiParam *dockerApiParamPtr;
struct _dockerApiParam {
    CURL *curl;
    struct curl_slist *headers;
    int valid_code;
    char *response;
    size_t size;
    virJSONValuePtr obj;
    virDomainStatsRecordPtr rec;
};


typedef struct _dockerJobsControl dockerJobsControl;
typedef dockerJobsControl *dockerJobsControlPtr;
struct _dockerJobsControl {
    CURLM *curlm;
    dockerApiParamPtr *jobs;
    size_t njobs;
    size_t njobs_max;
    unsigned int stats;
};


typedef int (*dockerJobHandlerPtr)(char *, dockerApiParamPtr, dockerJobsControlPtr);


typedef enum {
    DOCKER_WINRM_IDENTIFY = 0,
    DOCKER_WINRM_CREATE = 1,
    DOCKER_WINRM_INVOKE = 2,
    DOCKER_WINRM_DELETE = 3
} dockerWinrmRequestType;


typedef enum {
    DOCKER_API_HTTP_GET = 0,
    DOCKER_API_HTTP_POST = 1,
    DOCKER_API_HTTP_HEAD = 2,
    DOCKER_API_HTTP_DELETE = 3
} dockerApiHttpType;


/*
 * Get number of seconds elapsed since the Epoch, 1970-01-01 00:00:00 +0 (UTC)
 * from string in fromat RFC3339Nano
 * Expected format: CCYY-MM-DDTHH:MM:SS[.sssssssss][((+|-)HH:MM|Z)]
 */
static long long
dockerParseDateTime(const char *str)
{
    char *pch;
    struct tm tm;
    char sign;
    int nanosec, tz_hours, tz_minutes;
    int tz_offset = 0;

    pch = strptime(str, "%Y-%m-%dT%H:%M:%S", &tm);
    if (!pch) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Incorrect time format '%s'"), str);
        return -1;
    }

    if (*pch) {
        /* Skip .ssssssss part if present */
        if (*pch == '.' && virStrToLong_i(pch + 1, &pch, 10, &nanosec) < 0) {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Incorrect time format '%s'"), str);
            return -1;
        }

        /* Parse timezone offset if present */
        if (*pch == '+' || *pch == '-') {
            sign = *pch;

            if (virStrToLong_i(pch + 1, &pch, 10, &tz_hours) < 0 ||
                *pch != ':' ||
                virStrToLong_i(pch + 1, NULL, 10, &tz_minutes) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR,
                               _("Incorrect time format '%s'"), str);
                return -1;
            }

            tz_offset = tz_hours * 3600 + tz_minutes * 60;
            if (sign == '-')
                tz_offset = -tz_offset;
        } else if (*pch && *pch != 'Z') {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Incorrect time format '%s'"), str);
            return -1;
        }
    }

    return timegm(&tm) - tz_offset;
}


static void
dockerFreePrivate(dockerPrivatePtr *priv)
{
    if (priv || *priv)
        return;

    virObjectUnref((*priv)->xmlOpt);
    virObjectUnref((*priv)->hostCaps);
    VIR_FREE((*priv)->transport);
    VIR_FREE((*priv)->pkipath);
    VIR_FREE((*priv)->hostUri);
    VIR_FREE(*priv);
}


static bool
dockerIsCertPath(const char *path)
{
    if (!path)
        return false;

    if (virFileIsDir(path)) {
        VIR_AUTOFREE(char *) cert = NULL;
        VIR_AUTOFREE(char *) key = NULL;

        virAsprintf(&cert, "%s/cert.pem", path);
        virAsprintf(&key, "%s/key.pem", path);
        if (virFileExists(key) && virFileExists(cert))
            return true;
    }

    return false;
}


static char *
dockerDefaultPkiPath(void)
{
    VIR_AUTOFREE(char *) userdir = NULL;
    const char *envpath;
    char *pkipath;

    /* At first try to use env value */
    if ((envpath = virGetEnvAllowSUID("DOCKER_CERT_PATH")) &&
        dockerIsCertPath(envpath)) {
        VIR_STRDUP(pkipath, envpath);
        return pkipath;
    }

    /* At second try to use default path for Docker: ~/.docker */
    if (!(userdir = virGetUserDirectory()))
        return NULL;

    virAsprintf(&pkipath, "%s/.docker", userdir);
    if (dockerIsCertPath(pkipath))
        return pkipath;

    /* And finally we get default path */
    VIR_FREE(pkipath);
    VIR_STRDUP(pkipath, "/etc/pki/libvirt/docker");

    return pkipath;
}


static int
dockerParseUri(dockerPrivatePtr priv, virURIPtr uri)
{
    size_t i;

    if (!uri || !uri->scheme || STRCASENEQ(uri->scheme, "docker")) {
        virReportError(VIR_ERR_INVALID_ARG, "%s", _("Incorrect scheme"));
        return -1;
    }

    if (!uri->server) {
        virReportError(VIR_ERR_INVALID_ARG, "%s",
                       _("Only remote connection supported for Docker"));
        return -1;
    }

    VIR_STRDUP(priv->transport, "https");

    if (!(priv->pkipath = dockerDefaultPkiPath()))
        return -1;

    for (i = 0; i < uri->paramsCount; i++) {
        virURIParamPtr queryParam = &uri->params[i];
        if (STRCASEEQ(queryParam->name, "transport")) {
            if (STRCASENEQ(queryParam->value, "http") &&
                STRCASENEQ(queryParam->value, "https")) {
                virReportError(VIR_ERR_INVALID_ARG,
                               _("Query parameter 'transport' has unexpected value "
                                 "'%s' (should be http|https)"),
                               queryParam->value);
                return -1;
            }

            VIR_FREE(priv->transport);
            VIR_STRDUP(priv->transport, queryParam->value);
        } else if (STRCASEEQ(queryParam->name, "pkipath")) {
            if (!dockerIsCertPath(queryParam->value)) {
                virReportError(VIR_ERR_INVALID_ARG,
                               _("Query parameter 'pkipath' has unexpected value "
                                 "'%s'"), queryParam->value);
                return -1;
            }

            VIR_FREE(priv->pkipath);
            VIR_STRDUP(priv->pkipath, queryParam->value);
        } else if (STRCASEEQ(queryParam->name, "no_verify")) {
            if (virStrToLong_i(queryParam->value, NULL, 10, &(priv->noVerify)) < 0 ||
                (priv->noVerify != 0 && priv->noVerify != 1)) {
                virReportError(VIR_ERR_INVALID_ARG,
                               _("Query parameter 'no_verify' has unexpected value "
                                 "'%d' (should be 0 or 1)"),
                               priv->noVerify);
                return -1;
            }
        } else {
            VIR_WARN("Ignoring unexpected query parameter '%s'",
                     queryParam->name);
        }
    }

    if (uri->port == 0) {
        if (STRCASEEQ(priv->transport, "http"))
            uri->port = 2375;
        else
            uri->port = 2376;
    }

    VIR_INFO("Remote server: %s", uri->server);
    VIR_INFO("Remote port: %d", uri->port);
    VIR_INFO("Transport: %s", priv->transport);
    VIR_INFO("Certificate dir: %s", priv->pkipath);
    VIR_INFO("Verify server: %s", priv->noVerify ? "no" : "yes");

    return 0;
}


static int
dockerCapsInit(dockerPrivatePtr priv)
{
    int ret = -1;
    virCapsPtr caps = NULL;
    virCapsGuestPtr guest = NULL;

    if (!(caps = virCapabilitiesNew(priv->hostArch, true, false)))
        goto cleanup;

    if (priv->os == DOCKER_OS_WINDOWS &&
        virCapabilitiesAddStoragePool(caps, VIR_STORAGE_POOL_WSS) < 0)
        goto cleanup;

    if (!(guest = virCapabilitiesAddGuest(caps, VIR_DOMAIN_OSTYPE_EXE,
                                          priv->hostArch, NULL, NULL, 0, NULL)))
        goto cleanup;

    if (!virCapabilitiesAddGuestDomain(guest, VIR_DOMAIN_VIRT_DOCKER,
                                       NULL, NULL, 0, NULL))
        goto cleanup;

    VIR_STEAL_PTR(priv->hostCaps, caps);
    guest = NULL;
    ret = 0;

 cleanup:
    virCapabilitiesFreeGuest(guest);
    virObjectUnref(caps);

    return ret;
}


/*
 * Callback function for small patching of authentication
 * requests because we must set special authentication method
 * in http header for certificate authentication.
 */
static void
dockerWinrmAuthPatch(WsManClient *client,
                     wsman_auth_type_t type,
                     char **username,
                     char **password ATTRIBUTE_UNUSED)
{
    CURL *curl;
    struct curl_slist *headers = NULL;
    VIR_AUTOFREE(char *) content = NULL;

    if (type != WS_GSSNEGOTIATE_AUTH) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("'%s' doesn't support negotiate authentication"),
                       wsmc_get_endpoint(client));
        return;
    }

    curl = (CURL *)client->transport;
    curl_easy_getinfo(curl, CURLINFO_PRIVATE, &headers);
    if (headers) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Authentication for '%s' failed"),
                       wsmc_get_endpoint(client));
        return;
    }

    virAsprintf(&content, "Content-Type: application/soap+xml;charset=%s",
                wsmc_get_encoding(client));

    headers = curl_slist_append(headers, content);
    headers = curl_slist_append(headers, "User-Agent: " PACKAGE_STRING);
    headers = curl_slist_append(headers, "Authorization: "
                                "http://schemas.dmtf.org/wbem/wsman/1"
                                "/wsman/secprofile/https/mutual");
    curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);
    curl_easy_setopt(curl, CURLOPT_PRIVATE, headers);

    /* This is sing of successuful completion for opewsman */
    *username = u_strdup("None");
}


static WsXmlDocH
dockerWinrmRequest(virConnectPtr conn,
                   WsManClient **client,
                   dockerWinrmRequestType type,
                   const char *uri,
                   const char *method,
                   client_opt_t *options,
                   WsXmlDocH request)
{
    dockerPrivatePtr priv = conn->privateData;
    WsXmlDocH response = NULL;

    if (!*client) {
        if (!(*client = wsmc_create(conn->uri->server,
                                    STRCASEEQ(priv->transport, "https") ?
                                    5986 : 5985, NULL,
                                    priv->transport, NULL, NULL))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Could not create openwsman client"));
            return NULL;
        }

        /* Don't call global initializion by wsmc_transport_init */
        (*client)->initialized = 1;

        /* Set callback function for patching certificate authentication */
        wsmc_transport_set_auth_request_func(*client, dockerWinrmAuthPatch);

        wsman_transport_set_auth_method(*client, _WS_GSSNEGOTIATE_AUTH);
        wsman_transport_set_timeout(*client, options->timeout / 1000 + 10);

        /*
         * We set two lines in HTTP header at the same time:
         * first line with user agent value
         * second line with authorization value
         */
        wsman_transport_set_agent(*client, PACKAGE_STRING "\r\nAuthorization: "
                                  "http://schemas.dmtf.org/wbem/wsman/1"
                                  "/wsman/secprofile/https/mutual");

        if (STRCASEEQ(priv->transport, "https")) {
            VIR_AUTOFREE(char *) cert = NULL;
            VIR_AUTOFREE(char *) key = NULL;

            virAsprintf(&cert, "%s/cert.pem", priv->pkipath);
            virAsprintf(&key, "%s/key.pem", priv->pkipath);
            wsman_transport_set_cert(*client, cert);
            wsman_transport_set_key(*client, key);
            if (priv->noVerify) {
                wsman_transport_set_verify_peer(*client, 0);
                wsman_transport_set_verify_host(*client, 0);
            } else {
                VIR_AUTOFREE(char *) ca = NULL;

                virAsprintf(&ca, "%s/ca.pem", priv->pkipath);
                wsman_transport_set_cainfo(*client, ca);
                wsman_transport_set_verify_peer(*client, 1);
                wsman_transport_set_verify_host(*client, 2);
            }
        }
    }

    if ((*client)->transport)
        curl_easy_setopt((CURL *)(*client)->transport, CURLOPT_TIMEOUT,
                         options->timeout / 1000 + 10);

    switch (type) {
    case DOCKER_WINRM_IDENTIFY:
        response = wsmc_action_identify(*client, options);
        break;
    case DOCKER_WINRM_CREATE:
        response = wsmc_action_create(*client, uri, options, request);
        break;
    case DOCKER_WINRM_INVOKE:
        response = wsmc_action_invoke(*client, uri, options, method, request);
        break;
    case DOCKER_WINRM_DELETE:
        response = wsmc_action_delete(*client, uri, options);
    }

    if (!response) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Request to '%s' failed, code %d(%ld) %s"),
                       wsmc_get_endpoint(*client),
                       wsmc_get_last_error(*client),
                       wsmc_get_response_code(*client),
                       wsmc_get_fault_string(*client));
    } else if (wsmc_check_for_fault(response)) {
        WsManFault *fault = wsmc_fault_new();

        wsmc_get_fault_data(response, fault);
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Received fault from '%s': %s(%s) %s"),
                       wsmc_get_endpoint(*client),
                       fault->code, fault->subcode, fault->reason);

        wsmc_fault_destroy(fault);
        ws_xml_destroy_doc(response);
        response = NULL;
    }

    if ((*client)->transport) {
        CURL *curl = (CURL *)(*client)->transport;
        struct curl_slist *headers = NULL;

        curl_easy_getinfo(curl, CURLINFO_PRIVATE, &headers);
        if (headers) {
            curl_slist_free_all(headers);
            curl_easy_setopt(curl, CURLOPT_PRIVATE, NULL);
        }
    }

    return response;
}


static int
dockerWinrsCommand(virConnectPtr conn,
                   const char *cmd,
                   char **cmd_out)
{
    int exit_code, ret = -1;
    const char *ns = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell";
    VIR_AUTOCLEAN(virBuffer) buf = VIR_BUFFER_INITIALIZER;
    VIR_AUTOFREE(char *) uri = NULL;
    VIR_AUTOFREE(char *) shell = NULL;
    VIR_AUTOFREE(char *) invoke = NULL;
    VIR_AUTOFREE(char *) command = NULL;
    VIR_AUTOFREE(char *) terminate = NULL;
    client_opt_t *opt;
    WsManClient *client = NULL;
    WsXmlDocH request, response;
    WsXmlNodeH root, node;
    char *val;

    virAsprintf(&uri, "%s/cmd", ns);

    opt = wsmc_options_init();
    opt->timeout = 30000; /* timeout in msec */
    opt->max_envelope_size = 153600;
    wsmc_set_locale(opt, "en-US");
    wsmc_add_option(opt, "WINRS_NOPROFILE", "FALSE");
    wsmc_add_option(opt, "WINRS_CODEPAGE", "437");

    request = ws_xml_clone_and_create_doc(NULL, ns, "Shell");
    root = ws_xml_get_doc_root(request);
    ws_xml_add_child(root, ns, "Lifetime", "PT300.0S");
    ws_xml_add_child(root, ns, "IdleTimeOut", "PT300.0S");
    ws_xml_add_child(root, ns, "InputStreams", "");
    ws_xml_add_child(root, ns, "OutputStreams", "stdout");
    if (!(response = dockerWinrmRequest(conn, &client, DOCKER_WINRM_CREATE,
                                        uri, NULL, opt, request)))
        goto cleanup;

    if (!(val = ws_xml_get_xpath_value(response, (char *)
                                       "/s:Envelope/s:Body/rsp:Shell"
                                       "/rsp:ShellId"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Shell ID not found in response"));
        goto cleanup;
    }

    VIR_STRDUP(shell, val);
    ws_xml_destroy_doc(response);
    ws_xml_destroy_doc(request);
    wsmc_options_destroy(opt);

    virAsprintf(&invoke, "%s/Command", ns);

    opt = wsmc_options_init();
    opt->timeout = 30000; /* timeout in msec */
    opt->max_envelope_size = 153600;
    wsmc_set_locale(opt, "en-US");
    wsmc_add_selector(opt, "ShellId", shell);
    wsmc_add_option(opt, "WINRS_SKIP_CMD_SHELL", "FALSE");
    wsmc_add_option(opt, "WINRS_CONSOLEMODE_STDIN", "TRUE");

    request = ws_xml_clone_and_create_doc(NULL, ns, "CommandLine");
    root = ws_xml_get_doc_root(request);
    ws_xml_add_child(root, ns, "Command", cmd);
    if (!(response = dockerWinrmRequest(conn, &client, DOCKER_WINRM_INVOKE,
                                        uri, invoke, opt, request)))
        goto cleanup;

    if (!(val = ws_xml_get_xpath_value(response, (char *)
                                       "/s:Envelope/s:Body/rsp:CommandResponse"
                                       "/rsp:CommandId"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Command ID not found in response"));
        goto cleanup;
    }

    VIR_STRDUP(command, val);
    ws_xml_destroy_doc(response);
    ws_xml_destroy_doc(request);
    wsmc_options_destroy(opt);

    VIR_FREE(invoke);
    virAsprintf(&invoke, "%s/Receive", ns);

    opt = wsmc_options_init();
    opt->timeout = 240000; /* timeout in msec */
    opt->max_envelope_size = 153600;
    wsmc_set_locale(opt, "en-US");
    wsmc_add_selector(opt, "ShellId", shell);

    request = ws_xml_clone_and_create_doc(NULL, ns, "Receive");
    root = ws_xml_get_doc_root(request);
    node = ws_xml_add_child(root, ns, "DesiredStream", "stdout");
    ws_xml_add_node_attr(node, NULL, "CommandId", command);
    if (!(response = dockerWinrmRequest(conn, &client, DOCKER_WINRM_INVOKE,
                                        uri, invoke, opt, request)))
        goto cleanup;

    if (!(val = ws_xml_get_xpath_value(response, (char *)
                                       "/s:Envelope/s:Body/rsp:ReceiveResponse"
                                       "/rsp:CommandState/rsp:ExitCode"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Exit code not found in response"));
        goto cleanup;
    }

    if (virStrToLong_i(val, NULL, 10, &exit_code) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Invalid format of exit code in response"));
        goto cleanup;
    }

    if (cmd_out && !exit_code &&
        (node = ws_xml_get_soap_body(response)) &&
        (node = ws_xml_find_in_tree(node, NULL, "ReceiveResponse", 1))) {
        size_t i, count = ws_xml_get_child_count_by_qname(node, NULL, "Stream");

        for (i = 0; i < count; i++) {
            WsXmlNodeH item = ws_xml_get_child(node, i, NULL, "Stream");
            WsXmlAttrH attr = ws_xml_find_node_attr(item, NULL, "CommandId");
            if (attr && STREQ(command, ws_xml_get_attr_value(attr))) {
                VIR_AUTOFREE(char *) decode = NULL;
                size_t n;

                if ((val = ws_xml_get_node_text(item)) &&
                    base64_decode_alloc(val, strlen(val), &decode, &n))
                    virBufferAdd(&buf, decode, n);
            }
        }
    }

    ws_xml_destroy_doc(response);
    ws_xml_destroy_doc(request);

    VIR_FREE(invoke);
    virAsprintf(&invoke, "%s/Signal", ns);
    virAsprintf(&terminate, "%s/signal/terminate", ns);

    opt->timeout = 30000; /* timeout in msec */

    request = ws_xml_clone_and_create_doc(NULL, ns, "Signal");
    root = ws_xml_get_doc_root(request);
    ws_xml_add_node_attr(root, NULL, "CommandId", command);
    ws_xml_add_child(root, ns, "Code", terminate);
    if (!(response = dockerWinrmRequest(conn, &client, DOCKER_WINRM_INVOKE,
                                        uri, invoke, opt, request)))
        goto cleanup;

    ws_xml_destroy_doc(response);

    if (!(response = dockerWinrmRequest(conn, &client, DOCKER_WINRM_DELETE,
                                        uri, NULL, opt, NULL)))
        goto cleanup;

    if (exit_code) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Exit code from host '%d'"), exit_code);
    } else {
        if (cmd_out) {
            if (virBufferCheckError(&buf) < 0)
                goto cleanup;

            *cmd_out = virBufferContentAndReset(&buf);
        }
    }

    ret = exit_code;

 cleanup:
    if (response)
        ws_xml_destroy_doc(response);
    ws_xml_destroy_doc(request);
    wsmc_options_destroy(opt);
    if (client)
        wsmc_release(client);

    return ret;
}


static int
dockerWinrsPowershell(virConnectPtr conn,
                      const char *ps,
                      char **ps_out)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) base64 = NULL;
    VIR_AUTOFREE(uint16_t *) utf16le = NULL;
    size_t i, len;

    /*
     * Encode a Powershell script to base64 and execute
     * the powershell encoded script command.
     * Source Powershell script string must use utf16le!
     */
    len = strlen(ps);
    VIR_ALLOC_N(utf16le, len);

    for (i = 0; i < len; i++)
        utf16le[i] = ps[i];

    base64 = virStringEncodeBase64((uint8_t *)utf16le, len * 2);
    virAsprintf(&cmd, "powershell -EncodedCommand %s", base64);

    return dockerWinrsCommand(conn, cmd, ps_out);
}


static size_t
dockerWriteFunc(void *data, size_t size, size_t nmemb, void *userp)
{
    size_t realsize = size * nmemb;
    dockerApiParamPtr param = (dockerApiParamPtr)userp;

    VIR_REALLOC_N(param->response, param->size + realsize);
    memcpy(param->response + param->size, data, realsize);
    param->size += realsize;

    return realsize;
}


static dockerApiParamPtr
dockerPrepareApiRequest(dockerPrivatePtr priv,
                        const char *cmd,
                        bool ver,
                        dockerApiHttpType type,
                        const char *post,
                        int valid_code)
{
    CURL *curl;
    dockerApiParamPtr param = NULL;
    VIR_AUTOFREE(char *) url_request = NULL;

    if (!(curl = curl_easy_init())) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Initialization libcurl failed"));
        return NULL;
    }

    VIR_ALLOC(param);
    virAsprintf(&url_request, "%s%s/%s", priv->hostUri,
                ver ? "/v"DOCKER_API_VERSION : "", cmd);

    curl_easy_setopt(curl, CURLOPT_PROTOCOLS,
                     STRCASEEQ(priv->transport, "https") ?
                     CURLPROTO_HTTPS : CURLPROTO_HTTP);
    curl_easy_setopt(curl, CURLOPT_URL, url_request);
    /*
     * We set timeout for POST as 30 min because
     * Docker can pull and unpack image so long.
     */
    curl_easy_setopt(curl, CURLOPT_TIMEOUT, /* timeout in sec */
                     type == DOCKER_API_HTTP_POST ? 1800 : 30);
    curl_easy_setopt(curl, CURLOPT_NOSIGNAL, 1);
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, &dockerWriteFunc);
    curl_easy_setopt(curl, CURLOPT_WRITEDATA, param);

    switch (type) {
        case DOCKER_API_HTTP_GET:
            break;
        case DOCKER_API_HTTP_POST:
            curl_easy_setopt(curl, CURLOPT_POST, 1);
            if (!post) {
                curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, 0);
            } else {
                /* We send all our POST requests only in JSON format */
                param->headers = curl_slist_append(param->headers,
                                                   "Content-Type: application/json");
                param->headers = curl_slist_append(param->headers, "Expect:");
                curl_easy_setopt(curl, CURLOPT_HTTPHEADER, param->headers);
#ifdef CURLOPT_EXPECT_100_TIMEOUT_MS
                curl_easy_setopt(curl, CURLOPT_EXPECT_100_TIMEOUT_MS, 0);
#endif
                curl_easy_setopt(curl, CURLOPT_COPYPOSTFIELDS, post);
            }
            break;
        case DOCKER_API_HTTP_HEAD:
            curl_easy_setopt(curl, CURLOPT_NOBODY, 1);
            break;
        case DOCKER_API_HTTP_DELETE:
            curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "DELETE");
    }

    if (STRCASEEQ(priv->transport, "https")) {
        VIR_AUTOFREE(char *) cert = NULL;
        VIR_AUTOFREE(char *) key = NULL;

        virAsprintf(&cert, "%s/cert.pem", priv->pkipath);
        virAsprintf(&key, "%s/key.pem", priv->pkipath);
        curl_easy_setopt(curl, CURLOPT_SSLCERT, cert);
        curl_easy_setopt(curl, CURLOPT_SSLKEY, key);
        if (priv->noVerify) {
            curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, 0);
            curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST, 0);
        } else {
            VIR_AUTOFREE(char *) ca = NULL;

            virAsprintf(&ca, "%s/ca.pem", priv->pkipath);
            curl_easy_setopt(curl, CURLOPT_CAINFO, ca);
            curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, 1);
            curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST, 2);
        }
    }

    param->curl = curl;
    param->valid_code = valid_code;

    return param;
}


static char *
dockerCompleteApiRequest(dockerApiParamPtr param)
{
    long response_code = -1;
    char *response;

    VIR_EXPAND_N(param->response, param->size, 1);

    curl_easy_getinfo(param->curl, CURLINFO_RESPONSE_CODE, &response_code);
    if (response_code != param->valid_code) {
        VIR_AUTOPTR(virJSONValue) obj = NULL;
        const char *msg;

        if (!(obj = virJSONValueFromString(param->response)) ||
            !(msg = virJSONValueObjectGetString(obj, "message"))) {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Invalid response code '%ld'"),
                           response_code);
        } else {
            virReportError(VIR_ERR_INTERNAL_ERROR, _("%ld, %s"),
                           response_code, msg);
        }

        return NULL;
    }

    VIR_STEAL_PTR(response, param->response);

    return response;
}


static void
dockerCloseApiRequest(dockerApiParamPtr param)
{
    if (!param)
        return;

    curl_slist_free_all(param->headers);
    curl_easy_cleanup(param->curl);

    VIR_FREE(param->response);
    VIR_FREE(param);
}


static char *
dockerApiRequest(dockerPrivatePtr priv,
                 const char *cmd,
                 bool ver,
                 dockerApiHttpType type,
                 const char *post,
                 int valid_code)
{
    CURLcode code;
    dockerApiParamPtr param;
    char *response = NULL;

    if (!(param = dockerPrepareApiRequest(priv, cmd, ver, type,
                                          post, valid_code)))
        return NULL;

    if ((code = curl_easy_perform(param->curl)) != CURLE_OK) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Curl error %d: %s"), code,
                       curl_easy_strerror(code));
    } else {
        response = dockerCompleteApiRequest(param);
    }

    dockerCloseApiRequest(param);

    return response;
}


static int
dockerAddApiJob(dockerJobsControlPtr jobsctl, dockerApiParamPtr param)
{
    CURLMcode code;

    if (!jobsctl->curlm) {
        if (!(jobsctl->curlm = curl_multi_init())) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Initialization libcurl failed"));
            return -1;
       }
    }

    curl_easy_setopt(param->curl, CURLOPT_PRIVATE, param);

    if ((code = curl_multi_add_handle(jobsctl->curlm, param->curl)) != CURLM_OK) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Curl error %d: %s"), code,
                       curl_multi_strerror(code));
        return -1;
    }

    VIR_RESIZE_N(jobsctl->jobs, jobsctl->njobs_max, jobsctl->njobs, 1);
    jobsctl->jobs[jobsctl->njobs++] = param;

    return 0;
}


static int
dockerWaitAllJobs(dockerJobsControlPtr jobsctl, dockerJobHandlerPtr func)
{
    int num_alive = jobsctl->njobs;

    while (num_alive) {
        int dummy;
        CURLMsg *msg;
        CURLMcode code;

        if ((code = curl_multi_perform(jobsctl->curlm, &num_alive)) != CURLM_OK) {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Curl error %d: %s"), code,
                           curl_multi_strerror(code));
            return -1;
        }

        while ((msg = curl_multi_info_read(jobsctl->curlm, &dummy))) {
            char *response;
            dockerApiParamPtr param;

            if (msg->msg != CURLMSG_DONE || msg->data.result != CURLE_OK) {
                virReportError(VIR_ERR_INTERNAL_ERROR,
                               _("Curl error %d: %s"), msg->data.result,
                               curl_easy_strerror(msg->data.result));
                return -1;
            }

            curl_easy_getinfo(msg->easy_handle, CURLINFO_PRIVATE, &param);

            if (!(response = dockerCompleteApiRequest(param)) ||
                func(response, param, jobsctl) < 0)
                return -1;
        }

        if (!num_alive)
            break;

        curl_multi_wait(jobsctl->curlm, NULL, 0, 1000, NULL);
    }

    return 0;
}


static void
dockerCloseJobs(dockerJobsControlPtr jobsctl)
{
    size_t i;

    if (!jobsctl->njobs)
        return;

    for (i = 0; i < jobsctl->njobs; i++) {
        curl_multi_remove_handle(jobsctl->curlm, jobsctl->jobs[i]->curl);
        dockerCloseApiRequest(jobsctl->jobs[i]);
    }

    curl_multi_cleanup(jobsctl->curlm);
    jobsctl->curlm = NULL;

    VIR_FREE(jobsctl->jobs);
    jobsctl->njobs_max = 0;
    jobsctl->njobs = 0;
}


static virJSONValuePtr
dockerGetJSONObject(dockerPrivatePtr priv,
                    const char *cmd,
                    bool ver,
                    dockerApiHttpType type,
                    const char *post,
                    int valid_code)
{
    virJSONValuePtr obj;
    VIR_AUTOFREE(char *) response = NULL;

    if (!(response = dockerApiRequest(priv, cmd, ver, type,
                                      post, valid_code)))
        return NULL;

    if (!(obj = virJSONValueFromString(response)))
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received invalid JSON document"));

    return obj;
}


static int
dockerCheckApiVersion(dockerPrivatePtr priv)
{
    const char *maxver, *minver, *os_str, *arch_str;
    unsigned long req_ver, max_ver, min_ver;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    if (!(obj = dockerGetJSONObject(priv, "version", false,
                                    DOCKER_API_HTTP_GET,
                                    NULL, 200)))
        return -1;

    if (!(maxver = virJSONValueObjectGetString(obj, "ApiVersion")) ||
        !(minver = virJSONValueObjectGetString(obj, "MinAPIVersion"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Version fields not found in JSON document"));
        return -1;
    }

    if (virParseVersionString(maxver, &max_ver, true) < 0 ||
        virParseVersionString(minver, &min_ver, true) < 0 ||
        virParseVersionString(DOCKER_API_VERSION, &req_ver, true) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect format version fields in JSON document"));
        return -1;
    }

    if (req_ver > max_ver || req_ver < min_ver) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Unsupported version of Docker API"));
        return -1;
    }

    if (!(os_str = virJSONValueObjectGetString(obj, "Os")) ||
        !(arch_str = virJSONValueObjectGetString(obj, "Arch"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect format in JSON document"));
        return -1;
    }

    if (STREQ(os_str, "linux"))
        priv->os = DOCKER_OS_LINUX;
    else if (STREQ(os_str, "windows"))
        priv->os = DOCKER_OS_WINDOWS;
    else
        priv->os = DOCKER_OS_UNKNOWN;

    priv->hostArch = virArchFromString(arch_str);
    if (priv->hostArch == VIR_ARCH_NONE && STREQ(arch_str, "amd64"))
        priv->hostArch = VIR_ARCH_X86_64;

    return 0;
}


static int
dockerGetUUID(virJSONValuePtr obj, unsigned char *uuid)
{
    VIR_AUTOFREE(char *) uuid_str = NULL;
    virJSONValuePtr config, labels;
    const char *id_str;

    /* At first we try to get libvirt uuid from container labels */
    if (((labels = virJSONValueObjectGet(obj, "Labels")) ||
         ((config = virJSONValueObjectGet(obj, "Config")) &&
          (labels = virJSONValueObjectGet(config, "Labels")))) &&
        (id_str = virJSONValueObjectGetString(labels, LABEL_LIBVIRT_UUID))) {
        unsigned char libvirt_uuid[VIR_UUID_BUFLEN];

        /*
         * If value of label isn't suitable by format
         * we will use native Docker id because the label
         * could be created not by us.
         */
        if (virUUIDParse(id_str, libvirt_uuid) >= 0) {
            memcpy(uuid, libvirt_uuid, VIR_UUID_BUFLEN);
            return 0;
        }
    }

    if (!(id_str = virJSONValueObjectGetString(obj, "Id"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Id' field in JSON container"));
        return -1;
    }

    /*
     * We can get Docker id as short (12 bytes) hash string or
     * as very long (64 bytes) string.
     */
    if (strlen(id_str) < VIR_UUID_BUFLEN * 2)
        virAsprintf(&uuid_str, "%s 0000 0000 0000 0000 0000", id_str);
    else
        VIR_STRNDUP(uuid_str, id_str, VIR_UUID_BUFLEN * 2);

    if (virUUIDParse(uuid_str, uuid) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Could not parse UUID from string '%s'"), id_str);
        return -1;
    }

    return 0;
}


/*
 * Function seeks docker object by value of special label
 * in object configuration.
 * Returns -1 on error, 1 if found and 0 if not found.
 */
static int
dockerSeekLabel(dockerPrivatePtr priv,
                const unsigned char *uuid,
                bool from_networks,
                char **docker_id)
{
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOPTR(virJSONValue) list = NULL;
    virJSONValuePtr obj;
    const char *id_str;
    size_t size;

    /* Encoded filter: {"label":["key=value"]} */
    virAsprintf(&cmd, "%sfilters="
                "{%%22label%%22:[%%22"LABEL_LIBVIRT_UUID"%%3d%s%%22]}",
                (from_networks) ? "networks?" : "containers/json?all=1&",
                virUUIDFormat(uuid, uuidstr));

    if (!(list = dockerGetJSONObject(priv, cmd, true,
                                     DOCKER_API_HTTP_GET,
                                     NULL, 200)))
        return -1;

    if (!virJSONValueIsArray(list)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received not JSON array"));
        return -1;
    }

    if ((size = virJSONValueArraySize(list)) > 1) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Found more than one object"));
        return -1;
    }

    if (!docker_id)
        return size;

    if (!size) {
        virAsprintf(docker_id, "%02x%02x%02x%02x%02x%02x",
                    uuid[0], uuid[1], uuid[2], uuid[3],
                    uuid[4], uuid[5]);
        return 0;
    }

    obj = virJSONValueArrayGet(list, 0);
    if (!(id_str = virJSONValueObjectGetString(obj, "Id"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Id' field in JSON object"));
        return -1;
    }

    VIR_STRDUP(*docker_id, id_str);

    return 1;
}


static virDomainState
dockerGetContainerState(virJSONValuePtr obj)
{
    virDomainState ret = VIR_DOMAIN_NOSTATE;
    virJSONValuePtr state_obj;
    const char *state;

    if (!(state_obj = virJSONValueObjectGet(obj, "State"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Not found 'State' in JSON container"));
        return ret;
    }

    /*
     * We can receive 2 type of JSON container object:
     * short type where field 'State' is string
     * long type where 'State' is compound object with 'Status' field
     */
    if (virJSONValueGetType(state_obj) == VIR_JSON_TYPE_STRING)
        state = virJSONValueGetString(state_obj);
    else
        state = virJSONValueObjectGetString(state_obj, "Status");
    if (!state) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'State' in JSON container"));
        return ret;
    }

    if (STRCASEEQ(state, "running") || STRCASEEQ(state, "restarting"))
        ret = VIR_DOMAIN_RUNNING;
    else if (STRCASEEQ(state, "paused"))
        ret = VIR_DOMAIN_PAUSED;
    else if (STRCASEEQ(state, "created") || STRCASEEQ(state, "exited") ||
             STRCASEEQ(state, "removing"))
        ret = VIR_DOMAIN_SHUTOFF;
    else if (STRCASEEQ(state, "dead"))
        ret = VIR_DOMAIN_CRASHED;

    return ret;
}


static int
dockerGetId(virJSONValuePtr obj)
{
    virDomainState state;
    const char *id;

    /* We return -1 for inactive container */
    state = dockerGetContainerState(obj);
    if ((state != VIR_DOMAIN_RUNNING) && (state != VIR_DOMAIN_PAUSED))
        return -1;

    /* We return -2 for case of error */
    if (!(id = virJSONValueObjectGetString(obj, "Id"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Not found 'Id' in JSON container"));
        return -2;
    }

    /*
     * For active container make fictional process id
     * from first 4 symbols of native docker uuid
     */
    return (virHexToBin(id[0]) << 12) + (virHexToBin(id[1]) << 8) +
           (virHexToBin(id[2]) << 4) + virHexToBin(id[3]);
}


static int
dockerGetContainerAutostart(virJSONValuePtr obj, int *autostart)
{
    virJSONValuePtr host_conf, restart_pol;
    const char *policy;

    if (!(host_conf = virJSONValueObjectGet(obj, "HostConfig")) ||
        !(restart_pol = virJSONValueObjectGet(host_conf, "RestartPolicy")) ||
        !(policy = virJSONValueObjectGetString(restart_pol, "Name"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Not found 'RestartPolicy' in JSON container"));
        return -1;
    }

    if (STRCASEEQ(policy, "always") || STRCASEEQ(policy, "unless-stopped")) {
        *autostart = 1;
    } else if (!policy[0] || STRCASEEQ(policy, "no") ||
               STRCASEEQ(policy, "on-failure")) {
        *autostart = 0;
    } else {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Unknown value '%s' for 'RestartPolicy'"),
                       policy);
        return -1;
    }

    return 0;
}


static virDomainPtr
dockerGetDomainFromObject(virConnectPtr conn, virJSONValuePtr obj)
{
    const char *name;
    unsigned char uuid[VIR_UUID_BUFLEN];
    virJSONValuePtr names, first_name;
    int id;

    if (!virJSONValueIsObject(obj)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Container not JSON object"));
        return NULL;
    }

    if (dockerGetUUID(obj, uuid) < 0)
        return NULL;

    if ((id = dockerGetId(obj)) < -1)
        return NULL;

    if ((!(names = virJSONValueObjectGet(obj, "Names")) ||
         !virJSONValueIsArray(names) ||
         !(first_name = virJSONValueArrayGet(names, 0)) ||
         !(name = virJSONValueGetString(first_name))) &&
        !(name = virJSONValueObjectGetString(obj, "Name"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Names' field in JSON container"));
        return NULL;
    }

    /* First symbol in docker name always slash so we can skip it */
    return virGetDomain(conn, name + 1, uuid, id);
}


static virJSONValuePtr
dockerGetContainerById(dockerPrivatePtr priv, const char *id_str)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    virJSONValuePtr obj;

    virAsprintf(&cmd, "containers/%s/json", id_str);

    if (!(obj = dockerGetJSONObject(priv, cmd, true,
                                    DOCKER_API_HTTP_GET,
                                    NULL, 200)))
        return NULL;

    if (!virJSONValueIsObject(obj)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Container not JSON object"));
        virJSONValueFree(obj);
        return NULL;
    }

    return obj;
}


static virJSONValuePtr
dockerGetContainerByUUID(dockerPrivatePtr priv, const unsigned char *uuid)
{
    VIR_AUTOFREE(char *) id_str = NULL;

    if (dockerSeekLabel(priv, uuid, false, &id_str) < 0)
        return NULL;

    return dockerGetContainerById(priv, id_str);
}


static int
dockerContainerUpdateById(dockerPrivatePtr priv,
                          const char *id_str,
                          virJSONValuePtr obj)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) json = NULL;
    VIR_AUTOFREE(char *) response = NULL;

    virAsprintf(&cmd, "containers/%s/update", id_str);

    if (!(json = virJSONValueToString(obj, false)))
        return -1;

    if (!(response = dockerApiRequest(priv, cmd, true,
                                      DOCKER_API_HTTP_POST,
                                      json, 200)))
        return -1;

    return 0;
}


static int
dockerContainerUpdate(dockerPrivatePtr priv,
                      const unsigned char *uuid,
                      virJSONValuePtr obj)
{
    VIR_AUTOFREE(char *) id_str = NULL;

    if (dockerSeekLabel(priv, uuid, false, &id_str) < 0)
        return -1;

    return dockerContainerUpdateById(priv, id_str, obj);
}


static int
dockerContainerDelete(dockerPrivatePtr priv, const char *id_str)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;

    virAsprintf(&cmd, "containers/%s", id_str);

    if (!(response = dockerApiRequest(priv, cmd, true,
                                      DOCKER_API_HTTP_DELETE,
                                      NULL, 204)))
        return -1;

    return 0;
}


static virJSONValuePtr
dockerGetContainerStats(dockerPrivatePtr priv, const unsigned char *uuid)
{
    VIR_AUTOFREE(char *) id_str = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    virJSONValuePtr obj;

    if (dockerSeekLabel(priv, uuid, false, &id_str) < 0)
        return NULL;

    virAsprintf(&cmd, "containers/%s/stats?stream=0", id_str);

    if (!(obj = dockerGetJSONObject(priv, cmd, true,
                                    DOCKER_API_HTTP_GET,
                                    NULL, 200)))
        return NULL;

    if (!virJSONValueIsObject(obj)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received not JSON object"));
        virJSONValueFree(obj);
        return NULL;
    }

    return obj;
}


static int
dockerFillDomainState(virJSONValuePtr obj,
                      virDomainStatsRecordPtr rec,
                      int *maxparams)
{
    virDomainState state = dockerGetContainerState(obj);

    if (virTypedParamsAddInt(&rec->params, &rec->nparams,
                             maxparams, "state.state", state) < 0 ||
        virTypedParamsAddInt(&rec->params, &rec->nparams,
                             maxparams, "state.reason", 0) < 0)
        return -1;

    return 0;
}


static int
dockerFillCpuStats(virJSONValuePtr stats,
                   virDomainStatsRecordPtr rec,
                   int *maxparams)
{
    virJSONValuePtr cpu_stats, cpu_usage;
    unsigned long long total, system, user;

    if (!(cpu_stats = virJSONValueObjectGet(stats, "cpu_stats")) ||
        !(cpu_usage = virJSONValueObjectGet(cpu_stats, "cpu_usage")) ||
        virJSONValueObjectGetNumberUlong(cpu_usage, "total_usage", &total) < 0 ||
        virJSONValueObjectGetNumberUlong(cpu_usage, "usage_in_kernelmode", &system) < 0 ||
        virJSONValueObjectGetNumberUlong(cpu_usage, "usage_in_usermode", &user) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'cpu_stats' object in JSON document"));
        return -1;
    }

    if (virTypedParamsAddULLong(&rec->params, &rec->nparams,
                                maxparams, "cpu.time",
                                total * 100) < 0 || /* usage in 100ns to nanoseconds */
        virTypedParamsAddULLong(&rec->params, &rec->nparams,
                                maxparams, "cpu.user",
                                user * 100) < 0 ||
        virTypedParamsAddULLong(&rec->params, &rec->nparams,
                                maxparams, "cpu.system",
                                system * 100) < 0)
        return -1;

    return 0;
}


static int
dockerFillMemoryStats(virJSONValuePtr stats,
                      virDomainStatsRecordPtr rec,
                      int *maxparams)
{
    virJSONValuePtr mem_stats;
    unsigned long long memory;

    if (!(mem_stats = virJSONValueObjectGet(stats, "memory_stats")) ||
        virJSONValueObjectGetNumberUlong(mem_stats, "privateworkingset", &memory) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'memory_stats' object in JSON document"));
        return -1;
    }

    if (virTypedParamsAddULLong(&rec->params, &rec->nparams,
                                maxparams, "balloon.current",
                                memory / 1024) < 0) /* byte to kilobyte */
        return -1;

    return 0;
}


static int
dockerFillVcpuStats(virJSONValuePtr stats,
                    virDomainStatsRecordPtr rec,
                    int *maxparams)
{
    unsigned int nvcpu;
    size_t i;

    if (virJSONValueObjectGetNumberUint(stats, "num_procs", &nvcpu) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'num_procs' field in JSON document"));
        return -1;
    }

    if (virTypedParamsAddUInt(&rec->params, &rec->nparams,
                              maxparams, "vcpu.current",
                              nvcpu) < 0 ||
        virTypedParamsAddUInt(&rec->params, &rec->nparams,
                              maxparams, "vcpu.maximum",
                              nvcpu) < 0)
        return -1;

    for (i = 0; i < nvcpu; i++) {
        char buf[VIR_TYPED_PARAM_FIELD_LENGTH];

        snprintf(buf, VIR_TYPED_PARAM_FIELD_LENGTH, "vcpu.%zu.state", i);
        if (virTypedParamsAddInt(&rec->params, &rec->nparams,
                                 maxparams, buf, VIR_VCPU_RUNNING) < 0)
            return -1;
    }

    return 0;
}


static int
dockerFillNetStats(virJSONValuePtr obj,
                   virJSONValuePtr stats,
                   virDomainStatsRecordPtr rec,
                   int *maxparams)
{
    virJSONValuePtr net_sets, networks;
    unsigned int count;
    size_t i;

    if (!(net_sets = virJSONValueObjectGet(obj, "NetworkSettings")) ||
        !(networks = virJSONValueObjectGet(net_sets, "Networks"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Networks' object in JSON container"));
        return -1;
    }

    count = virJSONValueObjectKeysNumber(networks);
    if (virTypedParamsAddUInt(&rec->params, &rec->nparams,
                              maxparams, "net.count",
                              count) < 0)
        return -1;

    for (i = 0; i < count; i++) {
        const char *name = virJSONValueObjectGetKey(networks, i);
        char buf[VIR_TYPED_PARAM_FIELD_LENGTH];

        snprintf(buf, VIR_TYPED_PARAM_FIELD_LENGTH, "net.%zu.name", i);
        if (virTypedParamsAddString(&rec->params, &rec->nparams,
                                    maxparams, buf, name) < 0)
            return -1;

        if (stats) {
            virJSONValuePtr iface;
            unsigned long long rx_bytes, tx_bytes, rx_pkts, tx_pkts,
                               rx_errs, tx_errs, rx_drop, tx_drop;

            if (!(networks = virJSONValueObjectGet(stats, "networks")) ||
                !(iface = virJSONValueObjectGetValue(networks, i))) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect 'networks' object in JSON document"));
                return -1;
            }

            if (virJSONValueObjectGetNumberUlong(iface, "rx_bytes", &rx_bytes) < 0 ||
                virJSONValueObjectGetNumberUlong(iface, "tx_bytes", &tx_bytes) < 0 ||
                virJSONValueObjectGetNumberUlong(iface, "rx_packets", &rx_pkts) < 0 ||
                virJSONValueObjectGetNumberUlong(iface, "tx_packets", &tx_pkts) < 0 ||
                virJSONValueObjectGetNumberUlong(iface, "rx_errors", &rx_errs) < 0 ||
                virJSONValueObjectGetNumberUlong(iface, "tx_errors", &tx_errs) < 0 ||
                virJSONValueObjectGetNumberUlong(iface, "rx_dropped", &rx_drop) < 0 ||
                virJSONValueObjectGetNumberUlong(iface, "tx_dropped", &tx_drop) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR,
                               _("Incorrect '%s' object in JSON document"),
                               virJSONValueObjectGetKey(networks, i));
                return -1;
            }

#define DOCKER_ADD_NET_PARAM(stname, stvalue) \
    do { \
        snprintf(buf, VIR_TYPED_PARAM_FIELD_LENGTH, "net.%zu.%s", i, (stname)); \
        if (virTypedParamsAddULLong(&rec->params, &rec->nparams, \
                                    maxparams, buf, (stvalue)) < 0) \
            return -1; \
    } while (0)

            DOCKER_ADD_NET_PARAM("rx.bytes", rx_bytes);
            DOCKER_ADD_NET_PARAM("rx.pkts", rx_pkts);
            DOCKER_ADD_NET_PARAM("rx.errs", rx_errs);
            DOCKER_ADD_NET_PARAM("rx.drop", rx_drop);
            DOCKER_ADD_NET_PARAM("tx.bytes", tx_bytes);
            DOCKER_ADD_NET_PARAM("tx.pkts", tx_pkts);
            DOCKER_ADD_NET_PARAM("tx.errs", tx_errs);
            DOCKER_ADD_NET_PARAM("tx.drop", tx_drop);

#undef DOCKER_ADD_NET_STATS
        }
    }

    return 0;
}


static int
dockerFillBlockStats(virJSONValuePtr obj,
                     virJSONValuePtr stats,
                     virDomainStatsRecordPtr rec,
                     int *maxparams)
{
    const char *path;

    if (virTypedParamsAddUInt(&rec->params, &rec->nparams,
                              maxparams, "block.count",
                              1) < 0 ||
        virTypedParamsAddString(&rec->params, &rec->nparams,
                                maxparams, "block.0.name",
                                "vda") < 0)
        return -1;

    if (!(path = virJSONValueObjectGetString(obj, "Image"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Image' field in JSON container"));
        return -1;
    }

    if (virTypedParamsAddString(&rec->params, &rec->nparams,
                                maxparams, "block.0.path",
                                path) < 0)
        return -1;

    if (stats) {
        virJSONValuePtr stor_stats;
        unsigned long long rcn, wcn, read, write;

        if (!(stor_stats = virJSONValueObjectGet(stats, "storage_stats")) ||
            virJSONValueObjectGetNumberUlong(stor_stats, "read_count_normalized", &rcn) < 0 ||
            virJSONValueObjectGetNumberUlong(stor_stats, "write_count_normalized", &wcn) < 0 ||
            virJSONValueObjectGetNumberUlong(stor_stats, "read_size_bytes", &read) < 0 ||
            virJSONValueObjectGetNumberUlong(stor_stats, "write_size_bytes", &write) < 0) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'storage_stats' object in JSON document"));
            return -1;
        }

        if (virTypedParamsAddULLong(&rec->params, &rec->nparams,
                                    maxparams, "block.0.rd.reqs",
                                    rcn) < 0 ||
            virTypedParamsAddULLong(&rec->params, &rec->nparams,
                                    maxparams, "block.0.rd.bytes",
                                    read) < 0 ||
            virTypedParamsAddULLong(&rec->params, &rec->nparams,
                                    maxparams, "block.0.wr.reqs",
                                    wcn) < 0 ||
            virTypedParamsAddULLong(&rec->params, &rec->nparams,
                                    maxparams, "block.0.wr.bytes",
                                    write) < 0)
            return -1;
    }

    return 0;
}


static int
dockerStatsJobHandler(char *response,
                      dockerApiParamPtr param,
                      dockerJobsControlPtr jobsctl)
{
    VIR_AUTOPTR(virJSONValue) ostats = NULL;
    int maxparams = 0;

    ostats = virJSONValueFromString(response);
    VIR_FREE(response);

    if (!ostats || !virJSONValueIsObject(ostats)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received not JSON object"));
        return -1;
    }

    if ((jobsctl->stats & VIR_DOMAIN_STATS_STATE) &&
        dockerFillDomainState(param->obj, param->rec, &maxparams) < 0)
        return -1;

    if ((jobsctl->stats & VIR_DOMAIN_STATS_CPU_TOTAL) &&
        dockerFillCpuStats(ostats, param->rec, &maxparams) < 0)
        return -1;

    if ((jobsctl->stats & VIR_DOMAIN_STATS_BALLOON) &&
        dockerFillMemoryStats(ostats, param->rec, &maxparams) < 0)
        return -1;

    if ((jobsctl->stats & VIR_DOMAIN_STATS_VCPU) &&
        dockerFillVcpuStats(ostats, param->rec, &maxparams) < 0)
        return -1;

    if ((jobsctl->stats & VIR_DOMAIN_STATS_INTERFACE) &&
        dockerFillNetStats(param->obj, ostats, param->rec, &maxparams) < 0)
        return -1;

    if ((jobsctl->stats & VIR_DOMAIN_STATS_BLOCK) &&
        dockerFillBlockStats(param->obj, ostats, param->rec, &maxparams) < 0)
        return -1;

    return 0;
}


static int
dockerStartStatsJob(dockerPrivatePtr priv,
                    virJSONValuePtr obj,
                    virDomainStatsRecordPtr rec,
                    dockerJobsControlPtr jobsctl)
{
    int ret = -1;
    VIR_AUTOFREE(char *) id_str = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    dockerApiParamPtr param = NULL;

    if (dockerSeekLabel(priv, rec->dom->uuid, false, &id_str) < 0)
        goto cleanup;

    virAsprintf(&cmd, "containers/%s/stats?stream=0", id_str);

    if (!(param = dockerPrepareApiRequest(priv, cmd, true,
                                          DOCKER_API_HTTP_GET,
                                          NULL, 200)))
        goto cleanup;

    param->obj = obj;
    param->rec = rec;
    if (dockerAddApiJob(jobsctl, param) < 0)
        goto cleanup;

    param = NULL;
    ret = 0;

 cleanup:
    dockerCloseApiRequest(param);

    return ret;
}


static virNetworkPtr
dockerGetNetworkFromObject(virConnectPtr conn, virJSONValuePtr obj)
{
    const char *name;
    unsigned char uuid[VIR_UUID_BUFLEN];

    if (!virJSONValueIsObject(obj)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Network not JSON object"));
        return NULL;
    }

    if (dockerGetUUID(obj, uuid) < 0)
        return NULL;

    if (!(name = virJSONValueObjectGetString(obj, "Name"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Name' field in JSON network"));
        return NULL;
    }

    return virGetNetwork(conn, name, uuid);
}


static virJSONValuePtr
dockerGetNetworkJSON(dockerPrivatePtr priv, const unsigned char *uuid)
{
    VIR_AUTOFREE(char *) id_str = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    virJSONValuePtr obj;

    if (dockerSeekLabel(priv, uuid, true, &id_str) < 0)
        return NULL;

    virAsprintf(&cmd, "networks/%s", id_str);

    if (!(obj = dockerGetJSONObject(priv, cmd, true,
                                    DOCKER_API_HTTP_GET,
                                    NULL, 200)))
        return NULL;

    if (!virJSONValueIsObject(obj)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Network not JSON object"));
        virJSONValueFree(obj);
        return NULL;
    }

    return obj;
}


static virDomainPtr
dockerLookupContainer(virConnectPtr conn,
                      const char *name,
                      const char *value)
{
    VIR_AUTOCLEAN(virBuffer) buf = VIR_BUFFER_INITIALIZER;
    VIR_AUTOPTR(virJSONValue) values = NULL;
    VIR_AUTOPTR(virJSONValue) filter = NULL;
    VIR_AUTOPTR(virJSONValue) list = NULL;
    VIR_AUTOFREE(char *) filter_str = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    virJSONValuePtr obj = NULL;

    values = virJSONValueNewArray();
    filter = virJSONValueNewObject();

    if (virJSONValueArrayAppendString(values, value) < 0 ||
        virJSONValueObjectAppend(filter, name, values) < 0)
        return NULL;

    values = NULL;

    if (!(filter_str = virJSONValueToString(filter, false)))
        return NULL;

    virBufferAddLit(&buf, "containers/json?all=1&filters=");
    virBufferURIEncodeString(&buf, filter_str);
    if (virBufferCheckError(&buf) < 0 ||
        !(cmd = virBufferContentAndReset(&buf)))
        return NULL;

    if (!(list = dockerGetJSONObject(conn->privateData, cmd, true,
                                     DOCKER_API_HTTP_GET, NULL, 200)))
        return NULL;

    if (!virJSONValueIsArray(list)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received not JSON array"));
        return NULL;
    }

    if (virJSONValueArraySize(list) > 1 && STRNEQ(name, "name")) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Looked up more than one container"));
        return NULL;
    }

    /*
     * Docker can find more than one container by name "word" if
     * several containers have names like "word1", "words" and so on.
     */
    if (virJSONValueArraySize(list) && STREQ(name, "name")) {
        const char *cname;
        size_t i;

        for (i = 0; i < virJSONValueArraySize(list); i++) {
            virJSONValuePtr names, first_name;

            obj = virJSONValueArrayGet(list, i);
            if (!(names = virJSONValueObjectGet(obj, "Names")) ||
                !virJSONValueIsArray(names) ||
                !(first_name = virJSONValueArrayGet(names, 0)) ||
                !(cname = virJSONValueGetString(first_name))) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect 'Names' field in JSON container"));
                return NULL;
            }

            if (STREQ(value, cname + 1))
                break;
        }

        if (i >= virJSONValueArraySize(list))
            return NULL;
    } else {
        if (!(obj = virJSONValueArrayGet(list, 0)))
            return NULL;
    }

    /* Container looked up by id must be active */
    if (STREQ(name, "id") && strlen(value) < 12) {
        virDomainState state = dockerGetContainerState(obj);
        if ((state != VIR_DOMAIN_RUNNING) &&
            (state != VIR_DOMAIN_PAUSED)) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Looked up inactive container by id"));
            return NULL;
        }
    }

    return dockerGetDomainFromObject(conn, obj);
}


static virJSONValuePtr
dockerGetContainersList(dockerPrivatePtr priv, unsigned int flags)
{
    VIR_AUTOCLEAN(virBuffer) buf = VIR_BUFFER_INITIALIZER;
    VIR_AUTOPTR(virJSONValue) statelist = NULL;
    VIR_AUTOPTR(virJSONValue) filter = NULL;
    VIR_AUTOFREE(char *) filter_str = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    virJSONValuePtr list;

    /*
     * Check for filter combinations that return no results:
     * persistent: all docker containers are persistent
     * snapshot: the driver does not support snapshot management
     * managedsave: the driver does not support managedsave domains
     */
    if (((flags & VIR_CONNECT_LIST_DOMAINS_TRANSIENT) &&
         !(flags & VIR_CONNECT_LIST_DOMAINS_PERSISTENT)) ||
        ((flags & VIR_CONNECT_LIST_DOMAINS_HAS_SNAPSHOT) &&
         !(flags & VIR_CONNECT_LIST_DOMAINS_NO_SNAPSHOT)) ||
        ((flags & VIR_CONNECT_LIST_DOMAINS_MANAGEDSAVE) &&
         !(flags & VIR_CONNECT_LIST_DOMAINS_NO_MANAGEDSAVE)) ||
        ((flags & VIR_CONNECT_LIST_DOMAINS_HAS_CHECKPOINT) &&
         !(flags & VIR_CONNECT_LIST_DOMAINS_NO_CHECKPOINT))) {
        return virJSONValueNewArray();
    }

    virBufferAddLit(&buf, "containers/json?all=1");

    if ((!(flags & VIR_CONNECT_LIST_DOMAINS_INACTIVE) &&
         (flags & VIR_CONNECT_LIST_DOMAINS_ACTIVE)) ||
        (!(flags & VIR_CONNECT_LIST_DOMAINS_ACTIVE) &&
         (flags & VIR_CONNECT_LIST_DOMAINS_INACTIVE))) {
        statelist = virJSONValueNewArray();
        filter = virJSONValueNewObject();

        if (flags & VIR_CONNECT_LIST_DOMAINS_INACTIVE) {
            if ((virJSONValueArrayAppendString(statelist, "created") < 0) ||
                (virJSONValueArrayAppendString(statelist, "exited") < 0) ||
                (virJSONValueArrayAppendString(statelist, "removing") < 0) ||
                (virJSONValueArrayAppendString(statelist, "dead") < 0))
                return NULL;
        }

        if (flags & VIR_CONNECT_LIST_DOMAINS_ACTIVE) {
            if ((virJSONValueArrayAppendString(statelist, "running") < 0) ||
                (virJSONValueArrayAppendString(statelist, "restarting") < 0) ||
                (virJSONValueArrayAppendString(statelist, "paused") < 0))
                return NULL;
        }

        if (virJSONValueObjectAppend(filter, "status", statelist) < 0)
            return NULL;

        statelist = NULL;

        if (!(filter_str = virJSONValueToString(filter, false)))
            return NULL;

        virBufferAddLit(&buf, "&filters=");
        virBufferURIEncodeString(&buf, filter_str);
    }

    if (virBufferCheckError(&buf) < 0 ||
        !(cmd = virBufferContentAndReset(&buf)))
        return NULL;

    if (!(list = dockerGetJSONObject(priv, cmd, true,
                                     DOCKER_API_HTTP_GET,
                                     NULL, 200)))
        return NULL;

    if (!virJSONValueIsArray(list)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received not JSON array"));
        virJSONValueFree(list);
        return NULL;
    }

    return list;
}


VIR_DEFINE_AUTOPTR_FUNC(virDomainDiskDef, virDomainDiskDefFree);
VIR_DEFINE_AUTOPTR_FUNC(virDomainNetDef, virDomainNetDefFree);
VIR_DEFINE_AUTOPTR_FUNC(virDomainFSDef, virDomainFSDefFree);
VIR_DEFINE_AUTOPTR_FUNC(virDomainDef, virDomainDefFree);


static int
dockerAddNetToDef(virDomainDefPtr def,
                  virJSONValuePtr obj,
                  virJSONValuePtr host_conf,
                  virJSONValuePtr config)
{
    VIR_AUTOPTR(virDomainNetDef) net = NULL;
    virJSONValuePtr net_max, net_sets, networks;
    unsigned long long bandwidth = 0;
    size_t i;

    /*
     * Don't handle absence of field because standart Docker daemon
     * can't give us field 'NetMaximumBandwidth'.
     */
    if ((net_max = virJSONValueObjectGet(host_conf, "NetMaximumBandwidth"))) {
        if (virJSONValueGetNumberUlong(net_max, &bandwidth) < 0) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'NetMaximumBandwidth' object in "
                             "JSON container"));
            return -1;
        }
    }

    if (!(net_sets = virJSONValueObjectGet(obj, "NetworkSettings"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'NetworkSettings' object in JSON container"));
        return -1;
    }

    if (!(networks = virJSONValueObjectGet(net_sets, "Networks")) ||
        virJSONValueObjectKeysNumber(networks) < 1) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Networks' object in JSON container"));
        return -1;
    }

    for (i = 0; i < virJSONValueObjectKeysNumber(networks); i++) {
        virJSONValuePtr iface = virJSONValueObjectGetValue(networks, i);
        const char *macaddr, *ip_str, *ip6_str, *gateway, *gateway6;
        unsigned int prefix, prefix6;

        if (!(macaddr = virJSONValueObjectGetString(iface, "MacAddress"))) {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Incorrect '%s' object in JSON container"),
                           virJSONValueObjectGetKey(networks, i));
            return -1;
        }

        VIR_ALLOC(net);
        net->type = VIR_DOMAIN_NET_TYPE_NETWORK;
        VIR_STRDUP(net->data.network.name,
                   virJSONValueObjectGetKey(networks, i));

        /*
         * For first interface mac address can be preset in config object
         * instead of network object. For other interfaces it can be present
         * only in network object because other interfaces was added to
         * existing container late.
         */
        if (macaddr[0] ||
            (!i &&
             (macaddr = virJSONValueObjectGetString(config, "MacAddress")))) {
            if (virMacAddrParse(macaddr, &net->mac) < 0) {
                virReportError(VIR_ERR_INVALID_MAC, "%s", macaddr);
                return -1;
            }
        }

        if (virJSONValueObjectGetNumberUint(iface, "IPPrefixLen", &prefix)) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'IPPrefixLen' object in JSON container"));
            return -1;
        }

        if (!(ip_str = virJSONValueObjectGetString(iface, "IPAddress"))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'IPAddress' object in JSON container"));
            return -1;
        }

        if (ip_str[0]) {
            virNetDevIPAddrPtr ip_addr = NULL;

            VIR_ALLOC(ip_addr);
            VIR_APPEND_ELEMENT_COPY(net->guestIP.ips, net->guestIP.nips, ip_addr);

            if (virSocketAddrParse(&ip_addr->address, ip_str, AF_INET) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect 'IPAddress' object in JSON container"));
                return -1;
            }

            ip_addr->prefix = prefix;
        }

        if (!(gateway = virJSONValueObjectGetString(iface, "Gateway"))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'Gateway' object in JSON container"));
            return -1;
        }

        if (gateway[0]) {
            virNetDevIPRoutePtr route;

            VIR_ALLOC(route);
            VIR_APPEND_ELEMENT_COPY(net->guestIP.routes, net->guestIP.nroutes, route);

            if (virSocketAddrParse(&route->gateway, gateway, AF_INET) < 0 ||
                virSocketAddrMaskByPrefix(&route->gateway, prefix, &route->address) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect 'Gateway' object in JSON container"));
                return -1;
            }

            VIR_STRDUP(route->family, "ipv4");
            route->prefix = prefix;
            route->has_prefix = true;
        }

        if (virJSONValueObjectGetNumberUint(iface, "GlobalIPv6PrefixLen", &prefix6)) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'GlobalIPv6PrefixLen' object in JSON container"));
            return -1;
        }

        if (!(ip6_str = virJSONValueObjectGetString(iface, "GlobalIPv6Address"))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'GlobalIPv6Address' object in JSON container"));
            return -1;
        }

        if (ip6_str[0]) {
            virNetDevIPAddrPtr ip6_addr = NULL;

            VIR_ALLOC(ip6_addr);
            VIR_APPEND_ELEMENT_COPY(net->guestIP.ips, net->guestIP.nips, ip6_addr);

            if (virSocketAddrParse(&ip6_addr->address, ip6_str, AF_INET6) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect 'GlobalIPv6Address' object in JSON container"));
                return -1;
            }

            ip6_addr->prefix = prefix6;
        }

        if (!(gateway6 = virJSONValueObjectGetString(iface, "IPv6Gateway"))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'IPv6Gateway' object in JSON container"));
            return -1;
        }

        if (gateway6[0]) {
            virNetDevIPRoutePtr route6;

            VIR_ALLOC(route6);
            VIR_APPEND_ELEMENT_COPY(net->guestIP.routes, net->guestIP.nroutes, route6);

            if (virSocketAddrParse(&route6->gateway, gateway6, AF_INET6) < 0 ||
                virSocketAddrMaskByPrefix(&route6->gateway, prefix6, &route6->address) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect 'IPv6Gateway' object in JSON container"));
                return -1;
            }

            VIR_STRDUP(route6->family, "ipv6");
            route6->prefix = prefix6;
            route6->has_prefix = true;
        }

        if (bandwidth) {
            VIR_ALLOC(net->bandwidth);
            VIR_ALLOC(net->bandwidth->out);
            net->bandwidth->out->average = bandwidth / 1024; /* byte to kilobyte */
        }

        if (virDomainNetInsert(def, net) < 0)
            return -1;

        net = NULL;
    }

    return 0;
}


static int
dockerAddFSToDef(virDomainDefPtr def, virJSONValuePtr obj)
{
    virJSONValuePtr mounts;
    size_t i;

    if (!(mounts = virJSONValueObjectGet(obj, "Mounts")) ||
        !virJSONValueIsArray(mounts)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Mounts' array in JSON container"));
        return -1;
    }

    for (i = 0; i < virJSONValueArraySize(mounts); i++) {
        VIR_AUTOPTR(virDomainFSDef) fs = virDomainFSDefNew();
        virJSONValuePtr item = virJSONValueArrayGet(mounts, i);
        const char *type;
        const char *source;
        const char *dest;
        bool rw;

        if (!(type = virJSONValueObjectGetString(item, "Type")) ||
            !(source = virJSONValueObjectGetString(item, "Source")) ||
            !(dest = virJSONValueObjectGetString(item, "Destination")) ||
            virJSONValueObjectGetBoolean(item, "RW", &rw) < 0) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'Mounts' item in JSON container"));
            return -1;
        }

        if (STREQ(type, "bind")) {
            VIR_STRDUP(fs->src->path, source);
            fs->type = VIR_DOMAIN_FS_TYPE_MOUNT;
        } else if (STREQ(type, "volume")) {
            const char *volume;
            const char *pool;

            if (!(volume = virJSONValueObjectGetString(item, "Name")) ||
                !(pool = virJSONValueObjectGetString(item, "Driver"))) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect 'Name' field for item in JSON container"));
                return -1;
            }

            VIR_ALLOC(fs->src->srcpool);
            VIR_STRDUP(fs->src->srcpool->pool, pool);
            VIR_STRDUP(fs->src->srcpool->volume, volume);
            fs->type = VIR_DOMAIN_FS_TYPE_VOLUME;
        } else {
            VIR_WARN("Unsupported type '%s' for filesystem", type);
            continue;
        }

        VIR_STRDUP(fs->dst, dest);
        fs->readonly = !rw;

        if (virDomainFSInsert(def, fs) < 0)
            return -1;

        fs = NULL;
    }

    return 0;
}


static unsigned long long
dockerCalcPeriod(unsigned long long nanocpus, long long *quota)
{
    unsigned long long period;

    if (*quota > 0) {
        /* Restore source CPU period by computing it from nanocpus and quota */
        period = (*quota * 1000000000) / nanocpus;
    } else {
        /* Set CPU period and quota by computing it from nanocpus */
        period = 1000000; /* use max possible density for period */
        *quota = nanocpus > 1000000 ? nanocpus / 1000 : 1000;
    }

    return period;
}


static long long
dockerCalcNanoCpus(dockerPrivatePtr priv,
                   unsigned long long *period,
                   long long quota,
                   bool update)
{
    long long nanocpus = 0;

    /*
     * If we receive CPU period and quota we will compute and set
     * 'NanoCpus' option for docker. Also we always set quota value
     * in 'CpuQuota' option because docker save it option and we can
     * restore CPU period by computing it from 'NanoCpus' and quota.
     */
    if (*period && quota > 0)
        nanocpus = (quota * 1000000000) / *period;
    else if (update && quota < 0)
        nanocpus = -1;

    /*
     * Docker for Windows doesn't support CPU period in container
     * configuration so we cannot leave non-zero period value
     */
    if (*period && priv->os == DOCKER_OS_WINDOWS)
        *period = 0;

    return nanocpus;
}


static virDomainDefPtr
dockerJSONToDomainDef(dockerPrivatePtr priv, virJSONValuePtr obj)
{
    virDomainDefPtr ret;
    VIR_AUTOPTR(virDomainDef) def = NULL;
    VIR_AUTOPTR(virDomainDiskDef) disk = NULL;
    VIR_AUTOPTR(virJSONValue) info = NULL;
    virJSONValuePtr config, cmd, env, host_conf, dns, searchs;
    const char *name, *image, *initdir, *inituser, *hostname,
               *domainname, *netmode;
    unsigned long long memory, nanocpus;
    unsigned int maxvcpu;

    if (!(def = virDomainDefNew()))
        return NULL;

    def->virtType = VIR_DOMAIN_VIRT_DOCKER;
    def->os.type = VIR_DOMAIN_OSTYPE_EXE;
    def->os.arch = priv->hostArch;
    def->clock.offset = VIR_DOMAIN_CLOCK_OFFSET_LOCALTIME;
    def->onReboot = VIR_DOMAIN_LIFECYCLE_ACTION_RESTART;

    if (!(name = virJSONValueObjectGetString(obj, "Name"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Name' field in JSON container"));
        return NULL;
    }

    /* First symbol in docker name always slash so we can skip it */
    VIR_STRDUP(def->name, name + 1);

    if (dockerGetUUID(obj, def->uuid) < 0)
        return NULL;

    if ((def->id = dockerGetId(obj)) < -1)
        return NULL;

    if (!(config = virJSONValueObjectGet(obj, "Config")) ||
        !(image = virJSONValueObjectGetString(config, "Image"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Config' object in JSON container"));
        return NULL;
    }

    /* Get some optional parameters of container */
    initdir = virJSONValueObjectGetString(config, "WorkingDir");
    if (!virStringIsEmpty(initdir))
        VIR_STRDUP(def->os.initdir, initdir);

    inituser = virJSONValueObjectGetString(config, "User");
    if (!virStringIsEmpty(inituser))
        VIR_STRDUP(def->os.inituser, inituser);

    if ((cmd = virJSONValueObjectGetArray(config, "Cmd"))) {
        size_t i, nargv;

        nargv = virJSONValueArraySize(cmd);
        if (nargv > 1)
            VIR_ALLOC_N(def->os.initargv, nargv);

        for (i = 0; i < nargv; i++) {
            virJSONValuePtr item = virJSONValueArrayGet(cmd, i);
            const char *argp = virJSONValueGetString(item);

            if (virStringIsEmpty(argp)) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect 'Cmd' object in JSON container"));
                return NULL;
            }

            if (i)
                VIR_STRDUP(def->os.initargv[i - 1], argp);
            else
                VIR_STRDUP(def->os.init, argp);
        }
    }

    if (!def->os.init)
        VIR_STRDUP(def->os.init, "*");

    if ((env = virJSONValueObjectGetArray(config, "Env"))) {
        size_t i, nenv;

        nenv = virJSONValueArraySize(env);
        if (nenv)
            VIR_ALLOC_N(def->os.initenv, nenv + 1);

        for (i = 0; i < nenv; i++) {
            virJSONValuePtr item = virJSONValueArrayGet(env, i);
            const char *envp = virJSONValueGetString(item);
            char **envparts;

            if (virStringIsEmpty(envp)) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect 'Env' object in JSON container"));
                return NULL;
            }

            envparts = virStringSplit(envp, "=", 2);
            VIR_ALLOC(def->os.initenv[i]);
            VIR_STRDUP(def->os.initenv[i]->name, envparts[0]);
            VIR_STRDUP(def->os.initenv[i]->value,
                       envparts[1] ? envparts[1] : "");
            virStringListFree(envparts);
        }
    }

    hostname = virJSONValueObjectGetString(config, "Hostname");
    if (!virStringIsEmpty(hostname)) {
        domainname = virJSONValueObjectGetString(config, "Domainname");
        if (virStringIsEmpty(domainname))
            VIR_STRDUP(def->dns.hostname, hostname);
        else
            virAsprintf(&def->dns.hostname, "%s.%s", hostname,
                        domainname);
    }

    if (!(host_conf = virJSONValueObjectGet(obj, "HostConfig")) ||
        virJSONValueObjectGetNumberUint(host_conf, "CpuCount", &maxvcpu) < 0 ||
        virJSONValueObjectGetNumberUlong(host_conf, "CpuShares", &def->cputune.shares) < 0 ||
        virJSONValueObjectGetNumberUlong(host_conf, "CpuPeriod", &def->cputune.period) < 0 ||
        virJSONValueObjectGetNumberLong(host_conf, "CpuQuota", &def->cputune.quota) < 0 ||
        virJSONValueObjectGetNumberUlong(host_conf, "NanoCpus", &nanocpus) < 0 ||
        virJSONValueObjectGetNumberUlong(host_conf, "Memory", &memory) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'HostConfig' object in JSON container"));
        return NULL;
    }

    def->maxvcpus = maxvcpu;
    if (def->cputune.shares)
        def->cputune.sharesSpecified = true;
    if (nanocpus)
        def->cputune.period = dockerCalcPeriod(nanocpus, &def->cputune.quota);

    if (!memory) {
        if (!(info = dockerGetJSONObject(priv, "info", true,
                                         DOCKER_API_HTTP_GET,
                                         NULL, 200)))
            return NULL;

        if (!memory &&
            virJSONValueObjectGetNumberUlong(info, "MemTotal", &memory) < 0) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'MemTotal' field in JSON document"));
        }
    }

    virDomainDefSetMemoryTotal(def, memory / 1024); /* byte to kilobyte */
    if (def->id > 0)
        def->mem.cur_balloon = memory / 1024;

    if ((dns = virJSONValueObjectGetArray(host_conf, "Dns"))) {
        size_t i, ndns;

        ndns = virJSONValueArraySize(dns);
        if (ndns)
            VIR_EXPAND_N(def->dns.servers, def->dns.nservers, ndns);

        for (i = 0; i < ndns; i++) {
            virJSONValuePtr item = virJSONValueArrayGet(dns, i);
            const char *str = virJSONValueGetString(item);

            if (virSocketAddrParse(&def->dns.servers[i], str, AF_UNSPEC) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect 'Dns' object in JSON container"));
                return NULL;
            }
        }
    }

    if ((searchs = virJSONValueObjectGetArray(host_conf, "DnsSearch"))) {
        size_t i, nsearchs;

        nsearchs = virJSONValueArraySize(searchs);
        if (nsearchs)
            VIR_ALLOC_N(def->dns.searchs, nsearchs + 1);

        for (i = 0; i < nsearchs; i++) {
            virJSONValuePtr item = virJSONValueArrayGet(searchs, i);
            const char *str = virJSONValueGetString(item);
            if (virStringIsEmpty(str)) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect 'DnsSearch' object in JSON container"));
                return NULL;
            }

            VIR_STRDUP(def->dns.searchs[i], str);
        }
    }

    if (!(disk = virDomainDiskDefNew(priv->xmlOpt)) ||
        virDomainDiskSetSource(disk, image) < 0 ||
        virDomainDiskSetDriver(disk, "docker") < 0)
        return NULL;

    virDomainDiskSetType(disk, VIR_STORAGE_TYPE_TEMPLATE);
    VIR_STRDUP(disk->dst, "vda"); /* set fake dev */
    disk->bus = VIR_DOMAIN_DISK_BUS_VIRTIO; /* set fake bus without controller */

    if (virJSONValueObjectGetNumberUlong(host_conf, "IOMaximumBandwidth",
                                         &disk->blkdeviotune.total_bytes_sec) < 0 ||
        virJSONValueObjectGetNumberUlong(host_conf, "IOMaximumIOps",
                                         &disk->blkdeviotune.total_iops_sec) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect I/O maximum fields in JSON document"));
        return NULL;
    }

    if (virDomainDiskInsert(def, disk) < 0)
        return NULL;

    disk = NULL;

    if (!(netmode = virJSONValueObjectGetString(host_conf, "NetworkMode"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'NetworkMode' field in JSON container"));
        return NULL;
    }

    if (STRNEQ(netmode, "none") &&
        dockerAddNetToDef(def, obj, host_conf, config) < 0)
        return NULL;

    if (dockerAddFSToDef(def, obj) < 0)
        return NULL;

    VIR_STEAL_PTR(ret, def);

    return ret;
}


static virJSONValuePtr
dockerNetworkDefToJSON(virNetDevIPInfoPtr def, const char *macaddr)
{
    virJSONValuePtr obj;
    VIR_AUTOPTR(virJSONValue) ipam_conf = NULL;
    VIR_AUTOFREE(char *) ip_str = NULL;
    VIR_AUTOFREE(char *) ip6_str = NULL;
    VIR_AUTOFREE(char *) gateway = NULL;
    VIR_AUTOFREE(char *) gateway6 = NULL;
    int prefix = 0;
    int prefix6 = 0;
    size_t i;

    for (i = 0; i < def->nips; i++) {
        virNetDevIPAddrPtr ip = def->ips[i];

        if (VIR_SOCKET_ADDR_IS_FAMILY(&ip->address, AF_INET)) {
            if (ip_str) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Supported only one IP"));
                return NULL;
            }
            ip_str = virSocketAddrFormat(&ip->address);
            prefix = ip->prefix;
        } else if (VIR_SOCKET_ADDR_IS_FAMILY(&ip->address, AF_INET6)) {
            if (ip6_str) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Supported only one IPv6"));
                return NULL;
            }
            ip6_str = virSocketAddrFormat(&ip->address);
            prefix6 = ip->prefix;
        }
    }

    for (i = 0; i < def->nroutes; i++) {
        virNetDevIPRoutePtr route = def->routes[i];
        int prefix_route;

        if (VIR_SOCKET_ADDR_IS_FAMILY(&route->address, AF_INET)) {
            if (gateway) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Supported only one gateway"));
                return NULL;
            }

            if ((prefix_route = virNetDevIPRouteGetPrefix(route)) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect value for netmask"));
                return NULL;
            }

            if (!prefix) {
                prefix = prefix_route;
            } else if (prefix != prefix_route) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Prefixes for <ip> and <route> are not equal"));
                return NULL;
            }

            gateway = virSocketAddrFormat(&route->gateway);
        } else if (VIR_SOCKET_ADDR_IS_FAMILY(&route->address, AF_INET6)) {
            if (gateway6) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Supported only one IPv6 gateway"));
                return NULL;
            }

            if ((prefix_route = virNetDevIPRouteGetPrefix(route)) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect value for IPv6 netmask"));
                return NULL;
            }

            if (!prefix6) {
                prefix6 = prefix_route;
            } else if (prefix6 != prefix_route) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Prefixes for <ip> and <route> are not equal"));
                return NULL;
            }

            gateway6 = virSocketAddrFormat(&route->gateway);
        }
    }

    if ((ip_str || ip6_str) &&
        virJSONValueObjectCreate(&ipam_conf,
                                 "S:IPv4Address", ip_str,
                                 "S:IPv6Address", ip6_str,
                                 NULL) < 0)
        return NULL;

    virJSONValueObjectCreate(&obj,
                             "S:Gateway", gateway,
                             "S:IPv6Gateway", gateway6,
                             "S:IPAddress", ip_str,
                             "S:GlobalIPv6Address", ip6_str,
                             "u:IPPrefixLen", prefix,
                             "u:GlobalIPv6PrefixLen", prefix6,
                             "A:IPAMConfig", &ipam_conf,
                             "S:MacAddress", macaddr,
                             NULL);

    return obj;
}


static int
dockerCheckImage(dockerPrivatePtr priv, const char *image)
{
    VIR_AUTOCLEAN(virBuffer) buf = VIR_BUFFER_INITIALIZER;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;

    virAsprintf(&cmd, "images/%s/json", image);

    if ((response = dockerApiRequest(priv, cmd, true,
                                     DOCKER_API_HTTP_GET,
                                     NULL, 200)))
        return 0;

    VIR_FREE(cmd);
    virAsprintf(&cmd, "images/create?fromImage=%s", image);

    if (!(response = dockerApiRequest(priv, cmd, true,
                                      DOCKER_API_HTTP_POST,
                                      NULL, 200)))
        return -1;

    return 0;
}


static unsigned long long
dockerNetBandwidth(virDomainDefPtr def)
{
    size_t i;

    for (i = 0; i < def->nnets; i++) {
        if (def->nets[i]->bandwidth && def->nets[i]->bandwidth->out->average)
            /* kilobyte to byte */
            return def->nets[i]->bandwidth->out->average * 1024;
    }

    return 0;
}


static virJSONValuePtr
dockerDomainDefToJSON(dockerPrivatePtr priv,
                      virDomainDefPtr def,
                      bool interactive)
{
    virJSONValuePtr obj;
    VIR_AUTOPTR(virJSONValue) cmd = NULL;
    VIR_AUTOPTR(virJSONValue) env = NULL;
    VIR_AUTOPTR(virJSONValue) dns = NULL;
    VIR_AUTOPTR(virJSONValue) search = NULL;
    VIR_AUTOPTR(virJSONValue) labels = NULL;
    VIR_AUTOPTR(virJSONValue) mounts = NULL;
    VIR_AUTOPTR(virJSONValue) log_conf = NULL;
    VIR_AUTOPTR(virJSONValue) stor_opt = NULL;
    VIR_AUTOPTR(virJSONValue) restart_pol = NULL;
    VIR_AUTOPTR(virJSONValue) host_conf = NULL;
    VIR_AUTOPTR(virJSONValue) net_conf = NULL;
    VIR_AUTOFREE(char *) hostname = NULL;
    VIR_AUTOFREE(char *) domainname = NULL;
    unsigned long long memory, nanocpus;
    unsigned int maxvcpu;
    const char *netmode;
    char macaddr[VIR_MAC_STRING_BUFLEN] = { '\0', };
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    /*
     * If we want to set container command line as it defined
     * in base image we will define <init> as "*". In this case
     * we don't assign any value to command line in configuration
     * JSON at all.
     */
    if (def->os.init && STRNEQ(def->os.init, "*")) {
        size_t i;

        cmd = virJSONValueNewArray();
        if (virJSONValueArrayAppendString(cmd, def->os.init) < 0)
            return NULL;

        for (i = 0; def->os.initargv && def->os.initargv[i]; i++) {
            if (virJSONValueArrayAppendString(cmd, def->os.initargv[i]) < 0)
                return NULL;
        }
    }

    if (def->os.initenv) {
        size_t i;

        env = virJSONValueNewArray();

        for (i = 0; def->os.initenv[i]; i++) {
            VIR_AUTOFREE(char *) envp = NULL;

            virAsprintf(&envp, "%s=%s", def->os.initenv[i]->name,
                        def->os.initenv[i]->value);
            if (virJSONValueArrayAppendString(env, envp) < 0)
                return NULL;
        }
    }

    if (def->dns.servers) {
        size_t i;

        dns = virJSONValueNewArray();
        for (i = 0; i < def->dns.nservers; i++) {
            VIR_AUTOFREE(char *) str = virSocketAddrFormat(&def->dns.servers[i]);
            if (virJSONValueArrayAppendString(dns, str) < 0)
                return NULL;
        }
    }

    if (def->dns.searchs) {
        size_t i;

        search = virJSONValueNewArray();
        for (i = 0; def->dns.searchs[i]; i++) {
            if (virJSONValueArrayAppendString(search, def->dns.searchs[i]) < 0)
                return NULL;
        }
    }

    virUUIDFormat(def->uuid, uuidstr);
    if (virJSONValueObjectCreate(&labels,
                                 "s:"LABEL_LIBVIRT_UUID, uuidstr,
                                 NULL) < 0)
        return NULL;

    if (def->nfss) {
        size_t i;

        mounts = virJSONValueNewArray();

        for (i = 0; i < def->nfss; i++) {
            virDomainFSDefPtr fs = def->fss[i];
            VIR_AUTOPTR(virJSONValue) item = NULL;
            const char *type;
            const char *source;

            if (fs->type == VIR_DOMAIN_FS_TYPE_VOLUME) {
                type = "volume";
                source = fs->src->srcpool->volume;
            } else {
                type = "bind";
                source = fs->src->path;
            }

            if (virJSONValueObjectCreate(&item,
                                         "s:Type", type,
                                         "s:Source", source,
                                         "s:Target", fs->dst,
                                         "b:ReadOnly", fs->readonly,
                                         NULL) < 0)
                return NULL;

            if (virJSONValueArrayAppend(mounts, item) < 0)
                return NULL;

            item = NULL;
        }
    }

    if (!def->nnets) {
        netmode = "none";
    } else {
        VIR_AUTOPTR(virJSONValue) iface = NULL;
        VIR_AUTOPTR(virJSONValue) ep_conf = NULL;

        if (!(iface = dockerNetworkDefToJSON(&def->nets[0]->guestIP, NULL)))
            return NULL;

        ep_conf = virJSONValueNewObject();
        netmode = def->nets[0]->data.network.name;
        if (virJSONValueObjectAppend(ep_conf, netmode, iface) < 0)
            return NULL;

        iface = NULL;

        if (virJSONValueObjectCreate(&net_conf,
                                     "a:EndpointsConfig", &ep_conf,
                                     NULL) < 0)
            return NULL;

        virMacAddrFormat(&def->nets[0]->mac, macaddr);
    }

    if (virJSONValueObjectCreate(&log_conf,
                                 "s:Type", "", /* default: "json-file" */
                                 "n:Config",
                                 NULL) < 0)
        return NULL;

    if (virJSONValueObjectCreate(&stor_opt,
                                 "s:size", "20G",
                                 NULL) < 0)
        return NULL;

    if (virJSONValueObjectCreate(&restart_pol,
                                 "s:Name", "no",
                                 "u:MaximumRetryCount", 0,
                                 NULL) < 0)
        return NULL;

    maxvcpu = virDomainDefGetVcpusMax(def);

    nanocpus = dockerCalcNanoCpus(priv, &def->cputune.period,
                                  def->cputune.quota, false);

    memory = virDomainDefGetMemoryTotal(def) * 1024; /* kilobyte to byte */

    if (def->dns.hostname) {
        const char *dot;

        if ((dot = strchr(def->dns.hostname, '.'))) {
           VIR_STRNDUP(hostname, def->dns.hostname, dot - def->dns.hostname + 1);
           hostname[dot - def->dns.hostname] = '\0';
           VIR_STRDUP(domainname, dot + 1);
        } else {
           VIR_STRDUP(hostname, def->dns.hostname);
        }
    }

    if (virJSONValueObjectCreate(&host_conf,
                                 "b:AutoRemove", false,
                                 "s:NetworkMode", netmode,
                                 "u:CpuCount", maxvcpu,
                                 "P:CpuShares", def->cputune.shares,
                                 "P:CpuPeriod", def->cputune.period,
                                 "Z:CpuQuota", def->cputune.quota,
                                 "P:NanoCpus", nanocpus,
                                 "U:Memory", memory,
                                 "A:Mounts", &mounts,
                                 "a:LogConfig", &log_conf,
                                 "a:StorageOpt", &stor_opt,
                                 "a:RestartPolicy", &restart_pol,
                                 "A:Dns", &dns,
                                 "A:DnsSearch", &search,
                                 "P:IOMaximumBandwidth",
                                 def->disks[0]->blkdeviotune.total_bytes_sec,
                                 "P:IOMaximumIOps",
                                 def->disks[0]->blkdeviotune.total_iops_sec,
                                 "P:NetMaximumBandwidth",
                                 dockerNetBandwidth(def),
                                 NULL) < 0)
        return NULL;

    if (dockerCheckImage(priv, virDomainDiskGetSource(def->disks[0])) < 0)
        return NULL;

    virJSONValueObjectCreate(&obj,
                             "b:ArgsEscaped", true,
                             "b:AttachStdin", interactive,
                             "b:AttachStdout", interactive,
                             "b:AttachStderr", interactive,
                             "b:OpenStdin", interactive,
                             "b:StdinOnce", interactive,
                             "b:Tty", interactive,
                             "n:OnBuild",
                             "n:Entrypoint",
                             "S:Hostname", hostname,
                             "S:Domainname", domainname,
                             "S:User", def->os.inituser,
                             "S:WorkingDir", def->os.initdir,
                             "s:Image", virDomainDiskGetSource(def->disks[0]),
                             "A:Cmd", &cmd,
                             "A:Env", &env,
                             "a:Labels", &labels,
                             "a:HostConfig", &host_conf,
                             "A:NetworkingConfig", &net_conf,
                             "S:MacAddress", macaddr[0] ? macaddr : NULL,
                             NULL);

    return obj;
}


static int
dockerAddNetsToContainer(dockerPrivatePtr priv,
                         virDomainDefPtr def,
                         const char *id_str)
{
    size_t i;

    for (i = 1; i < def->nnets; i++) {
        VIR_AUTOPTR(virJSONValue) ep_conf = NULL;
        VIR_AUTOPTR(virJSONValue) connect = NULL;
        VIR_AUTOFREE(char *) cmd = NULL;
        VIR_AUTOFREE(char *) json = NULL;
        VIR_AUTOFREE(char *) response = NULL;
        char macaddr[VIR_MAC_STRING_BUFLEN];

        virMacAddrFormat(&def->nets[i]->mac, macaddr);
        if (!(ep_conf = dockerNetworkDefToJSON(&def->nets[i]->guestIP, macaddr)))
            return -1;

        if (virJSONValueObjectCreate(&connect,
                                     "s:Container", id_str,
                                     "a:EndpointConfig", &ep_conf,
                                     NULL) < 0)
            return -1;

        if (!(json = virJSONValueToString(connect, false)))
            return -1;

        virAsprintf(&cmd, "networks/%s/connect",
                    def->nets[i]->data.network.name);

        if (!(response = dockerApiRequest(priv, cmd, true,
                                          DOCKER_API_HTTP_POST,
                                          json, 200)))
            return -1;
    }

    return 0;
}


static const char *
dockerCompareDomainDefs(virDomainDefPtr old, virDomainDefPtr new)
{
    size_t i;

    if (STRNEQ(old->name, new->name))
        return "name";

    if (STRNEQ_NULLABLE(old->os.init, new->os.init))
        return "init";

    if (STRNEQ_NULLABLE(old->os.inituser, new->os.inituser))
        return "inituser";

    if (STRNEQ_NULLABLE(old->os.initdir, new->os.initdir))
        return "initdir";

    if ((old->os.initargv && !new->os.initargv) ||
        (!old->os.initargv && new->os.initargv && new->os.initargv[0]))
        return "initargv";

    for (i = 0; old->os.initargv &&
                (old->os.initargv[i] || new->os.initargv[i]); i++)
        if (STRNEQ_NULLABLE(old->os.initargv[i], new->os.initargv[i]))
            return "initargv";

    if ((old->os.initenv && !new->os.initenv) ||
        (!old->os.initenv && new->os.initenv && new->os.initenv[0]))
        return "initenv";

    for (i = 0; old->os.initenv &&
                (old->os.initenv[i] || new->os.initenv[i]); i++)
        if (!old->os.initenv[i] || !new->os.initenv[i] ||
            STRNEQ(old->os.initenv[i]->name, new->os.initenv[i]->name) ||
            STRNEQ(old->os.initenv[i]->value, new->os.initenv[i]->value))
            return "initenv";

    if (STRNEQ_NULLABLE(old->dns.hostname, new->dns.hostname))
        return "hostname";

    if (old->dns.nservers != new->dns.nservers)
        return "server";

    for (i = 0; i < old->dns.nservers; i++)
        if (!virSocketAddrEqual(&old->dns.servers[i], &new->dns.servers[i]))
            return "server";

    if ((old->dns.searchs && !new->dns.searchs) ||
        (!old->dns.searchs && new->dns.searchs && new->dns.searchs[0]))
        return "search";

    for (i = 0; old->dns.searchs &&
                (old->dns.searchs[i] || new->dns.searchs[i]); i++)
        if (STRNEQ_NULLABLE(old->dns.searchs[i], new->dns.searchs[i]))
            return "search";

    if (virDomainDefGetVcpusMax(old) != virDomainDefGetVcpusMax(new))
        return "vcpu";

    if (STRNEQ(virDomainDiskGetSource(old->disks[0]),
               virDomainDiskGetSource(new->disks[0])))
        return "disk";

    if (old->nfss != new->nfss)
        return "filesystem";

    for (i = 0; i < old->nfss; i++) {
        virDomainFSDefPtr ofs = old->fss[i];
        virDomainFSDefPtr nfs;
        int pos;

        /*
         * We use search by target for handling possible
         * rearrangement of filesystem elements because
         * Docker can permute items in 'Mounts' JSON array.
         */
        if ((pos = virDomainFSIndexByName(new, ofs->dst)) < 0)
            return "filesystem/target";

        nfs = new->fss[pos];

        if (ofs->type != nfs->type)
            return "filesystem/type";

        switch (ofs->type) {
        case VIR_DOMAIN_FS_TYPE_VOLUME:
            if (STRNEQ(ofs->src->srcpool->volume,
                       nfs->src->srcpool->volume))
                return "filesystem/source";
            break;
        case VIR_DOMAIN_FS_TYPE_MOUNT:
             if (STRNEQ(ofs->src->path, nfs->src->path))
                return "filesystem/source";
            break;
        }

        if (ofs->readonly != nfs->readonly)
            return "filesystem/readonly";
    }

    if (old->nnets != new->nnets)
        return "interface";

    for (i = 0; i < old->nnets; i++) {
        virNetDevIPInfoPtr ogip = &old->nets[i]->guestIP;
        virNetDevIPInfoPtr ngip = &new->nets[i]->guestIP;
        size_t j;

        if (STRNEQ(old->nets[i]->data.network.name,
                   new->nets[i]->data.network.name))
            return "interface/source";

        if (ogip->nips != ngip->nips)
            return "interface/ip";

        for (j = 0; j < ogip->nips; j++)
            if (!virSocketAddrEqual(&ogip->ips[j]->address,
                                    &ngip->ips[j]->address) ||
                ogip->ips[j]->prefix != ngip->ips[j]->prefix)
                return "interface/ip";

        if (ogip->nroutes != ngip->nroutes)
            return "interface/route";

        for (j = 0; j < ogip->nroutes; j++)
            if (!virSocketAddrEqual(&ogip->routes[j]->address,
                                    &ngip->routes[j]->address) ||
                !virSocketAddrEqual(&ogip->routes[j]->gateway,
                                    &ngip->routes[j]->gateway))
                return "interface/route";

        if (virMacAddrCmp(&old->nets[i]->mac, &new->nets[i]->mac))
            return "interface/mac";
    }

    return NULL;
}


static virDomainPtr
dockerDomainRedefine(virConnectPtr conn,
                     const char *id_str,
                     virDomainDefPtr def)
{
    dockerPrivatePtr priv = conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    VIR_AUTOPTR(virJSONValue) update = NULL;
    VIR_AUTOPTR(virDomainDef) old = NULL;
    long long cpushares, nanocpus, bandwidth;
    unsigned long long memory, tbs, tios;
    const char *element;

    if (!(obj = dockerGetContainerById(priv, id_str)))
        return NULL;

    if (!(old = dockerJSONToDomainDef(priv, obj)))
        return NULL;

    if ((element = dockerCompareDomainDefs(old, def))) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Cannot modify <%s> in domain configuration"),
                       element);
        return NULL;
    }

    /* Use negative value to remove limit */
    cpushares = def->cputune.sharesSpecified && def->cputune.shares ?
                def->cputune.shares : old->cputune.sharesSpecified ? -1 : 0;
    nanocpus = dockerCalcNanoCpus(priv, &def->cputune.period,
                                  def->cputune.quota, old->cputune.quota > 0);
    memory = virDomainDefGetMemoryTotal(def) * 1024; /* kilobyte to byte */

    bandwidth = dockerNetBandwidth(def);
    if (!bandwidth && dockerNetBandwidth(old))
        bandwidth = -1;

    if (virJSONValueObjectCreate(&update,
                                 "Z:CpuShares", cpushares,
                                 "P:CpuPeriod", def->cputune.period,
                                 "Z:CpuQuota", def->cputune.quota,
                                 "Z:NanoCpus", nanocpus,
                                 "U:Memory", memory,
                                 "Z:NetMaximumBandwidth", bandwidth,
                                 NULL) < 0)
        return NULL;

    /* Use value 18446744073709551615(-1llu) to remove limit */
    tbs = def->disks[0]->blkdeviotune.total_bytes_sec ?
          def->disks[0]->blkdeviotune.total_bytes_sec :
          (old->disks[0]->blkdeviotune.total_bytes_sec ? -1llu : 0);
    tios = def->disks[0]->blkdeviotune.total_iops_sec ?
           def->disks[0]->blkdeviotune.total_iops_sec :
           (old->disks[0]->blkdeviotune.total_iops_sec ? -1llu : 0);
    if (virJSONValueObjectAppendNumberUlong(update, "IOMaximumBandwidth",
                                            tbs) < 0 ||
        virJSONValueObjectAppendNumberUlong(update, "IOMaximumIOps",
                                            tios) < 0)
        return NULL;

    if (dockerContainerUpdateById(priv, id_str, update) < 0)
        return NULL;

    return virGetDomain(conn, def->name, def->uuid, old->id);
}


static int
dockerDomainDefValidate(const virDomainDef *def,
                        virCapsPtr caps ATTRIBUTE_UNUSED,
                        void *opaque ATTRIBUTE_UNUSED)
{
    if (STREQ(def->os.init, "*") &&
        virStringListLength((const char * const *)def->os.initargv)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("<init>*</init> not compatible with <initargv>"));
        return -1;
    }

    if (!def->ndisks) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Configuration of container must have one disk"));
        return -1;
    }

    if (def->ndisks > 1) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Supported only one disk"));
        return -1;
    }

    if (def->nnets > 1) {
        unsigned long long bandwidth = 0;
        size_t i;

        for (i = 0; i < def->nnets; i++) {
            if (def->nets[i]->bandwidth &&
                def->nets[i]->bandwidth->out->average) {
                if (bandwidth) {
                    if (bandwidth != def->nets[i]->bandwidth->out->average) {
                        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                                       _("Supported only same "
                                         "network bandwidth"));
                        return -1;
                    }
                } else {
                    bandwidth = def->nets[i]->bandwidth->out->average;
                }
            }
        }
    }

    return 0;
}


static int
dockerDeviceDefValidate(const virDomainDeviceDef *dev,
                        const virDomainDef *def ATTRIBUTE_UNUSED,
                        void *opaque ATTRIBUTE_UNUSED)
{
    switch (dev->type) {
    case VIR_DOMAIN_DEVICE_DISK:
        if (dev->data.disk->device != VIR_DOMAIN_DISK_DEVICE_DISK) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                           _("Unsupported type of device '%s'"),
                           virDomainDiskDeviceTypeToString(dev->data.disk->device));
            return -1;
        }
        if (virDomainDiskGetDriver(dev->data.disk) &&
            STRNEQ(virDomainDiskGetDriver(dev->data.disk), "docker")) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                           _("Unsupported type of driver name '%s'"),
                           virDomainDiskGetDriver(dev->data.disk));
            return -1;
        }
        if (virDomainDiskGetType(dev->data.disk) != VIR_STORAGE_TYPE_TEMPLATE) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                           _("Unsupported type of source '%s'"),
                           virStorageTypeToString(virDomainDiskGetType(dev->data.disk)));
            return -1;
        }
        if (dev->data.disk->bus != VIR_DOMAIN_DISK_BUS_VIRTIO) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                           _("Unsupported bus of disk '%s'"),
                           virDomainDiskBusTypeToString(dev->data.disk->bus));
            return -1;
        }
        if (dev->data.disk->blkdeviotune.read_bytes_sec ||
            dev->data.disk->blkdeviotune.write_bytes_sec ||
            dev->data.disk->blkdeviotune.read_iops_sec ||
            dev->data.disk->blkdeviotune.write_iops_sec ||
            dev->data.disk->blkdeviotune.total_bytes_sec_max ||
            dev->data.disk->blkdeviotune.read_bytes_sec_max ||
            dev->data.disk->blkdeviotune.write_bytes_sec_max ||
            dev->data.disk->blkdeviotune.total_iops_sec_max ||
            dev->data.disk->blkdeviotune.read_iops_sec_max ||
            dev->data.disk->blkdeviotune.write_iops_sec_max ||
            dev->data.disk->blkdeviotune.total_bytes_sec_max_length ||
            dev->data.disk->blkdeviotune.read_bytes_sec_max_length ||
            dev->data.disk->blkdeviotune.write_bytes_sec_max_length ||
            dev->data.disk->blkdeviotune.total_iops_sec_max_length ||
            dev->data.disk->blkdeviotune.read_iops_sec_max_length ||
            dev->data.disk->blkdeviotune.write_iops_sec_max_length ||
            dev->data.disk->blkdeviotune.size_iops_sec ||
            dev->data.disk->blkdeviotune.group_name) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Supported only <total_bytes_sec> and <total_iops_sec>"));
            return -1;
        }
        break;

    case VIR_DOMAIN_DEVICE_NET:
        if (dev->data.net->type != VIR_DOMAIN_NET_TYPE_NETWORK) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                           _("Unsupported type of network '%s'"),
                           virDomainNetTypeToString(dev->data.net->type));
            return -1;
        }
        if (dev->data.net->bandwidth) {
            if (dev->data.net->bandwidth->in) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("Supported only outbound network bandwidth"));
                return -1;
            }
            if (dev->data.net->bandwidth->out->peak ||
                dev->data.net->bandwidth->out->floor ||
                dev->data.net->bandwidth->out->burst) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("Supported only average network bandwidth"));
                return -1;
            }
        }
        break;

    case VIR_DOMAIN_DEVICE_FS:
        if (dev->data.fs->type != VIR_DOMAIN_FS_TYPE_MOUNT &&
            dev->data.fs->type != VIR_DOMAIN_FS_TYPE_VOLUME) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                           _("Unsupported type of filesystem '%s'"),
                           virDomainFSTypeToString(dev->data.fs->type));
            return -1;
        }
        break;

    default:
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Unsupported type of devices '%s'"),
                       virDomainDeviceTypeToString(dev->type));
        return -1;
    }

    return 0;
}


static virDomainDefParserConfig dockerParserConfig = {
    .domainValidateCallback = dockerDomainDefValidate,
    .deviceValidateCallback = dockerDeviceDefValidate,
};


/*
 * Callback functions for the Docker driver
 */
static virDrvOpenStatus
dockerConnectOpen(virConnectPtr conn,
                  virConnectAuthPtr auth ATTRIBUTE_UNUSED,
                  virConfPtr conf ATTRIBUTE_UNUSED,
                  unsigned int flags)
{
    virDrvOpenStatus ret = VIR_DRV_OPEN_ERROR;
    dockerPrivatePtr priv = NULL;

    virCheckFlags(VIR_CONNECT_RO, VIR_DRV_OPEN_ERROR);

    VIR_ALLOC(priv);

    if (dockerParseUri(priv, conn->uri) < 0)
        goto cleanup;

    virAsprintf(&priv->hostUri, "%s://%s:%d", priv->transport,
                conn->uri->server, conn->uri->port);

    if (dockerCheckApiVersion(priv) < 0)
        goto cleanup;

    if (dockerCapsInit(priv) < 0)
        goto cleanup;

    if (!(priv->xmlOpt = virDomainXMLOptionNew(&dockerParserConfig,
                                               NULL, NULL, NULL, NULL)))
        goto cleanup;

    VIR_STEAL_PTR(conn->privateData, priv);
    ret = VIR_DRV_OPEN_SUCCESS;

 cleanup:
    dockerFreePrivate(&priv);

    return ret;
}


static int
dockerConnectClose(virConnectPtr conn)
{
    dockerPrivatePtr priv = conn->privateData;

    dockerFreePrivate(&priv);
    conn->privateData = NULL;

    return 0;
}


static const char *
dockerConnectGetType(virConnectPtr conn ATTRIBUTE_UNUSED)
{
    return "Docker";
}


static int
dockerConnectGetVersion(virConnectPtr conn, unsigned long *version)
{
    dockerPrivatePtr priv = conn->privateData;
    const char *ver;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    if (!(obj = dockerGetJSONObject(priv, "version", true,
                                    DOCKER_API_HTTP_GET,
                                    NULL, 200)))
        return -1;

    if (!(ver = virJSONValueObjectGetString(obj, "Version")) ||
        virParseVersionString(ver, version, true) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect format 'Version' field in JSON document"));
        return -1;
    }

    return 0;
}


static char *
dockerConnectGetHostname(virConnectPtr conn)
{
    dockerPrivatePtr priv = conn->privateData;
    const char *name;
    char *hostname;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    if (!(obj = dockerGetJSONObject(priv, "info", true,
                                    DOCKER_API_HTTP_GET,
                                    NULL, 200)))
        return NULL;

    if (!(name = virJSONValueObjectGetString(obj, "Name"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect format 'Name' field in JSON document"));
        return NULL;
    }

    VIR_STRDUP(hostname, name);

    return hostname;
}


static int
dockerConnectGetMaxVcpus(virConnectPtr conn, const char *type ATTRIBUTE_UNUSED)
{
    dockerPrivatePtr priv = conn->privateData;
    unsigned int ncpu;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    if (!(obj = dockerGetJSONObject(priv, "info", true,
                                    DOCKER_API_HTTP_GET,
                                    NULL, 200)))
        return -1;

    if (virJSONValueObjectGetNumberUint(obj, "NCPU", &ncpu) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect format 'NCPU' field in JSON document"));
        return -1;
    }

    return ncpu;
}


static char *
dockerConnectGetCapabilities(virConnectPtr conn)
{
    dockerPrivatePtr priv = conn->privateData;

    return virCapabilitiesFormatXML(priv->hostCaps);
}


static int
dockerConnectIsEncrypted(virConnectPtr conn)
{
    dockerPrivatePtr priv = conn->privateData;

    if (STRCASEEQ(priv->transport, "https"))
        return 1;

    return 0;
}


static int
dockerConnectIsSecure(virConnectPtr conn)
{
    dockerPrivatePtr priv = conn->privateData;

    if (STRCASEEQ(priv->transport, "https"))
        return 1;

    return 0;
}


static int
dockerConnectIsAlive(virConnectPtr conn)
{
    dockerPrivatePtr priv = conn->privateData;
    VIR_AUTOFREE(char *) response = NULL;

    if (!(response = dockerApiRequest(priv, "_ping", false,
                                      DOCKER_API_HTTP_HEAD,
                                      NULL, 200)))
        return 0;

    return 1;
}


static int
dockerConnectListAllDomains(virConnectPtr conn,
                            virDomainPtr **domains,
                            unsigned int flags)
{
    dockerPrivatePtr priv = conn->privateData;
    int ret = -1;
    int ndoms = 0;
    size_t i, ndoms_max = 0;
    virDomainPtr *doms = NULL;
    VIR_AUTOPTR(virJSONValue) list = NULL;

    virCheckFlags(VIR_CONNECT_LIST_DOMAINS_FILTERS_ALL, -1);

    if (!(list = dockerGetContainersList(priv, flags)))
        goto cleanup;

    for (i = 0; i < virJSONValueArraySize(list); i++) {
        virJSONValuePtr item = virJSONValueArrayGet(list, i);

        if (flags & VIR_CONNECT_LIST_DOMAINS_FILTERS_STATE) {
            virDomainState state = dockerGetContainerState(item);
            if (!(((flags & VIR_CONNECT_LIST_DOMAINS_RUNNING) &&
                   state == VIR_DOMAIN_RUNNING) ||
                  ((flags & VIR_CONNECT_LIST_DOMAINS_PAUSED) &&
                   state == VIR_DOMAIN_PAUSED) ||
                  ((flags & VIR_CONNECT_LIST_DOMAINS_SHUTOFF) &&
                   state == VIR_DOMAIN_SHUTOFF) ||
                  ((flags & VIR_CONNECT_LIST_DOMAINS_OTHER) &&
                   (state != VIR_DOMAIN_RUNNING &&
                    state != VIR_DOMAIN_PAUSED &&
                    state != VIR_DOMAIN_SHUTOFF))))
                continue;
        }

        if (flags & VIR_CONNECT_LIST_DOMAINS_FILTERS_AUTOSTART) {
            VIR_AUTOPTR(virJSONValue) obj = NULL;
            const char *id_str;
            int autostart;

            if (!(id_str = virJSONValueObjectGetString(item, "Id"))) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Incorrect 'Id' field in JSON container"));
                goto cleanup;
            }

            if (!(obj = dockerGetContainerById(priv, id_str)) ||
                dockerGetContainerAutostart(obj, &autostart) < 0)
                goto cleanup;

            if (!(((flags & VIR_CONNECT_LIST_DOMAINS_AUTOSTART) &&
                   autostart) ||
                  ((flags & VIR_CONNECT_LIST_DOMAINS_NO_AUTOSTART) &&
                   !autostart)))
                continue;
        }

        /* Only if we have pointer for saving domains */
        if (domains) {
            VIR_RESIZE_N(doms, ndoms_max, ndoms, 1);
            if (!(doms[ndoms] = dockerGetDomainFromObject(conn, item)))
                goto cleanup;
        }

        ndoms++;
    }

    if (domains) {
        VIR_RESIZE_N(doms, ndoms_max, ndoms, 1);
        VIR_STEAL_PTR(*domains, doms);
    }
    ret = ndoms;

 cleanup:
    virObjectListFreeCount(doms, ndoms);

    return ret;
}


static int
dockerConnectGetAllDomainStats(virConnectPtr conn,
                               virDomainPtr *domains,
                               unsigned int ndomains,
                               unsigned int stats,
                               virDomainStatsRecordPtr **retStats,
                               unsigned int flags)
{
    dockerPrivatePtr priv = conn->privateData;
    int ret = -1;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    virJSONValuePtr list;
    int ndoms = 0;
    VIR_AUTOFREE(int *) items = NULL;
    size_t i, ndoms_max = 0;
    size_t items_max = 0;
    virDomainPtr *doms = NULL;
    virDomainStatsRecordPtr *tmpstats = NULL;
    dockerJobsControl jobsctl = { NULL, NULL, 0, 0, 0 };
    const unsigned int supported = VIR_DOMAIN_STATS_STATE |
        VIR_DOMAIN_STATS_CPU_TOTAL | VIR_DOMAIN_STATS_BALLOON |
        VIR_DOMAIN_STATS_VCPU | VIR_DOMAIN_STATS_INTERFACE |
        VIR_DOMAIN_STATS_BLOCK;

    virCheckFlags(VIR_CONNECT_GET_ALL_DOMAINS_STATS_ACTIVE |
                  VIR_CONNECT_GET_ALL_DOMAINS_STATS_INACTIVE |
                  VIR_CONNECT_GET_ALL_DOMAINS_STATS_OTHER |
                  VIR_CONNECT_GET_ALL_DOMAINS_STATS_PAUSED |
                  VIR_CONNECT_GET_ALL_DOMAINS_STATS_PERSISTENT |
                  VIR_CONNECT_GET_ALL_DOMAINS_STATS_RUNNING |
                  VIR_CONNECT_GET_ALL_DOMAINS_STATS_SHUTOFF |
                  VIR_CONNECT_GET_ALL_DOMAINS_STATS_NOWAIT |
                  VIR_CONNECT_GET_ALL_DOMAINS_STATS_ENFORCE_STATS, -1);

    if (!stats) {
        stats = supported;
    } else if ((flags & VIR_CONNECT_GET_ALL_DOMAINS_STATS_ENFORCE_STATS) &&
               (stats & ~supported)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED,
                       _("Stats types (0x%x) don't supported"),
                       stats & ~supported);
        return -1;
    }

    if (!(obj = dockerGetJSONObject(priv, "system/df", true,
                                    DOCKER_API_HTTP_GET,
                                    NULL, 200)))
        goto cleanup;

    if (!(list = virJSONValueObjectGet(obj, "Containers")) ||
        !virJSONValueIsArray(list)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Containers' field in JSON document"));
        goto cleanup;
    }

    for (i = 0; i < virJSONValueArraySize(list); i++) {
        virJSONValuePtr item = virJSONValueArrayGet(list, i);
        virDomainState state = dockerGetContainerState(item);
        virDomainPtr domain;

        if ((flags & VIR_CONNECT_LIST_DOMAINS_FILTERS_ACTIVE) &&
            !(((flags & VIR_CONNECT_GET_ALL_DOMAINS_STATS_ACTIVE) &&
               (state == VIR_DOMAIN_RUNNING ||
                state == VIR_DOMAIN_PAUSED)) ||
              ((flags & VIR_CONNECT_GET_ALL_DOMAINS_STATS_INACTIVE) &&
               (state != VIR_DOMAIN_RUNNING &&
                state != VIR_DOMAIN_PAUSED))))
            continue;

        if ((flags & VIR_CONNECT_LIST_DOMAINS_FILTERS_STATE) &&
            !(((flags & VIR_CONNECT_GET_ALL_DOMAINS_STATS_RUNNING) &&
               state == VIR_DOMAIN_RUNNING) ||
              ((flags & VIR_CONNECT_GET_ALL_DOMAINS_STATS_PAUSED) &&
                state == VIR_DOMAIN_PAUSED) ||
              ((flags & VIR_CONNECT_GET_ALL_DOMAINS_STATS_SHUTOFF) &&
                state == VIR_DOMAIN_SHUTOFF) ||
              ((flags & VIR_CONNECT_GET_ALL_DOMAINS_STATS_OTHER) &&
                (state != VIR_DOMAIN_RUNNING &&
                 state != VIR_DOMAIN_PAUSED &&
                 state != VIR_DOMAIN_SHUTOFF))))
            continue;

        if (!(domain = dockerGetDomainFromObject(conn, item)))
            goto cleanup;

        if (domains) {
            size_t j;

            for (j = 0; j < ndomains; j++) {
                if (!memcmp(domain->uuid, domains[j]->uuid, VIR_UUID_BUFLEN))
                    break;
            }

            if (j >= ndomains) {
                virObjectUnref(domain);
                continue;
            }
        }

        VIR_RESIZE_N(items, items_max, ndoms, 1);
        VIR_RESIZE_N(doms, ndoms_max, ndoms, 1);
        items[ndoms] = i;
        doms[ndoms++] = domain;
    }

    VIR_ALLOC_N(tmpstats, ndoms + 1);

    /* Return empty list of domains */
    if (!ndoms) {
        VIR_STEAL_PTR(*retStats, tmpstats);
        ret = 0;
        goto cleanup;
    }

    jobsctl.stats = stats;
    for (i = 0; i < ndoms; i++) {
        virJSONValuePtr item = virJSONValueArrayGet(list, items[i]);
        virDomainState state = dockerGetContainerState(item);

        VIR_ALLOC(tmpstats[i]);
        VIR_STEAL_PTR(tmpstats[i]->dom, doms[i]);

        if ((state == VIR_DOMAIN_RUNNING || state == VIR_DOMAIN_PAUSED) &&
            (stats & (VIR_DOMAIN_STATS_CPU_TOTAL | VIR_DOMAIN_STATS_BALLOON |
                      VIR_DOMAIN_STATS_VCPU | VIR_DOMAIN_STATS_INTERFACE |
                      VIR_DOMAIN_STATS_BLOCK))) {
            /* Only for active container we do async API request */
            if (dockerStartStatsJob(priv, item, tmpstats[i], &jobsctl) < 0)
                goto cleanup;
        } else {
            int maxparams = 0;

            if ((stats & VIR_DOMAIN_STATS_STATE) &&
                dockerFillDomainState(item, tmpstats[i], &maxparams) < 0)
                goto cleanup;

            if ((stats & VIR_DOMAIN_STATS_INTERFACE) &&
                dockerFillNetStats(item, NULL, tmpstats[i], &maxparams) < 0)
                goto cleanup;

            if ((stats & VIR_DOMAIN_STATS_BLOCK) &&
                dockerFillBlockStats(item, NULL, tmpstats[i], &maxparams) < 0)
                goto cleanup;
        }
    }

    if (dockerWaitAllJobs(&jobsctl, dockerStatsJobHandler) < 0)
        goto cleanup;

    VIR_STEAL_PTR(*retStats, tmpstats);
    ret = ndoms;

 cleanup:
    dockerCloseJobs(&jobsctl);
    virDomainStatsRecordListFree(tmpstats);
    virObjectListFreeCount(doms, ndoms);

    return ret;
}


static int
dockerConnectSupportsFeature(virConnectPtr conn ATTRIBUTE_UNUSED, int feature)
{
    switch ((virDrvFeature)feature) {
    case VIR_DRV_FEATURE_MIGRATION_PARAMS:
    case VIR_DRV_FEATURE_MIGRATION_OFFLINE:
    case VIR_DRV_FEATURE_TYPED_PARAM_STRING:
    case VIR_DRV_FEATURE_XML_MIGRATABLE:
        return 1;
    case VIR_DRV_FEATURE_MIGRATION_V1:
    case VIR_DRV_FEATURE_MIGRATION_V2:
    case VIR_DRV_FEATURE_MIGRATION_V3:
    case VIR_DRV_FEATURE_MIGRATION_P2P:
    case VIR_DRV_FEATURE_MIGRATION_DIRECT:
    case VIR_DRV_FEATURE_MIGRATE_CHANGE_PROTECTION:
    case VIR_DRV_FEATURE_FD_PASSING:
    case VIR_DRV_FEATURE_PROGRAM_KEEPALIVE:
    case VIR_DRV_FEATURE_REMOTE:
    case VIR_DRV_FEATURE_REMOTE_EVENT_CALLBACK:
    case VIR_DRV_FEATURE_REMOTE_CLOSE_CALLBACK:
    case VIR_DRV_FEATURE_MIGRATION_VSTORAGE:
    default:
        return 0;
    }
}


static virDomainPtr
dockerDomainLookupByID(virConnectPtr conn, int id)
{
    VIR_AUTOFREE(char *) id_str = NULL;

    virAsprintf(&id_str, "%04x", id);
    return dockerLookupContainer(conn, "id", id_str);
}


static virDomainPtr
dockerDomainLookupByUUID(virConnectPtr conn, const unsigned char *uuid)
{
    dockerPrivatePtr priv = conn->privateData;
    VIR_AUTOFREE(char *) id_str = NULL;

    if (dockerSeekLabel(priv, uuid, false, &id_str) < 0)
        return NULL;

    return dockerLookupContainer(conn, "id", id_str);
}


static virDomainPtr
dockerDomainLookupByName(virConnectPtr conn, const char *name)
{
    return dockerLookupContainer(conn, "name", name);
}


static int
dockerDomainIsActive(virDomainPtr domain)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    virDomainState state;

    if (!(obj = dockerGetContainerByUUID(priv, domain->uuid)))
        return -1;

    state = dockerGetContainerState(obj);

    return ((state == VIR_DOMAIN_RUNNING) ||
            (state == VIR_DOMAIN_PAUSED)) ? 1 : 0;
}


static int
dockerDomainIsPersistent(virDomainPtr domain)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    if (!(obj = dockerGetContainerByUUID(priv, domain->uuid)))
        return -1;

    /* For Docker all containers are persistent */
    return 1;
}


static int
dockerDomainIsUpdated(virDomainPtr domain)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    if (!(obj = dockerGetContainerByUUID(priv, domain->uuid)))
        return -1;

    /* Unfortunately we cannot look after Docker containers */
    return 0;
}


static int
dockerDomainHasManagedSaveImage(virDomainPtr domain, unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    virCheckFlags(0, -1);

    if (!(obj = dockerGetContainerByUUID(priv, domain->uuid)))
        return -1;

    /* Docker doesn't support managed save image */
    return 0;
}


static int
dockerDomainGetAutostart(virDomainPtr domain, int *autostart)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    if (!(obj = dockerGetContainerByUUID(priv, domain->uuid)))
        return -1;

    if (dockerGetContainerAutostart(obj, autostart) < 0)
        return -1;

    return 0;
}


static int
dockerDomainSetAutostart(virDomainPtr domain, int autostart)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    VIR_AUTOPTR(virJSONValue) restart_pol = NULL;

    if (virJSONValueObjectCreate(&restart_pol,
                                 "s:Name", autostart ? "always" : "no",
                                 "u:MaximumRetryCount", 0,
                                 NULL) < 0)
        return -1;

    if (virJSONValueObjectCreate(&obj,
                                 "a:RestartPolicy", &restart_pol,
                                 NULL) < 0)
        return -1;

    if (dockerContainerUpdate(priv, domain->uuid, obj) < 0)
        return -1;

    return 0;
}


static int
dockerDomainGetInfo(virDomainPtr domain, virDomainInfoPtr info)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    VIR_AUTOPTR(virJSONValue) sinfo = NULL;
    VIR_AUTOPTR(virJSONValue) stats = NULL;
    virJSONValuePtr host_conf;
    unsigned int nvcpu;
    unsigned long long maxmem;

    if (!(obj = dockerGetContainerByUUID(priv, domain->uuid)))
        return -1;

    if (!(host_conf = virJSONValueObjectGet(obj, "HostConfig")) ||
        virJSONValueObjectGetNumberUint(host_conf, "CpuCount", &nvcpu) < 0 ||
        virJSONValueObjectGetNumberUlong(host_conf, "Memory", &maxmem) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'HostConfig' object in JSON container"));
        return -1;
    }

    if (!nvcpu || !maxmem) {
        if (!(sinfo = dockerGetJSONObject(priv, "info", true,
                                         DOCKER_API_HTTP_GET,
                                         NULL, 200)))
            return -1;

        if (!nvcpu &&
            virJSONValueObjectGetNumberUint(sinfo, "NCPU", &nvcpu) < 0) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'NCPU' field in JSON document"));
            return -1;
        }

        if (!maxmem &&
            virJSONValueObjectGetNumberUlong(sinfo, "MemTotal", &maxmem) < 0) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'MemTotal' field in JSON document"));
            return -1;
        }
    }

    memset(info, 0, sizeof(*info));
    info->state = dockerGetContainerState(obj);
    info->maxMem = maxmem / 1024; /* byte to kilobyte */

    if ((info->state == VIR_DOMAIN_RUNNING) || (info->state == VIR_DOMAIN_PAUSED)) {
        virJSONValuePtr mem_stats, cpu_stats, cpu_usage;
        unsigned long long memory, usage;

        if (!(stats = dockerGetContainerStats(priv, domain->uuid)))
            return -1;

        if (virJSONValueObjectGetNumberUint(stats, "num_procs", &nvcpu) < 0) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'num_procs' field in JSON document"));
            return -1;
        }

        if (!(mem_stats = virJSONValueObjectGet(stats, "memory_stats")) ||
            virJSONValueObjectGetNumberUlong(mem_stats, "privateworkingset", &memory) < 0) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'memory_stats' object in JSON document"));
            return -1;
        }

        if (!(cpu_stats = virJSONValueObjectGet(stats, "cpu_stats")) ||
            !(cpu_usage = virJSONValueObjectGet(cpu_stats, "cpu_usage")) ||
            virJSONValueObjectGetNumberUlong(cpu_usage, "total_usage", &usage) < 0) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'cpu_stats' object in JSON document"));
            return -1;
        }

        info->memory = memory / 1024; /* byte to kilobyte */
        info->cpuTime = usage * 100; /* usage in 100ns to nanoseconds */
    }

    info->nrVirtCpu = nvcpu;

    return 0;
}


static int
dockerDomainGetState(virDomainPtr domain,
                     int *state,
                     int *reason,
                     unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    virCheckFlags(0, -1);

    if (!(obj = dockerGetContainerByUUID(priv, domain->uuid)))
        return -1;

    *state = dockerGetContainerState(obj);

    if (reason)
        *reason = 0;

    return 0;
}


static char *
dockerDomainGetOSType(virDomainPtr domain)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    char *ostype;

    if (!(obj = dockerGetContainerByUUID(priv, domain->uuid)))
        return NULL;

    VIR_STRDUP(ostype, virDomainOSTypeToString(VIR_DOMAIN_OSTYPE_EXE));

    return ostype;
}


static int
dockerCheckDomainAffectFlags(virDomainPtr domain, unsigned int flags)
{
    if (dockerDomainIsActive(domain)) {
        if (!(flags & VIR_DOMAIN_AFFECT_LIVE) ||
            !(flags & VIR_DOMAIN_AFFECT_CONFIG)) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Cannot modify independently persistent "
                             "and active configs"));
            return -1;
        }
    } else {
        if (flags & VIR_DOMAIN_AFFECT_LIVE) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Cannot modify active config of inactive domain"));
            return -1;
        }
    }

    return 0;
}


static int
dockerDomainSetMemoryFlags(virDomainPtr domain,
                           unsigned long memory,
                           unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    long long bytes;

    virCheckFlags(VIR_DOMAIN_MEM_LIVE |
                  VIR_DOMAIN_MEM_CONFIG |
                  VIR_DOMAIN_MEM_MAXIMUM, -1);

    if ((flags & (VIR_DOMAIN_MEM_LIVE | VIR_DOMAIN_MEM_CONFIG)) &&
        dockerCheckDomainAffectFlags(domain, flags) < 0)
        return -1;

    /* Use negative value to remove limit */
    bytes = memory ? memory * 1024 : -1; /* kilobyte to byte */

    if (virJSONValueObjectCreate(&obj, "I:Memory", bytes, NULL) < 0)
        return -1;

    if (dockerContainerUpdate(priv, domain->uuid, obj) < 0)
        return -1;

    return 0;
}


static int
dockerDomainSetMemory(virDomainPtr domain, unsigned long memory)
{
    return dockerDomainSetMemoryFlags(domain, memory, 0);
}


static char *
dockerDomainGetSchedulerType(virDomainPtr domain, int *nparams)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    char *type;

    if (!(obj = dockerGetContainerByUUID(priv, domain->uuid)))
        return NULL;

    VIR_STRDUP(type, "posix");

    if (nparams)
        *nparams = 3; /* cpu_shares, vpcu_period, vcpu_quota */

    return type;
}


static int
dockerDomainGetSchedulerParametersFlags(virDomainPtr domain,
                                        virTypedParameterPtr params,
                                        int *nparams,
                                        unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    virJSONValuePtr host_conf;
    unsigned long long cpushares, nanocpus, period;
    long long quota;
    int param = 0;

    /* We don't return strings so we can support string flag */
    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG |
                  VIR_TYPED_PARAM_STRING_OKAY, -1);

    if (!(obj = dockerGetContainerByUUID(priv, domain->uuid)))
        return -1;

    if (flags & VIR_DOMAIN_AFFECT_LIVE) {
        virDomainState state = dockerGetContainerState(obj);
        if (state != VIR_DOMAIN_RUNNING && state != VIR_DOMAIN_PAUSED) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Flag 'live' used for not active domain"));
            return -1;
        }
    }

    if (!(host_conf = virJSONValueObjectGet(obj, "HostConfig")) ||
        virJSONValueObjectGetNumberUlong(host_conf, "CpuShares",
                                         &cpushares) < 0 ||
        virJSONValueObjectGetNumberUlong(host_conf, "CpuPeriod",
                                         &period) < 0 ||
        virJSONValueObjectGetNumberLong(host_conf, "CpuQuota",
                                        &quota) < 0 ||
        virJSONValueObjectGetNumberUlong(host_conf, "NanoCpus",
                                         &nanocpus) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'HostConfig' object in JSON container"));
        return -1;
    }

    if (nanocpus)
        period = dockerCalcPeriod(nanocpus, &quota);

    if (virTypedParameterAssign(&params[param++],
                                VIR_DOMAIN_SCHEDULER_CPU_SHARES,
                                VIR_TYPED_PARAM_ULLONG,
                                cpushares) < 0)
        return -1;

    if (param < *nparams &&
        virTypedParameterAssign(&params[param++],
                                VIR_DOMAIN_SCHEDULER_VCPU_PERIOD,
                                VIR_TYPED_PARAM_ULLONG,
                                period) < 0)
        return -1;

    if (param < *nparams &&
        virTypedParameterAssign(&params[param++],
                                VIR_DOMAIN_SCHEDULER_VCPU_QUOTA,
                                VIR_TYPED_PARAM_LLONG,
                                quota) < 0)
        return -1;

    *nparams = param;

    return 0;
}


static int
dockerDomainGetSchedulerParameters(virDomainPtr domain,
                                   virTypedParameterPtr params,
                                   int *nparams)
{
    return dockerDomainGetSchedulerParametersFlags(domain, params, nparams, 0);
}


static int
dockerDomainSetSchedulerParametersFlags(virDomainPtr domain,
                                        virTypedParameterPtr params,
                                        int nparams,
                                        unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    unsigned long long period = 0;
    long long cpushares = 0;
    long long quota = 0;
    long long nanocpus;
    size_t i;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (virTypedParamsValidate(params, nparams,
                               VIR_DOMAIN_SCHEDULER_CPU_SHARES,
                               VIR_TYPED_PARAM_ULLONG,
                               VIR_DOMAIN_SCHEDULER_VCPU_PERIOD,
                               VIR_TYPED_PARAM_ULLONG,
                               VIR_DOMAIN_SCHEDULER_VCPU_QUOTA,
                               VIR_TYPED_PARAM_LLONG,
                               NULL) < 0)
        return -1;

    if (flags && dockerCheckDomainAffectFlags(domain, flags) < 0)
        return -1;

    for (i = 0; i < nparams; i++) {
        if (STREQ(params[i].field, VIR_DOMAIN_SCHEDULER_CPU_SHARES)) {
            /* Use negative value to remove limit */
            cpushares = params[i].value.ul ? params[i].value.ul : -1;
        } else if (STREQ(params[i].field, VIR_DOMAIN_SCHEDULER_VCPU_PERIOD)) {
            period = params[i].value.ul;
            if (period && (period < 1000 || period > 1000000)) {
                virReportError(VIR_ERR_INVALID_ARG, "%s",
                               _("Value of 'period' is out of range "
                                 "[1000, 1000000]"));
                return -1;
            }
        } else if (STREQ(params[i].field, VIR_DOMAIN_SCHEDULER_VCPU_QUOTA)) {
            quota = params[i].value.l;
            if (quota > 0 && (quota < 1000 || quota > 18446744073709551)) {
                virReportError(VIR_ERR_INVALID_ARG, "%s",
                               _("Value of 'quota' is out of range "
                                 "[1000, 18446744073709551]"));
                return -1;
            }
        }
    }

    nanocpus = dockerCalcNanoCpus(priv, &period, quota, true);

    if (virJSONValueObjectCreate(&obj,
                                 "I:CpuShares", cpushares,
                                 "P:CpuPeriod", period,
                                 "I:CpuQuota", quota,
                                 "I:NanoCpus", nanocpus,
                                 NULL) < 0)
        return -1;

    if (dockerContainerUpdate(priv, domain->uuid, obj) < 0)
        return -1;

    return 0;
}


static int
dockerDomainSetSchedulerParameters(virDomainPtr domain,
                                   virTypedParameterPtr params,
                                   int nparams)
{
    return dockerDomainSetSchedulerParametersFlags(domain, params, nparams, 0);
}


static int
dockerDomainGetBlockIoTune(virDomainPtr domain,
                           const char *path,
                           virTypedParameterPtr params,
                           int *nparams,
                           unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    virJSONValuePtr host_conf;
    unsigned long long bytes, iops;
    int param = 0;

    /* We don't return strings so we can support string flag */
    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG |
                  VIR_TYPED_PARAM_STRING_OKAY, -1);

    if (!*nparams) {
        *nparams = 2; /* total_bytes_sec and total_iops_sec */
        return 0;
    }

    if (STRNEQ(path, "vda")) {
        virReportError(VIR_ERR_DEVICE_MISSING, "%s", path);
        return -1;
    }

    if (!(obj = dockerGetContainerByUUID(priv, domain->uuid)))
        return -1;

    if (flags & VIR_DOMAIN_AFFECT_LIVE) {
        virDomainState state = dockerGetContainerState(obj);
        if (state != VIR_DOMAIN_RUNNING && state != VIR_DOMAIN_PAUSED) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Flag 'live' used for not active domain"));
            return -1;
        }
    }

    if (!(host_conf = virJSONValueObjectGet(obj, "HostConfig")) ||
        virJSONValueObjectGetNumberUlong(host_conf, "IOMaximumBandwidth",
                                         &bytes) < 0 ||
        virJSONValueObjectGetNumberUlong(host_conf, "IOMaximumIOps",
                                         &iops) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'HostConfig' object in JSON container"));
        return -1;
    }

    if (virTypedParameterAssign(&params[param++],
                                VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_BYTES_SEC,
                                VIR_TYPED_PARAM_ULLONG,
                                bytes) < 0)
        return -1;

    if (param < *nparams &&
        virTypedParameterAssign(&params[param++],
                                VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_IOPS_SEC,
                                VIR_TYPED_PARAM_ULLONG,
                                iops) < 0)
        return -1;

    *nparams = param;

    return 0;
}


static int
dockerDomainSetBlockIoTune(virDomainPtr domain,
                           const char *path,
                           virTypedParameterPtr params,
                           int nparams,
                           unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    size_t i;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (STRNEQ(path, "vda")) {
        virReportError(VIR_ERR_DEVICE_MISSING, "%s", path);
        return -1;
    }

    if (virTypedParamsValidate(params, nparams,
                               VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_BYTES_SEC,
                               VIR_TYPED_PARAM_ULLONG,
                               VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_IOPS_SEC,
                               VIR_TYPED_PARAM_ULLONG,
                               NULL) < 0)
        return -1;

    if (flags && dockerCheckDomainAffectFlags(domain, flags) < 0)
        return -1;

    obj = virJSONValueNewObject();

    for (i = 0; i < nparams; i++) {
        /* Use value 18446744073709551615(-1llu) to remove limit */
        if (STREQ(params[i].field, VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_BYTES_SEC)) {
            if (virJSONValueObjectAppendNumberUlong(obj, "IOMaximumBandwidth",
                                                    params[i].value.ul ?
                                                    params[i].value.ul :
                                                    -1llu) < 0)
                return -1;
        } else if (STREQ(params[i].field,
                   VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_IOPS_SEC)) {
            if (virJSONValueObjectAppendNumberUlong(obj, "IOMaximumIOps",
                                                    params[i].value.ul ?
                                                    params[i].value.ul :
                                                    -1llu) < 0)
                return -1;
        }
    }

    if (dockerContainerUpdate(priv, domain->uuid, obj) < 0)
        return -1;

    return 0;
}


static char *
dockerDomainGetXMLDesc(virDomainPtr domain, unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    VIR_AUTOPTR(virDomainDef) def = NULL;

    virCheckFlags(VIR_DOMAIN_XML_COMMON_FLAGS, NULL);

    if (!(obj = dockerGetContainerByUUID(priv, domain->uuid)))
        return NULL;

    if (!(def = dockerJSONToDomainDef(priv, obj)))
        return NULL;

    return virDomainDefFormat(def, NULL,
                              virDomainDefFormatConvertXMLFlags(flags));
}


static virDomainPtr
dockerDomainDefineXMLFlags(virConnectPtr conn,
                           const char *xml,
                           unsigned int flags)
{
    dockerPrivatePtr priv = conn->privateData;
    VIR_AUTOPTR(virDomainDef) def = NULL;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    VIR_AUTOPTR(virJSONValue) id = NULL;
    unsigned int parse_flags = VIR_DOMAIN_DEF_PARSE_INACTIVE;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) json = NULL;
    VIR_AUTOFREE(char *) id_str = NULL;
    int seek;

    virCheckFlags(VIR_DOMAIN_DEFINE_VALIDATE, NULL);

    if (flags & VIR_DOMAIN_DEFINE_VALIDATE)
        parse_flags |= VIR_DOMAIN_DEF_PARSE_VALIDATE_SCHEMA;

    if (!(def = virDomainDefParseString(xml, priv->hostCaps, priv->xmlOpt,
                                        NULL, parse_flags)))
        return NULL;

    if (!virStringMatch(def->name, DOCKER_NAME_REGEXP)) {
        virReportError(VIR_ERR_INVALID_ARG, "%s",
                       _("Defined incorrect name of container"));
        return NULL;
    }

    if ((seek = dockerSeekLabel(priv, def->uuid, false, &id_str)) < 0)
        return NULL;

    /* Define XML for existing domain so try to update it */
    if (seek)
        return dockerDomainRedefine(conn, id_str, def);

    if (!(obj = dockerDomainDefToJSON(priv, def, true)))
        return NULL;

    if (!(json = virJSONValueToString(obj, false)))
        return NULL;

    virAsprintf(&cmd, "containers/create?name=%s", def->name);

    if (!(id = dockerGetJSONObject(priv, cmd, true,
                                    DOCKER_API_HTTP_POST,
                                    json, 201)))
        return NULL;

    if (def->nnets > 1) {
        const char *id_str_new;

        if (!(id_str_new = virJSONValueObjectGetString(id, "Id"))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Incorrect 'Id' field in JSON document"));
            return NULL;
        }

        if (dockerAddNetsToContainer(priv, def, id_str_new) < 0) {
            dockerContainerDelete(priv, id_str_new);
            return NULL;
        }
    }

    return virGetDomain(conn, def->name, def->uuid, -1);
}


static virDomainPtr
dockerDomainDefineXML(virConnectPtr conn, const char *xml)
{
    return dockerDomainDefineXMLFlags(conn, xml, 0);
}


static int
dockerDomainUndefineFlags(virDomainPtr domain, unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOFREE(char *) id_str = NULL;

    virCheckFlags(0, -1);

    if (dockerSeekLabel(priv, domain->uuid, false, &id_str) < 0)
        return -1;

    return dockerContainerDelete(priv, id_str);
}


static int
dockerDomainUndefine(virDomainPtr domain)
{
    return dockerDomainUndefineFlags(domain, 0);
}


static int
dockerDomainCreateWithFlags(virDomainPtr domain, unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOFREE(char *) id_str = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;

    virCheckFlags(VIR_DOMAIN_NONE, -1);

    if (dockerSeekLabel(priv, domain->uuid, false, &id_str) < 0)
        return -1;

    virAsprintf(&cmd, "containers/%s/start", id_str);

    if (!(response = dockerApiRequest(priv, cmd, true,
                                      DOCKER_API_HTTP_POST,
                                      NULL, 204)))
        return -1;

    return 0;
}


static int
dockerDomainCreate(virDomainPtr domain)
{
    return dockerDomainCreateWithFlags(domain, 0);
}


static int
dockerDomainDestroyFlags(virDomainPtr domain, unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOFREE(char *) id_str = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;

    virCheckFlags(VIR_DOMAIN_DESTROY_GRACEFUL, -1);

    if (dockerSeekLabel(priv, domain->uuid, false, &id_str) < 0)
        return -1;

    /*
     * By default we send SIGTERM and after that
     * wait 3 sec before sending SIGKILL
     */
    virAsprintf(&cmd, "containers/%s/stop%s", id_str,
                (flags & VIR_DOMAIN_DESTROY_GRACEFUL) ?
                    "" : "?t=3");

    if (!(response = dockerApiRequest(priv, cmd, true,
                                      DOCKER_API_HTTP_POST,
                                      NULL, 204)))
        return -1;

    return 0;
}


static int
dockerDomainDestroy(virDomainPtr domain)
{
    return dockerDomainDestroyFlags(domain, 0);
}


static int
dockerDomainSuspend(virDomainPtr domain)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOFREE(char *) id_str = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;

    if (dockerSeekLabel(priv, domain->uuid, false, &id_str) < 0)
        return -1;

    virAsprintf(&cmd, "containers/%s/pause", id_str);

    if (!(response = dockerApiRequest(priv, cmd, true,
                                      DOCKER_API_HTTP_POST,
                                      NULL, 204)))
        return -1;

    return 0;
}


static int
dockerDomainResume(virDomainPtr domain)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOFREE(char *) id_str = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;

    if (dockerSeekLabel(priv, domain->uuid, false, &id_str) < 0)
        return -1;

    virAsprintf(&cmd, "containers/%s/unpause", id_str);

    if (!(response = dockerApiRequest(priv, cmd, true,
                                      DOCKER_API_HTTP_POST,
                                      NULL, 204)))
        return -1;

    return 0;
}


static int
dockerDomainShutdownFlags(virDomainPtr domain, unsigned int flags)
{
    virCheckFlags(VIR_DOMAIN_SHUTDOWN_SIGNAL, -1);

    return dockerDomainDestroyFlags(domain, VIR_DOMAIN_DESTROY_GRACEFUL);
}


static int
dockerDomainShutdown(virDomainPtr domain)
{
    return dockerDomainShutdownFlags(domain, 0);
}


static int
dockerDomainReboot(virDomainPtr domain, unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOFREE(char *) id_str = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;

    virCheckFlags(VIR_DOMAIN_REBOOT_SIGNAL, -1);

    if (dockerSeekLabel(priv, domain->uuid, false, &id_str) < 0)
        return -1;

    virAsprintf(&cmd, "containers/%s/restart", id_str);

    if (!(response = dockerApiRequest(priv, cmd, true,
                                      DOCKER_API_HTTP_POST,
                                      NULL, 204)))
        return -1;

    return 0;
}


static int
dockerDomainRename(virDomainPtr domain,
                   const char *new_name,
                   unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOFREE(char *) id_str = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;

    virCheckFlags(0, -1);

    if (!virStringMatch(new_name, DOCKER_NAME_REGEXP)) {
        virReportError(VIR_ERR_INVALID_ARG, "%s",
                       _("Defined incorrect name of container"));
        return -1;
    }

    if (dockerSeekLabel(priv, domain->uuid, false, &id_str) < 0)
        return -1;

    virAsprintf(&cmd, "containers/%s/rename?name=%s", id_str, new_name);

    if (!(response = dockerApiRequest(priv, cmd, true,
                                      DOCKER_API_HTTP_POST,
                                      NULL, 204)))
        return -1;

    return 0;
}


static int
dockerDomainReset(virDomainPtr domain, unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOFREE(char *) id_str = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;

    virCheckFlags(0, -1);

    if (dockerSeekLabel(priv, domain->uuid, false, &id_str) < 0)
        return -1;

    virAsprintf(&cmd, "containers/%s/restart?t=0", id_str);

    if (!(response = dockerApiRequest(priv, cmd, true,
                                      DOCKER_API_HTTP_POST,
                                      NULL, 204)))
        return -1;

    return 0;
}


static int
dockerDomainBlockStats(virDomainPtr domain,
                       const char *path,
                       virDomainBlockStatsPtr stats)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    virJSONValuePtr stor_stats;
    unsigned long long rcn, wcn, read, write;

    if (path[0] && STRNEQ(path, "vda")) {
        virReportError(VIR_ERR_DEVICE_MISSING, "%s", path);
        return -1;
    }

    if (!(obj = dockerGetContainerStats(priv, domain->uuid)))
        return -1;

    if (!(stor_stats = virJSONValueObjectGet(obj, "storage_stats")) ||
        virJSONValueObjectGetNumberUlong(stor_stats, "read_count_normalized", &rcn) < 0 ||
        virJSONValueObjectGetNumberUlong(stor_stats, "write_count_normalized", &wcn) < 0 ||
        virJSONValueObjectGetNumberUlong(stor_stats, "read_size_bytes", &read) < 0 ||
        virJSONValueObjectGetNumberUlong(stor_stats, "write_size_bytes", &write) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'storage_stats' object in JSON document"));
        return -1;
    }

    stats->rd_req = rcn;
    stats->wr_req = wcn;
    stats->rd_bytes = read;
    stats->wr_bytes = write;
    stats->errs = -1;

    return 0;
}


static int
dockerDomainMemoryStats(virDomainPtr domain,
                        virDomainMemoryStatPtr stats,
                        unsigned int nr_stats,
                        unsigned int flags)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    virJSONValuePtr mem_stats;
    const char *timestamp;
    long long updated;
    unsigned long long cb, pws;
    int ret;

    virCheckFlags(0, -1);

    if (!(obj = dockerGetContainerStats(priv, domain->uuid)))
        return -1;

    if (!(mem_stats = virJSONValueObjectGet(obj, "memory_stats"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'memory_stats' object in JSON document"));
        return -1;
    }

    /* If container isn't running than memory_stats object is empty */
    if (!virJSONValueObjectKeysNumber(mem_stats))
        return 0;

    if (!(timestamp = virJSONValueObjectGetString(obj, "read")) ||
        (updated = dockerParseDateTime(timestamp)) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'read' field in JSON document"));
        return -1;
    }

    if (virJSONValueObjectGetNumberUlong(mem_stats, "commitbytes", &cb) < 0 ||
        virJSONValueObjectGetNumberUlong(mem_stats, "privateworkingset", &pws) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'memory_stats' object in JSON document"));
        return -1;
    }

    ret = 0;
    if (ret < nr_stats) {
        stats[ret].tag = VIR_DOMAIN_MEMORY_STAT_ACTUAL_BALLOON;
        stats[ret].val = pws / 1024; /* byte to kilobyte */
        ret++;
    }
    if (ret < nr_stats) {
        stats[ret].tag = VIR_DOMAIN_MEMORY_STAT_RSS;
        stats[ret].val = cb / 1024; /* byte to kilobyte */
        ret++;
    }
    if (ret < nr_stats) {
        stats[ret].tag = VIR_DOMAIN_MEMORY_STAT_LAST_UPDATE;
        stats[ret].val = updated;
        ret++;
    }

    return ret;
}


static int
dockerDomainInterfaceStats(virDomainPtr domain,
                           const char *device,
                           virDomainInterfaceStatsPtr stats)
{
    dockerPrivatePtr priv = domain->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    VIR_AUTOPTR(virJSONValue) ostats = NULL;
    virJSONValuePtr net_sets, networks, iface;
    size_t i;

    if (!(obj = dockerGetContainerByUUID(priv, domain->uuid)))
        return -1;

    if (!(net_sets = virJSONValueObjectGet(obj, "NetworkSettings")) ||
        !(networks = virJSONValueObjectGet(net_sets, "Networks")) ||
        virJSONValueObjectKeysNumber(networks) < 1) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Networks' object in JSON container"));
        return -1;
    }

    /* Check requested device as network name or mac address */
    for (i = 0; i < virJSONValueObjectKeysNumber(networks); i++) {
        virJSONValuePtr config;
        const char *macaddr;

        iface = virJSONValueObjectGetValue(networks, i);
        if (STREQ(device, virJSONValueObjectGetKey(networks, i)) ||
            ((macaddr = virJSONValueObjectGetString(iface, "MacAddress")) &&
             STREQ(device, macaddr)))
            break;

        /*
         * For first interface mac address can be preset in config object
         * instead of network object. For other interfaces it can be present
         * only in network object because other interfaces was added to
         * existing container late.
         */
        if (!i &&
            (config = virJSONValueObjectGet(obj, "Config")) &&
            (macaddr = virJSONValueObjectGetString(config, "MacAddress")) &&
             STREQ(device, macaddr))
            break;
    }

    if (i >= virJSONValueObjectKeysNumber(networks)) {
        virReportError(VIR_ERR_NO_INTERFACE, "%s", device);
        return -1;
    }

    if (!(ostats = dockerGetContainerStats(priv, domain->uuid)))
        return -1;

    if (!(networks = virJSONValueObjectGet(ostats, "networks")) ||
        !(iface = virJSONValueObjectGetValue(networks, i))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'networks' object in JSON document"));
        return -1;
    }

    if (virJSONValueObjectGetNumberLong(iface, "rx_bytes", &stats->rx_bytes) < 0 ||
        virJSONValueObjectGetNumberLong(iface, "tx_bytes", &stats->tx_bytes) < 0 ||
        virJSONValueObjectGetNumberLong(iface, "rx_packets", &stats->rx_packets) < 0 ||
        virJSONValueObjectGetNumberLong(iface, "tx_packets", &stats->tx_packets) < 0 ||
        virJSONValueObjectGetNumberLong(iface, "rx_errors", &stats->rx_errs) < 0 ||
        virJSONValueObjectGetNumberLong(iface, "tx_errors", &stats->tx_errs) < 0 ||
        virJSONValueObjectGetNumberLong(iface, "rx_dropped", &stats->rx_drop) < 0 ||
        virJSONValueObjectGetNumberLong(iface, "tx_dropped", &stats->tx_drop) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Incorrect '%s' object in JSON document"),
                       virJSONValueObjectGetKey(networks, i));
        return -1;
    }

    return 0;
}


#define DOCKER_MIGRATION_FLAGS (VIR_MIGRATE_OFFLINE | \
                                VIR_MIGRATE_PERSIST_DEST | \
                                VIR_MIGRATE_UNDEFINE_SOURCE)


#define DOCKER_MIGRATION_PARAMS \
    VIR_MIGRATE_PARAM_URI,       VIR_TYPED_PARAM_STRING, \
    VIR_MIGRATE_PARAM_DEST_XML,  VIR_TYPED_PARAM_STRING, \
    VIR_MIGRATE_PARAM_DEST_NAME, VIR_TYPED_PARAM_STRING, \
    NULL


static char *
dockerDomainMigrateBegin3Params(virDomainPtr domain,
                                virTypedParameterPtr params,
                                int nparams,
                                char **cookieout ATTRIBUTE_UNUSED,
                                int *cookieoutlen ATTRIBUTE_UNUSED,
                                unsigned int flags)
{
    virCheckFlags(DOCKER_MIGRATION_FLAGS, NULL);

    if (virTypedParamsValidate(params, nparams, DOCKER_MIGRATION_PARAMS) < 0)
        return NULL;

    if (!(flags & VIR_MIGRATE_OFFLINE)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Supported only offline migration"));
        return NULL;
    }

    if (!(flags & VIR_MIGRATE_PERSIST_DEST)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Offline migration needs to persist destination"));
        return NULL;
    }

    if (dockerDomainIsActive(domain) > 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Source domain '%s' is active now"), domain->name);
        return NULL;
    }

    return dockerDomainGetXMLDesc(domain, VIR_DOMAIN_XML_INACTIVE |
                                          VIR_DOMAIN_XML_MIGRATABLE);
}


static int
dockerDomainMigratePrepare3Params(virConnectPtr dconn,
                                  virTypedParameterPtr params,
                                  int nparams,
                                  const char *cookiein ATTRIBUTE_UNUSED,
                                  int cookieinlen ATTRIBUTE_UNUSED,
                                  char **cookieout ATTRIBUTE_UNUSED,
                                  int *cookieoutlen ATTRIBUTE_UNUSED,
                                  char **uri_out,
                                  unsigned int flags)
{
    dockerPrivatePtr priv = dconn->privateData;
    VIR_AUTOPTR(virDomainDef) def = NULL;
    const char *uri_in = NULL;
    const char *dom_xml = NULL;
    const char *dname = NULL;
    virDomainPtr domain;
    int seek;

    virCheckFlags(DOCKER_MIGRATION_FLAGS, -1);

    if (virTypedParamsValidate(params, nparams, DOCKER_MIGRATION_PARAMS) < 0)
        return -1;

    if (virTypedParamsGetString(params, nparams, VIR_MIGRATE_PARAM_URI,
                                &uri_in) < 0 ||
        virTypedParamsGetString(params, nparams, VIR_MIGRATE_PARAM_DEST_XML,
                                &dom_xml) < 0 ||
        virTypedParamsGetString(params, nparams, VIR_MIGRATE_PARAM_DEST_NAME,
                                &dname) < 0)
        return -1;

    if (!dom_xml) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("No domain XML passed"));
        return -1;
    }

    if (!(def = virDomainDefParseString(dom_xml, priv->hostCaps,
                                        priv->xmlOpt, NULL,
                                        VIR_DOMAIN_DEF_PARSE_INACTIVE)))
        return -1;

    if (dname) {
        if (!virStringMatch(dname, DOCKER_NAME_REGEXP)) {
            virReportError(VIR_ERR_INVALID_ARG, "%s",
                           _("Defined incorrect name of container"));
            return -1;
        }

        VIR_FREE(def->name);
        VIR_STRDUP(def->name, dname);
    }

    if ((domain = dockerLookupContainer(dconn, "name", def->name))) {
        virObjectUnref(domain);
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Destination domain '%s' already exists"), def->name);
        return -1;
    }

    if ((seek = dockerSeekLabel(priv, def->uuid, false, NULL)) < 0)
        return -1;

    if (seek) {
        char uuidstr[VIR_UUID_STRING_BUFLEN];

        virUUIDFormat(def->uuid, uuidstr);
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Destination domain '%s' already exists"), uuidstr);
        return -1;
    }

    if (dname) {
        VIR_AUTOFREE(char *) xml = NULL;

        if (!(xml = virDomainDefFormat(def, NULL,
                                       VIR_DOMAIN_DEF_FORMAT_INACTIVE |
                                       VIR_DOMAIN_DEF_FORMAT_MIGRATABLE)))
            return -1;

        if (virTypedParamsReplaceString(&params, &nparams,
                                        VIR_MIGRATE_PARAM_DEST_XML,
                                        xml) < 0)
            return -1;
    }

    if (!uri_in)
        *uri_out = virConnectGetURI(dconn);

    return 0;
}


static int
dockerDomainMigratePerform3Params(virDomainPtr domain ATTRIBUTE_UNUSED,
                                  const char *dconnuri ATTRIBUTE_UNUSED,
                                  virTypedParameterPtr params,
                                  int nparams,
                                  const char *cookiein ATTRIBUTE_UNUSED,
                                  int cookieinlen ATTRIBUTE_UNUSED,
                                  char **cookieout ATTRIBUTE_UNUSED,
                                  int *cookieoutlen ATTRIBUTE_UNUSED,
                                  unsigned int flags)
{
    virCheckFlags(DOCKER_MIGRATION_FLAGS, -1);

    if (virTypedParamsValidate(params, nparams, DOCKER_MIGRATION_PARAMS) < 0)
        return -1;

    /* Do nothing here for offline migration */
    return 0;
}


static virDomainPtr
dockerDomainMigrateFinish3Params(virConnectPtr dconn,
                                 virTypedParameterPtr params,
                                 int nparams,
                                 const char *cookiein ATTRIBUTE_UNUSED,
                                 int cookieinlen ATTRIBUTE_UNUSED,
                                 char **cookieout ATTRIBUTE_UNUSED,
                                 int *cookieoutlen ATTRIBUTE_UNUSED,
                                 unsigned int flags,
                                 int cancelled)
{
    const char *dom_xml = NULL;

    virCheckFlags(DOCKER_MIGRATION_FLAGS, NULL);

    if (virTypedParamsValidate(params, nparams, DOCKER_MIGRATION_PARAMS) < 0)
        return NULL;

    if (cancelled)
        return NULL;

    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_DEST_XML,
                                &dom_xml) < 0)
        return NULL;

    if (!dom_xml) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("No domain XML passed"));
        return NULL;
    }

    return dockerDomainDefineXMLFlags(dconn, dom_xml, 0);
}


static int
dockerDomainMigrateConfirm3Params(virDomainPtr domain,
                                  virTypedParameterPtr params,
                                  int nparams,
                                  const char *cookiein ATTRIBUTE_UNUSED,
                                  int cookieinlen ATTRIBUTE_UNUSED,
                                  unsigned int flags,
                                  int cancelled)
{
    virCheckFlags(DOCKER_MIGRATION_FLAGS, -1);

    if (virTypedParamsValidate(params, nparams, DOCKER_MIGRATION_PARAMS) < 0)
        return -1;

    if (cancelled)
        return 0;

    if ((flags & VIR_MIGRATE_UNDEFINE_SOURCE) &&
        dockerDomainUndefineFlags(domain, 0) < 0)
        return -1;

    return 0;
}


static int
dockerNodeGetInfo(virConnectPtr conn, virNodeInfoPtr info)
{
    dockerPrivatePtr priv = conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    unsigned int ncpu;
    unsigned long long memory;

    if (!(obj = dockerGetJSONObject(priv, "info", true,
                                    DOCKER_API_HTTP_GET,
                                    NULL, 200)))
        return -1;

    if (virJSONValueObjectGetNumberUint(obj, "NCPU", &ncpu) < 0 ||
        virJSONValueObjectGetNumberUlong(obj, "MemTotal", &memory) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect format 'NCPU' or 'MemTotal' fields in JSON document"));
        return -1;
    }

    memset(info, 0, sizeof(*info));
    info->memory = memory / 1024; /* byte to kilobyte */
    info->cpus = ncpu;
    info->nodes = 1;
    info->sockets = 1;
    info->cores = ncpu;
    info->threads = 1;

    return 0;
}


static int
dockerConnectListAllNetworks(virConnectPtr conn,
                             virNetworkPtr **networks,
                             unsigned int flags)
{
    dockerPrivatePtr priv = conn->privateData;
    int ret = -1;
    int nnets = 0;
    VIR_AUTOPTR(virJSONValue) list = NULL;
    virNetworkPtr *nets = NULL;
    size_t i;

    virCheckFlags(VIR_CONNECT_LIST_NETWORKS_FILTERS_ALL, -1);

    /*
     * Check for filter combinations that return no results:
     * active: all docker networks are always active
     * persistent: all docker networks are always persistent
     * autostart: all docker networks are always autostart
     */
    if (((flags & VIR_CONNECT_LIST_NETWORKS_INACTIVE) &&
         !(flags & VIR_CONNECT_LIST_NETWORKS_ACTIVE)) ||
        ((flags & VIR_CONNECT_LIST_NETWORKS_TRANSIENT) &&
         !(flags & VIR_CONNECT_LIST_NETWORKS_PERSISTENT)) ||
        ((flags & VIR_CONNECT_LIST_NETWORKS_NO_AUTOSTART) &&
         !(flags & VIR_CONNECT_LIST_NETWORKS_AUTOSTART))) {
        if (networks)
            VIR_ALLOC(*networks);
        return 0;
    }

    /* Encoded filter: {"scope":["local"]} */
    if (!(list = dockerGetJSONObject(priv, "networks?filters="
                                     "{%22scope%22:[%22local%22]}",
                                    true, DOCKER_API_HTTP_GET,
                                    NULL, 200)))
        return -1;

    if (!virJSONValueIsArray(list)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received not JSON array"));
        return -1;
    }

    nnets = virJSONValueArraySize(list);
    if (!networks)
        return nnets;

    VIR_ALLOC_N(nets, nnets + 1);

    for (i = 0; i < nnets; i++) {
        virJSONValuePtr obj = virJSONValueArrayGet(list, i);
        if (!(nets[i] = dockerGetNetworkFromObject(conn, obj)))
            goto cleanup;
    }

    VIR_STEAL_PTR(*networks, nets);
    ret = nnets;

 cleanup:
    virObjectListFreeCount(nets, nnets);

    return ret;
}


static virNetworkPtr
dockerNetworkLookupByUUID(virConnectPtr conn, const unsigned char *uuid)
{
    dockerPrivatePtr priv = conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    if (!(obj = dockerGetNetworkJSON(priv, uuid)))
        return NULL;

    return dockerGetNetworkFromObject(conn, obj);
}


static virNetworkPtr
dockerNetworkLookupByName(virConnectPtr conn,
                          const char *name)
{
    dockerPrivatePtr priv = conn->privateData;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    virAsprintf(&cmd, "networks/%s", name);

    if (!(obj = dockerGetJSONObject(priv, cmd, true,
                                    DOCKER_API_HTTP_GET,
                                    NULL, 200)))
        return NULL;

    return dockerGetNetworkFromObject(conn, obj);
}


VIR_DEFINE_AUTOPTR_FUNC(virNetworkDef, virNetworkDefFree);


static char *
dockerNetworkGetXMLDesc(virNetworkPtr network, unsigned int flags)
{
    dockerPrivatePtr priv = network->conn->privateData;
    VIR_AUTOPTR(virNetworkDef) def = NULL;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    virJSONValuePtr conns, options, ipam, config;
    const char *driver, *dev;
    size_t i;

    virCheckFlags(VIR_NETWORK_XML_INACTIVE, NULL);

    if (!(obj = dockerGetNetworkJSON(priv, network->uuid)))
        return NULL;

    VIR_ALLOC(def);
    VIR_STRDUP(def->name, network->name);
    memcpy(def->uuid, network->uuid, VIR_UUID_BUFLEN);

    if (!(conns = virJSONValueObjectGet(obj, "Containers"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Containers' field in JSON network"));
        return NULL;
    }

    def->connections = virJSONValueObjectKeysNumber(conns);

    if (!(driver = virJSONValueObjectGetString(obj, "Driver"))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Driver' field in JSON network"));
        return NULL;
    }

    if (STREQ(driver, "null")) {
        def->forward.type = VIR_NETWORK_FORWARD_NONE;
    } else if (STREQ(driver, "nat") || STREQ(driver, "ics")) {
        def->forward.type = VIR_NETWORK_FORWARD_NAT;
    } else if (STREQ(driver, "transparent")) {
        def->forward.type = VIR_NETWORK_FORWARD_ROUTE;
    } else {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Unsupported 'Driver' value '%s' for network"),
                       driver);
        return NULL;
    }

    /* JSON-object 'Options' and any fields in it can absent */
    if ((options = virJSONValueObjectGet(obj, "Options")) &&
        (dev = virJSONValueObjectGetString(options,
                                           "com.docker.network.windowsshim.interface"))) {
        VIR_ALLOC_N(def->forward.ifs, 1);
        VIR_STRDUP(def->forward.ifs[0].device.dev, dev);
        def->forward.ifs[0].type = VIR_NETWORK_FORWARD_HOSTDEV_DEVICE_NETDEV;
        def->forward.nifs = 1;
    }

    if (!(ipam = virJSONValueObjectGet(obj, "IPAM")) ||
        !(config = virJSONValueObjectGet(ipam, "Config")) ||
        !virJSONValueIsArray(config)) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Incorrect 'Config' object in JSON network"));
        return NULL;
    }

    for (i = 0; i < virJSONValueArraySize(config); i++) {
        virJSONValuePtr item = virJSONValueArrayGet(config, i);
        const char *subnet;

        if ((subnet = virJSONValueObjectGetString(item, "Subnet"))) {
            virNetworkIPDef ip;
            char **ipparts = virStringSplit(subnet, "/", 2);

            memset(&ip, 0, sizeof(ip));

            if (virStringListLength((const char * const *)ipparts) == 2 &&
                virSocketAddrParse(&ip.address, ipparts[0], AF_UNSPEC) >= 0 &&
                virStrToLong_uip(ipparts[1], NULL, 10, &ip.prefix) >= 0) {
                const char *gateway;

                /*
                 * If we don't found correct 'Gateway' field
                 * we ignore the docker info and don't add <ip>
                 */
                if (!(gateway = virJSONValueObjectGetString(item, "Gateway")) ||
                    virSocketAddrParse(&ip.address, gateway, AF_UNSPEC) < 0)
                    continue;

                VIR_STRDUP(ip.family,
                           (virSocketAddrNumericFamily(ipparts[0]) == AF_INET) ?
                           "ipv4" : "ipv6");

                VIR_APPEND_ELEMENT(def->ips, def->nips, ip);
            }

            virStringListFree(ipparts);
        }
    }

    return virNetworkDefFormat(def, NULL, flags);
}


static virNetworkPtr
dockerNetworkDefineXML(virConnectPtr conn, const char *xml)
{
    dockerPrivatePtr priv = conn->privateData;
    VIR_AUTOPTR(virNetworkDef) def = NULL;
    VIR_AUTOPTR(virJSONValue) labels = NULL;
    VIR_AUTOPTR(virJSONValue) options = NULL;
    VIR_AUTOPTR(virJSONValue) config = NULL;
    VIR_AUTOPTR(virJSONValue) ipam = NULL;
    VIR_AUTOPTR(virJSONValue) obj = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) json = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    const char *driver;
    bool ipv6;
    int seek;
    size_t i;

    if (!(def = virNetworkDefParseString(xml, NULL)))
        return NULL;

    if ((seek = dockerSeekLabel(priv, def->uuid, true, NULL)) < 0)
        return NULL;

    virUUIDFormat(def->uuid, uuidstr);
    if (seek) {
        virReportError(VIR_ERR_NETWORK_EXIST, "%s", uuidstr);
        return NULL;
    }

    if (virJSONValueObjectCreate(&labels,
                                 "s:"LABEL_LIBVIRT_UUID, uuidstr,
                                 NULL) < 0)
        return NULL;

    /* Special options for Docker network driver on windows */
    if (virJSONValueObjectCreate(&options,
                                 "s:com.docker.network.windowsshim.networkname",
                                 def->name,
                                 "S:com.docker.network.windowsshim.interface",
                                 virNetworkDefForwardIf(def, 0),
                                 NULL) < 0)
        return NULL;

    config = virJSONValueNewArray();

    for (i = 0; i < def->nips; i++) {
        virNetworkIPDefPtr ip = &def->ips[i];
        VIR_AUTOPTR(virJSONValue) item = NULL;
        VIR_AUTOFREE(char *) subnet = NULL;
        VIR_AUTOFREE(char *) gw = virSocketAddrFormat(&ip->address);

        if (STREQ(gw, VIR_SOCKET_ADDR_IPV4_ALL) ||
            STREQ(gw, VIR_SOCKET_ADDR_IPV6_ALL)) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                           _("Unsupported address '%s'"), gw);
            return NULL;
        }

        virAsprintf(&subnet, "%s/%d", gw,
                    virNetworkIPDefPrefix(ip));
        if (virJSONValueObjectCreate(&item,
                                     "s:Subnet", subnet,
                                     "s:Gateway", gw,
                                     NULL) < 0)
            return NULL;

        virJSONValueArrayAppend(config, item);
        item = NULL;
    }

    if (virJSONValueObjectCreate(&ipam,
                                 "s:Driver", "default",
                                 "a:Config", &config,
                                 NULL) < 0)
        return NULL;

    switch (def->forward.type) {
    case VIR_NETWORK_FORWARD_NONE:
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("Cannot create network with type 'none'"));
        return NULL;
    case VIR_NETWORK_FORWARD_NAT:
        driver = "nat";
        break;
    case VIR_NETWORK_FORWARD_ROUTE:
        driver = "transparent";
        break;
    default:
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Unsupported type of network '%s'"),
                       virNetworkForwardTypeToString(def->forward.type));
        return NULL;
    }

    ipv6 = virNetworkDefGetIPByIndex(def, AF_INET6, 0);

    if (virJSONValueObjectCreate(&obj,
                                 "s:Name", def->name,
                                 "s:Driver", driver,
                                 "b:EnableIPv6", ipv6,
                                 "b:CheckDuplicate", true,
                                 "a:IPAM", &ipam,
                                 "a:Labels", &labels,
                                 "a:Options", &options,
                                 NULL) < 0)
        return NULL;

    if (!(json = virJSONValueToString(obj, false)))
        return NULL;

    if (!(response = dockerApiRequest(priv, "networks/create",
                                      true, DOCKER_API_HTTP_POST,
                                      json, 201)))
        return NULL;

    return virGetNetwork(conn, def->name, def->uuid);
}


static int
dockerNetworkUndefine(virNetworkPtr network)
{
    dockerPrivatePtr priv = network->conn->privateData;
    VIR_AUTOFREE(char *) id_str = NULL;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;

    if (dockerSeekLabel(priv, network->uuid, true, &id_str) < 0)
        return -1;

    virAsprintf(&cmd, "networks/%s", id_str);

    if (!(response = dockerApiRequest(priv, cmd, true,
                                      DOCKER_API_HTTP_DELETE,
                                      NULL, 204)))
        return -1;

    return 0;
}


static int
dockerNetworkGetAutostart(virNetworkPtr network, int *autostart)
{
    dockerPrivatePtr priv = network->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    if (!(obj = dockerGetNetworkJSON(priv, network->uuid)))
        return -1;

    /* Docker networks are always autostart */
    *autostart = 1;

    return 0;
}


static int
dockerNetworkSetAutostart(virNetworkPtr network, int autostart)
{
    dockerPrivatePtr priv = network->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    if (!(obj = dockerGetNetworkJSON(priv, network->uuid)))
        return -1;

    /* Accept autostart activation but fail on autostart deactivation */
    if (!autostart) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Cannot deactivate network autostart"));
        return -1;
    }

    return 0;
}


static int
dockerNetworkIsActive(virNetworkPtr network)
{
    dockerPrivatePtr priv = network->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    if (!(obj = dockerGetNetworkJSON(priv, network->uuid)))
        return -1;

    /* Docker networks are always active */
    return 1;
}


static int
dockerNetworkIsPersistent(virNetworkPtr network)
{
    dockerPrivatePtr priv = network->conn->privateData;
    VIR_AUTOPTR(virJSONValue) obj = NULL;

    if (!(obj = dockerGetNetworkJSON(priv, network->uuid)))
        return -1;

    /* Docker networks are always persistent */
    return 1;
}


static int
dockerConnectListAllStoragePools(virConnectPtr conn,
                                 virStoragePoolPtr **pools,
                                 unsigned int flags)
{
    int ret = -1;
    int npools = 0;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    VIR_AUTOSTRINGLIST lines = NULL;
    virStoragePoolPtr *pls = NULL;
    size_t i, nlines, npools_max = 0;

    virCheckFlags(VIR_CONNECT_LIST_STORAGE_POOLS_FILTERS_ACTIVE |
                  VIR_CONNECT_LIST_STORAGE_POOLS_FILTERS_PERSISTENT |
                  VIR_CONNECT_LIST_STORAGE_POOLS_FILTERS_AUTOSTART |
                  VIR_CONNECT_LIST_STORAGE_POOLS_WSS, -1);

    /*
     * Check for filter combinations that return no results:
     * persistent: all storage pools are always persistent
     * autostart: all storage pools are always autostart
     */
    if (((flags & VIR_CONNECT_LIST_STORAGE_POOLS_TRANSIENT) &&
         !(flags & VIR_CONNECT_LIST_STORAGE_POOLS_PERSISTENT)) ||
        ((flags & VIR_CONNECT_LIST_STORAGE_POOLS_NO_AUTOSTART) &&
         !(flags & VIR_CONNECT_LIST_STORAGE_POOLS_AUTOSTART))) {
        if (pools)
            VIR_ALLOC(*pools);
        return 0;
    }

    virAsprintf(&cmd, "(Get-StoragePool -IsPrimordial:0%s).UniqueId",
                flags & VIR_CONNECT_LIST_STORAGE_POOLS_ACTIVE ?
                (flags & VIR_CONNECT_LIST_STORAGE_POOLS_INACTIVE ?
                 "" : "|Where-Object HealthStatus -ne 'Unknown'") :
                (flags & VIR_CONNECT_LIST_STORAGE_POOLS_INACTIVE ?
                 "|Where-Object HealthStatus -eq 'Unknown'" : ""));

    if (dockerWinrsPowershell(conn, cmd, &response))
        return -1;

    if (virStringIsEmpty(response)) {
        if (pools)
            VIR_ALLOC(*pools);
        return 0;
    }

    lines = virStringSplit(response, "\r\n", 0);
    nlines = virStringListLength((const char * const *)lines);

    for (i = 0; i < nlines; i++) {
        char uuidstr[VIR_UUID_STRING_BUFLEN];

        if (virStringIsEmpty(lines[i]))
            continue;

        if (sscanf(lines[i], "{%36s}", uuidstr) != 1)
            continue;

        if (pools) {
            unsigned char uuid[VIR_UUID_BUFLEN];

            if (virUUIDParse(uuidstr, uuid) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR,
                               _("Could not parse UUID from string '%s'"),
                               uuidstr);
                goto cleanup;
            }

            VIR_RESIZE_N(pls, npools_max, npools, 1);
            pls[npools] = virGetStoragePool(conn, uuidstr, uuid, NULL, NULL);
        }

        npools++;
    }

    if (pools) {
        VIR_RESIZE_N(pls, npools_max, npools, 1);
        VIR_STEAL_PTR(*pools, pls);
    }
    ret = npools;

 cleanup:
    virObjectListFreeCount(pls, npools);

    return ret;
}


static virStoragePoolPtr
dockerStoragePoolLookupByUUID(virConnectPtr conn, const unsigned char *uuid)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    virAsprintf(&cmd, "(Get-StoragePool -IsPrimordial:0 -UniqueId:'{%s}')"
                ".Name", virUUIDFormat(uuid, uuidstr));

    if (dockerWinrsPowershell(conn, cmd, NULL))
        return NULL;

    return virGetStoragePool(conn, uuidstr, uuid, NULL, NULL);
}


static virStoragePoolPtr
dockerStoragePoolLookupByName(virConnectPtr conn, const char *name)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    unsigned char uuid[VIR_UUID_BUFLEN];

    virAsprintf(&cmd, "Get-StoragePool -IsPrimordial:0|"
                "Where-Object UniqueId -like '{%s*}'|"
                "ForEach-Object {$c=0} {if($c -ne 0){throw}"
                "$c=1;return $_.UniqueId}", name);

    if (dockerWinrsPowershell(conn, cmd, &response))
        return NULL;

    if (virStringIsEmpty(response)) {
        virReportError(VIR_ERR_NO_STORAGE_VOL,
                       _("Storage pool with name '%s' not found"), name);
        return NULL;
    }

    if (sscanf(response, "{%36s}", uuidstr) != 1 ||
        virUUIDParse(uuidstr, uuid) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Could not parse UUID from string '%s'"), uuidstr);
        return NULL;
    }

    return virGetStoragePool(conn, uuidstr, uuid, NULL, NULL);
}


static virStoragePoolPtr
dockerStoragePoolLookupByVolume(virStorageVolPtr vol)
{
    unsigned char uuid[VIR_UUID_BUFLEN];

    if (virUUIDParse(vol->pool, uuid) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Could not parse UUID from string '%s'"), vol->pool);
        return NULL;
    }

    return virGetStoragePool(vol->conn, vol->pool, uuid, NULL, NULL);
}


static char *
dockerStoragePoolGetXMLDesc(virStoragePoolPtr pool, unsigned int flags)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    VIR_AUTOPTR(virStoragePoolDef) def = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    unsigned long long allocated;

    virCheckFlags(0, NULL);

    virAsprintf(&cmd, "(Get-StoragePool -UniqueId:'{%s}'|"
                "Format-Table -HideTableHeaders "
                "Size,AllocatedSize|Out-String).Trim()",
                virUUIDFormat(pool->uuid, uuidstr));

    if (dockerWinrsPowershell(pool->conn, cmd, &response))
        return NULL;

    if (virStringIsEmpty(response)) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Cannot get info on storage pool '%s'"), uuidstr);
        return NULL;
    }

    VIR_ALLOC(def);
    if (sscanf(response, "%llu %llu", &def->capacity,
               &allocated) != 2) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received badly formatted data from host"));
        return NULL;
    }

    VIR_STRDUP(def->name, uuidstr);
    memcpy(def->uuid, pool->uuid, VIR_UUID_BUFLEN);
    def->type = VIR_STORAGE_POOL_WSS;
    def->allocation = def->capacity;
    def->available = def->capacity - allocated;
    VIR_STRDUP(def->target.path, "");

    return virStoragePoolDefFormat(def);
}


static int
dockerStoragePoolGetInfo(virStoragePoolPtr pool, virStoragePoolInfoPtr info)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    unsigned long long allocated;
    unsigned int status;

    virAsprintf(&cmd, "(Get-StoragePool -UniqueId:'{%s}'|"
                "Format-Table -HideTableHeaders Size,AllocatedSize,"
                "{$_.CimInstanceProperties['HealthStatus'].Value}|"
                "Out-String).Trim()", virUUIDFormat(pool->uuid, uuidstr));

    if (dockerWinrsPowershell(pool->conn, cmd, &response))
        return -1;

    if (virStringIsEmpty(response)) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Cannot get info on storage pool '%s'"), uuidstr);
        return -1;
    }

    if (sscanf(response, "%llu %llu %u", &info->capacity,
               &allocated, &status) != 3) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received badly formatted data from host"));
        return -1;
    }

    switch (status) {
    case 0: /* Healthy */
        info->state = VIR_STORAGE_POOL_RUNNING;
        break;
    case 1: /* Warning */
        info->state = VIR_STORAGE_POOL_DEGRADED;
        break;
    case 2: /* Unhealthy */
        info->state = VIR_STORAGE_POOL_INACCESSIBLE;
        break;
    default: /* Unknown */
        info->state = VIR_STORAGE_POOL_INACTIVE;
    }

    info->allocation = info->capacity;
    info->available = info->capacity - allocated;

    return 0;
}


static int
dockerStoragePoolGetAutostart(virStoragePoolPtr pool, int *autostart)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    virAsprintf(&cmd, "(Get-StoragePool -UniqueId:'{%s}').Name",
                virUUIDFormat(pool->uuid, uuidstr));

    if (dockerWinrsPowershell(pool->conn, cmd, NULL))
        return -1;

    /* Windows storage pools are always autostart */
    *autostart = 1;

    return 0;
}


static int
dockerStoragePoolSetAutostart(virStoragePoolPtr pool, int autostart)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    virAsprintf(&cmd, "(Get-StoragePool -UniqueId:'{%s}').Name",
                virUUIDFormat(pool->uuid, uuidstr));

    if (dockerWinrsPowershell(pool->conn, cmd, NULL))
        return -1;

    /* Accept autostart activation but fail on autostart deactivation */
    if (!autostart) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Cannot deactivate storage pool autostart"));
        return -1;
    }

    return 0;
}


static int
dockerStoragePoolIsActive(virStoragePoolPtr pool)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    unsigned int status;

    virAsprintf(&cmd, "(Get-StoragePool -UniqueId:'{%s}')"
                ".CimInstanceProperties['HealthStatus'].Value",
                virUUIDFormat(pool->uuid, uuidstr));

    if (dockerWinrsPowershell(pool->conn, cmd, &response))
        return -1;

    if (virStringIsEmpty(response)) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Cannot get the status of storage pool '%s'"),
                       uuidstr);
        return -1;
    }

    if (sscanf(response, "%u", &status) != 1) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received badly formatted data from host"));
        return -1;
    }

    return status < 3 ? 1 : 0;
}


static int
dockerStoragePoolIsPersistent(virStoragePoolPtr pool)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    virAsprintf(&cmd, "(Get-StoragePool -UniqueId:'{%s}').Name",
                virUUIDFormat(pool->uuid, uuidstr));

    if (dockerWinrsPowershell(pool->conn, cmd, NULL))
        return -1;

    /* All Windows storage pools are persistent */
    return 1;
}


static int
dockerStoragePoolRefresh(virStoragePoolPtr pool, unsigned int flags)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    virCheckFlags(0, -1);

    virAsprintf(&cmd, "Update-StorageProviderCache -RootObject"
                "([ref](Get-StoragePool -UniqueId:'{%s}'))",
                virUUIDFormat(pool->uuid, uuidstr));

    if (dockerWinrsPowershell(pool->conn, cmd, NULL))
        return -1;

    return 0;
}


static int
dockerStoragePoolListAllVolumes(virStoragePoolPtr pool,
                                virStorageVolPtr **vols,
                                unsigned int flags)
{
    int nvols = 0;
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    VIR_AUTOSTRINGLIST lines = NULL;
    virStorageVolPtr *vls = NULL;
    size_t i, nlines, nvols_max = 0;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    virCheckFlags(0, -1);

    virAsprintf(&cmd, "(Get-Volume -StoragePool"
                "(Get-StoragePool -UniqueId:'{%s}')|"
                "Format-Table -HideTableHeaders "
                "UniqueId,FileSystemLabel|Out-String).Trim()",
                virUUIDFormat(pool->uuid, uuidstr));

    if (dockerWinrsPowershell(pool->conn, cmd, &response))
        return -1;

    if (virStringIsEmpty(response)) {
        if (vols)
            VIR_ALLOC(*vols);
        return 0;
    }

    lines = virStringSplit(response, "\r\n", 0);
    nlines = virStringListLength((const char * const *)lines);

    for (i = 0; i < nlines; i++) {
        VIR_AUTOFREE(char *) key = NULL;
        VIR_AUTOFREE(char *) name = NULL;

        if (virStringIsEmpty(lines[i]))
            continue;

        if (sscanf(lines[i], "%ms %ms", &key, &name) != 2)
            continue;

        if (vols) {
            VIR_RESIZE_N(vls, nvols_max, nvols, 1);
            vls[nvols] = virGetStorageVol(pool->conn, uuidstr,
                                          name, key, NULL, NULL);
        }

        nvols++;
    }

    if (vols) {
        VIR_RESIZE_N(vls, nvols_max, nvols, 1);
        VIR_STEAL_PTR(*vols, vls);
    }

    return nvols;
}


static virStorageVolPtr
dockerStorageVolLookupByName(virStoragePoolPtr pool, const char *name)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    virAsprintf(&cmd, "(Get-Volume -StoragePool"
                "(Get-StoragePool -UniqueId:'{%s}')|"
                "Where-Object FileSystemLabel -eq '%s'|"
                "Select-Object -First:1).UniqueId",
                virUUIDFormat(pool->uuid, uuidstr), name);

    if (dockerWinrsPowershell(pool->conn, cmd, &response))
        return NULL;

    if (virStringIsEmpty(response)) {
        virReportError(VIR_ERR_NO_STORAGE_VOL,
                       _("Volume with name '%s' not found"), name);
        return NULL;
    }

    /* Remove trailing "\r\n" from volume path */
    virTrimSpaces(response, NULL);
    virUUIDFormat(pool->uuid, uuidstr);

    return virGetStorageVol(pool->conn, uuidstr,
                            name, response, NULL, NULL);
}


static virStorageVolPtr
dockerStorageVolLookupByPath(virConnectPtr conn, const char *path)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) winpath = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    VIR_AUTOFREE(char *) name = NULL;
    VIR_AUTOFREE(char *) key = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    winpath = virStringReplace(path, "/", "\\");
    virAsprintf(&cmd, "(Get-StoragePool -Volume"
                "($v=Get-Volume -FilePath:'%s')|Format-Table "
                "-HideTableHeaders UniqueId,{$v.UniqueId},"
                "{$v.FileSystemLabel}|Out-String).Trim()", winpath);

    if (dockerWinrsPowershell(conn, cmd, &response))
        return NULL;

    if (virStringIsEmpty(response)) {
        virReportError(VIR_ERR_NO_STORAGE_VOL,
                       _("Volume with path '%s' not found"), winpath);
        return NULL;
    }

    if (sscanf(response, "{%36s} %ms %ms",
               uuidstr, &key, &name) != 3) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received badly formatted data from host"));
        return NULL;
    }

    return virGetStorageVol(conn, uuidstr, name,
                            key, NULL, NULL);
}


static virStorageVolPtr
dockerStorageVolLookupByKey(virConnectPtr conn, const char *key)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) winkey = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    VIR_AUTOFREE(char *) name = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    winkey = virStringReplace(key, "/", "\\");
    virAsprintf(&cmd, "(Get-StoragePool -Volume"
                "($v=Get-Volume -UniqueId:'%s')|Format-Table "
                "-HideTableHeaders UniqueId,{$v.FileSystemLabel}|"
                "Out-String).Trim()", winkey);

    if (dockerWinrsPowershell(conn, cmd, &response))
        return NULL;

    if (virStringIsEmpty(response)) {
        virReportError(VIR_ERR_NO_STORAGE_VOL,
                       _("Volume '%s' not found"), key);
        return NULL;
    }

    if (sscanf(response, "{%36s} %ms", uuidstr, &name) != 2) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received badly formatted data from host"));
        return NULL;
    }

    return virGetStorageVol(conn, uuidstr, name,
                            winkey, NULL, NULL);
}


static char *
dockerStorageVolGetXMLDesc(virStorageVolPtr vol, unsigned int flags)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    VIR_AUTOPTR(virStoragePoolDef) pooldef = NULL;
    VIR_AUTOPTR(virStorageVolDef) def = NULL;

    virCheckFlags(0, NULL);

    virAsprintf(&cmd, "(Get-VirtualDisk -Disk"
                "(Get-Disk -Partition(Get-Partition -Volume"
                "($v=Get-Volume -UniqueId:'%s')))|Format-Table "
                "-HideTableHeaders FootprintOnPool,{$v.Size},"
                "{(Get-WmiObject Win32_Volume|Where-Object DeviceID -eq "
                "$v.UniqueId).Name}|Out-String).Trim()", vol->key);

    if (dockerWinrsPowershell(vol->conn, cmd, &response))
        return NULL;

    if (virStringIsEmpty(response)) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Cannot get info on volume '%s'"), vol->key);
        return NULL;
    }

    VIR_ALLOC(def);
    if (sscanf(response, "%llu %llu %ms", &def->target.allocation,
               &def->target.capacity, &def->target.path) < 2) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received badly formatted data from host"));
        return NULL;
    }

    VIR_STRDUP(def->name, vol->name);
    VIR_STRDUP(def->key, vol->key);
    def->type = VIR_STORAGE_VOL_DIR;

    VIR_ALLOC(pooldef);
    pooldef->type = VIR_STORAGE_POOL_WSS;

    return virStorageVolDefFormat(pooldef, def);
}


static virStorageVolPtr
dockerStorageVolCreateXML(virStoragePoolPtr pool,
                          const char *xml,
                          unsigned int flags)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    VIR_AUTOPTR(virStoragePoolDef) pooldef = NULL;
    VIR_AUTOPTR(virStorageVolDef) def = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    virCheckFlags(0, NULL);

    VIR_ALLOC(pooldef);
    pooldef->type = VIR_STORAGE_POOL_WSS;

    if (!(def = virStorageVolDefParseString(pooldef, xml, 0)))
        return NULL;

    virUUIDFormat(pool->uuid, uuidstr);
    virAsprintf(&cmd, "(New-Volume -StoragePoolUniqueId:'{%s}' "
                "-FriendlyName:'%s' -Size:%llu).UniqueId",
                uuidstr, def->name, def->target.capacity);

    if (dockerWinrsPowershell(pool->conn, cmd, &response))
        return NULL;

    if (virStringIsEmpty(response)) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Failed to create volume '%s'"), def->name);
        return NULL;
    }

    /* Remove trailing "\r\n" from volume path */
    virTrimSpaces(response, NULL);

    return virGetStorageVol(pool->conn, uuidstr, def->name,
                            response, NULL, NULL);
}


static virStorageVolPtr
dockerStorageVolCreateXMLFrom(virStoragePoolPtr pool,
                              const char *xml,
                              virStorageVolPtr clonevol,
                              unsigned int flags)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    VIR_AUTOPTR(virStoragePoolDef) pooldef = NULL;
    VIR_AUTOPTR(virStorageVolDef) def = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    unsigned long long capacity;

    virCheckFlags(0, NULL);

    VIR_ALLOC(pooldef);
    pooldef->type = VIR_STORAGE_POOL_WSS;

    if (!(def = virStorageVolDefParseString(pooldef, xml, 0)))
        return NULL;

    virAsprintf(&cmd, "(Get-Volume -UniqueId:'%s').Size",
                clonevol->key);

    if (dockerWinrsPowershell(pool->conn, cmd, &response))
        return NULL;

    if (virStringIsEmpty(response)) {
        virReportError(VIR_ERR_NO_STORAGE_VOL,
                       _("Cannot get the size of volume '%s'"),
                       clonevol->key);
        return NULL;
    }

    if (sscanf(response, "%llu", &capacity) != 1) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received badly formatted data from host"));
        return NULL;
    }

    /*
     * Use the capacity of original volume in case the new
     * capacity is less than that, or it was omitted.
     */
    if (def->target.capacity < capacity)
        def->target.capacity = capacity;

    VIR_FREE(cmd);
    VIR_FREE(response);
    virUUIDFormat(pool->uuid, uuidstr);
    virAsprintf(&cmd, "(New-Volume -StoragePoolUniqueId:'{%s}' "
                "-FriendlyName:'%s' -Size:%llu).UniqueId",
                uuidstr, def->name, def->target.capacity);

    if (dockerWinrsPowershell(pool->conn, cmd, &response))
        return NULL;

    if (virStringIsEmpty(response)) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Failed to create volume '%s'"), def->name);
        return NULL;
    }

    /* Remove trailing "\r\n" from volume path */
    virTrimSpaces(response, NULL);

    VIR_FREE(cmd);
    virAsprintf(&cmd, "Copy-Item -Path((Get-WmiObject Win32_Volume|"
                "Where-Object DeviceID -eq '%s').Name+'*') "
                "-Destination(Get-WmiObject Win32_Volume|"
                "Where-Object DeviceID -eq '%s').Name "
                "-Recurse:1 -Confirm:0", clonevol->key, response);

    if (dockerWinrsPowershell(pool->conn, cmd, NULL)) {
        VIR_FREE(cmd);
        virAsprintf(&cmd, "Get-VirtualDisk -Disk"
                    "(Get-Disk -Partition(Get-Partition -Volume"
                    "(Get-Volume -UniqueId:'%s')))|"
                    "Remove-VirtualDisk -Confirm:0", response);
        dockerWinrsPowershell(pool->conn, cmd, NULL);

        return NULL;
    }

    return virGetStorageVol(pool->conn, uuidstr, def->name,
                            response, NULL, NULL);
}


static int
dockerStorageVolDelete(virStorageVolPtr vol, unsigned int flags)
{
    VIR_AUTOFREE(char *) cmd = NULL;

    virCheckFlags(0, -1);

    virAsprintf(&cmd, "Get-VirtualDisk -Disk"
                "(Get-Disk -Partition(Get-Partition -Volume"
                "(Get-Volume -UniqueId:'%s')))|"
                "Remove-VirtualDisk -Confirm:0",
                vol->key);

    if (dockerWinrsPowershell(vol->conn, cmd, NULL))
        return -1;

    return 0;
}


static int
dockerStorageVolGetInfoFlags(virStorageVolPtr vol,
                             virStorageVolInfoPtr info,
                             unsigned int flags)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;

    virCheckFlags(0, -1);

    virAsprintf(&cmd, "(Get-VirtualDisk -Disk"
                "(Get-Disk -Partition(Get-Partition -Volume"
                "($v=Get-Volume -UniqueId:'%s')))|Format-Table "
                "-HideTableHeaders FootprintOnPool,{$v.Size}|"
                "Out-String).Trim()", vol->key);

    if (dockerWinrsPowershell(vol->conn, cmd, &response))
        return -1;

    if (virStringIsEmpty(response)) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Cannot get info on volume '%s'"), vol->key);
        return -1;
    }

    if (sscanf(response, "%llu %llu", &info->allocation,
               &info->capacity) != 2) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received badly formatted data from host"));
        return -1;
    }

    info->type = VIR_STORAGE_VOL_DIR;

    return 0;
}


static int
dockerStorageVolGetInfo(virStorageVolPtr vol, virStorageVolInfoPtr info)
{
    return dockerStorageVolGetInfoFlags(vol, info, 0);
}


static char *
dockerStorageVolGetPath(virStorageVolPtr vol)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    char *path;

    virAsprintf(&cmd, "(Get-WmiObject Win32_Volume|"
                "Where-Object DeviceID -eq '%s').Name", vol->key);

    if (dockerWinrsPowershell(vol->conn, cmd, &response))
        return NULL;

    if (virStringIsEmpty(response)) {
        virReportError(VIR_ERR_NO_STORAGE_VOL,
                       _("Cannot get the path of volume '%s'"), vol->key);
        return NULL;
    }

    /* Remove trailing "\r\n" from volume path */
    virTrimSpaces(response, NULL);
    VIR_STRDUP(path, response);

    return path;
}


static int
dockerStorageVolResize(virStorageVolPtr vol,
                       unsigned long long capacity,
                       unsigned int flags)
{
    VIR_AUTOFREE(char *) cmd = NULL;
    VIR_AUTOFREE(char *) response = NULL;
    unsigned long long cur_capacity, delta, available;
    int is_tiered;

    virCheckFlags(VIR_STORAGE_VOL_RESIZE_ALLOCATE |
                  VIR_STORAGE_VOL_RESIZE_DELTA, -1);

    virAsprintf(&cmd, "(Get-VirtualDiskSupportedSize "
                "-StoragePoolUniqueId(Get-StoragePool "
                "-VirtualDisk($d=Get-VirtualDisk -Disk"
                "(Get-Disk -Partition(Get-Partition -Volume"
                "($v=Get-Volume -UniqueId:'%s'))))).UniqueId|"
                "Format-Table -HideTableHeaders VirtualDiskSizeMax,"
                "{$v.Size},{[int]$d.IsTiered}|Out-String).Trim()", vol->key);

    if (dockerWinrsPowershell(vol->conn, cmd, &response))
        return -1;

    if (sscanf(response, "%llu %llu %d", &available,
               &cur_capacity, &is_tiered) != 3) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Received badly formatted data from host"));
        return -1;
    }

    if (flags & VIR_STORAGE_VOL_RESIZE_DELTA) {
        delta = capacity;
    } else {
        if (capacity < cur_capacity) {
            virReportError(VIR_ERR_INVALID_ARG, "%s",
                           _("Cannot shrink capacity if the shrink flag is not set"));
            return -1;
        }

        delta = capacity - cur_capacity;
    }

    if (delta > available) {
         virReportError(VIR_ERR_OPERATION_FAILED, "%s",
                        _("Not enough space left in the storage pool"));
         return -1;
    }

    /* Phase 1: increase capacity of virtual disk or tiers */
    VIR_FREE(cmd);
    virAsprintf(&cmd, is_tiered ?
                "($t=Get-StorageTier -VirtualDisk"
                "(Get-VirtualDisk -Disk(Get-Disk -Partition"
                "(Get-Partition -Volume(Get-Volume "
                "-UniqueId:'%s')))))|ForEach-Object {$_|"
                "Resize-StorageTier -Size($_.Size+%llu*$_.Size/($t|"
                "Measure-Object Size -Sum).Sum)}" :
                "($v=Get-VirtualDisk -Disk(Get-Disk -Partition"
                "(Get-Partition -Volume(Get-Volume "
                "-UniqueId:'%s'))))|Resize-VirtualDisk "
                "-Size($v.Size+%llu)", vol->key, delta);

    if (dockerWinrsPowershell(vol->conn, cmd, NULL))
        return -1;

    /* Phase 2: change partition size for filling virtual disk */
    VIR_FREE(cmd);
    virAsprintf(&cmd, "($p=Get-Partition -Volume(Get-Volume "
                "-UniqueId:'%s'))|Resize-Partition -Size"
                "(($p|Get-PartitionSupportedSize).SizeMax)",
                vol->key);

    if (dockerWinrsPowershell(vol->conn, cmd, NULL))
        return -1;

    return 0;
}


static virHypervisorDriver dockerHypervisorDriver = {
    .name = "Docker",
    .connectOpen = dockerConnectOpen, /* 5.6.0 */
    .connectClose = dockerConnectClose, /* 5.6.0 */
    .connectGetType = dockerConnectGetType, /* 5.6.0 */
    .connectGetVersion = dockerConnectGetVersion, /* 5.6.0 */
    .connectIsEncrypted = dockerConnectIsEncrypted, /* 5.6.0 */
    .connectIsSecure = dockerConnectIsSecure, /* 5.6.0 */
    .connectIsAlive = dockerConnectIsAlive, /* 5.6.0 */
    .connectGetHostname = dockerConnectGetHostname, /* 5.6.0 */
    .connectGetMaxVcpus = dockerConnectGetMaxVcpus, /* 5.6.0 */
    .connectGetCapabilities = dockerConnectGetCapabilities, /* 5.6.0 */
    .connectListAllDomains = dockerConnectListAllDomains, /* 5.6.0 */
    .connectGetAllDomainStats = dockerConnectGetAllDomainStats, /* 5.6.0 */
    .connectSupportsFeature = dockerConnectSupportsFeature, /* 5.6.0 */
    .domainLookupByID = dockerDomainLookupByID, /* 5.6.0 */
    .domainLookupByUUID = dockerDomainLookupByUUID, /* 5.6.0 */
    .domainLookupByName = dockerDomainLookupByName, /* 5.6.0 */
    .domainIsActive = dockerDomainIsActive, /* 5.6.0 */
    .domainIsPersistent = dockerDomainIsPersistent, /* 5.6.0 */
    .domainIsUpdated = dockerDomainIsUpdated, /* 5.6.0 */
    .domainHasManagedSaveImage = dockerDomainHasManagedSaveImage, /* 5.6.0 */
    .domainGetAutostart = dockerDomainGetAutostart, /* 5.6.0 */
    .domainSetAutostart = dockerDomainSetAutostart, /* 5.6.0 */
    .domainGetInfo = dockerDomainGetInfo, /* 5.6.0 */
    .domainGetState = dockerDomainGetState, /* 5.6.0 */
    .domainGetOSType = dockerDomainGetOSType, /* 5.6.0 */
    .domainSetMemory = dockerDomainSetMemory, /* 5.6.0 */
    .domainSetMemoryFlags = dockerDomainSetMemoryFlags, /* 5.6.0 */
    .domainGetSchedulerType = dockerDomainGetSchedulerType, /* 5.6.0 */
    .domainGetSchedulerParameters = dockerDomainGetSchedulerParameters, /* 5.6.0 */
    .domainGetSchedulerParametersFlags = dockerDomainGetSchedulerParametersFlags, /* 5.6.0 */
    .domainSetSchedulerParameters = dockerDomainSetSchedulerParameters, /* 5.6.0 */
    .domainSetSchedulerParametersFlags = dockerDomainSetSchedulerParametersFlags, /* 5.6.0 */
    .domainGetBlockIoTune = dockerDomainGetBlockIoTune, /* 5.6.0 */
    .domainSetBlockIoTune = dockerDomainSetBlockIoTune, /* 5.6.0 */
    .domainGetXMLDesc = dockerDomainGetXMLDesc, /* 5.6.0 */
    .domainDefineXML = dockerDomainDefineXML, /* 5.6.0 */
    .domainDefineXMLFlags = dockerDomainDefineXMLFlags, /* 5.6.0 */
    .domainUndefine = dockerDomainUndefine, /* 5.6.0 */
    .domainUndefineFlags = dockerDomainUndefineFlags, /* 5.6.0 */
    .domainCreate = dockerDomainCreate, /* 5.6.0 */
    .domainCreateWithFlags = dockerDomainCreateWithFlags, /* 5.6.0 */
    .domainDestroy = dockerDomainDestroy, /* 5.6.0 */
    .domainDestroyFlags = dockerDomainDestroyFlags, /* 5.6.0 */
    .domainSuspend = dockerDomainSuspend, /* 5.6.0 */
    .domainShutdown = dockerDomainShutdown, /* 5.6.0 */
    .domainShutdownFlags = dockerDomainShutdownFlags, /* 5.6.0 */
    .domainResume = dockerDomainResume, /* 5.6.0 */
    .domainReboot = dockerDomainReboot, /* 5.6.0 */
    .domainRename = dockerDomainRename, /* 5.6.0 */
    .domainReset = dockerDomainReset, /* 5.6.0 */
    .domainBlockStats = dockerDomainBlockStats, /* 5.6.0 */
    .domainMemoryStats = dockerDomainMemoryStats, /* 5.6.0 */
    .domainInterfaceStats = dockerDomainInterfaceStats, /* 5.6.0 */
    .domainMigrateBegin3Params = dockerDomainMigrateBegin3Params, /* 5.6.0 */
    .domainMigratePrepare3Params = dockerDomainMigratePrepare3Params, /* 5.6.0 */
    .domainMigratePerform3Params = dockerDomainMigratePerform3Params, /* 5.6.0 */
    .domainMigrateFinish3Params = dockerDomainMigrateFinish3Params, /* 5.6.0 */
    .domainMigrateConfirm3Params = dockerDomainMigrateConfirm3Params, /* 5.6.0 */
    .nodeGetInfo = dockerNodeGetInfo, /* 5.6.0 */
};


virNetworkDriver dockerNetworkDriver = {
    .name = "Docker",
    .connectListAllNetworks = dockerConnectListAllNetworks, /* 5.6.0 */
    .networkLookupByUUID = dockerNetworkLookupByUUID, /* 5.6.0 */
    .networkLookupByName = dockerNetworkLookupByName, /* 5.6.0 */
    .networkGetXMLDesc = dockerNetworkGetXMLDesc, /* 5.6.0 */
    .networkDefineXML = dockerNetworkDefineXML, /* 5.6.0 */
    .networkUndefine = dockerNetworkUndefine, /* 5.6.0 */
    .networkGetAutostart = dockerNetworkGetAutostart, /* 5.6.0 */
    .networkSetAutostart = dockerNetworkSetAutostart, /* 5.6.0 */
    .networkIsActive = dockerNetworkIsActive, /* 5.6.0 */
    .networkIsPersistent = dockerNetworkIsPersistent, /* 5.6.0 */
};


virStorageDriver dockerStorageDriver = {
    .name = "Winrm",
    .connectListAllStoragePools = dockerConnectListAllStoragePools, /* 5.6.0 */
    .storagePoolLookupByUUID = dockerStoragePoolLookupByUUID, /* 5.6.0 */
    .storagePoolLookupByName = dockerStoragePoolLookupByName, /* 5.6.0 */
    .storagePoolLookupByVolume = dockerStoragePoolLookupByVolume, /* 5.6.0 */
    .storagePoolGetXMLDesc = dockerStoragePoolGetXMLDesc, /* 5.6.0 */
    .storagePoolGetInfo = dockerStoragePoolGetInfo, /* 5.6.0 */
    .storagePoolGetAutostart = dockerStoragePoolGetAutostart, /* 5.6.0 */
    .storagePoolSetAutostart = dockerStoragePoolSetAutostart, /* 5.6.0 */
    .storagePoolIsActive = dockerStoragePoolIsActive, /* 5.6.0 */
    .storagePoolIsPersistent = dockerStoragePoolIsPersistent, /* 5.6.0 */
    .storagePoolRefresh = dockerStoragePoolRefresh, /* 5.6.0 */
    .storagePoolListAllVolumes = dockerStoragePoolListAllVolumes, /* 5.6.0 */
    .storageVolLookupByName = dockerStorageVolLookupByName, /* 5.6.0 */
    .storageVolLookupByPath = dockerStorageVolLookupByPath, /* 5.6.0 */
    .storageVolLookupByKey = dockerStorageVolLookupByKey, /* 5.6.0 */
    .storageVolGetXMLDesc = dockerStorageVolGetXMLDesc, /* 5.6.0 */
    .storageVolCreateXML = dockerStorageVolCreateXML, /* 5.6.0 */
    .storageVolCreateXMLFrom = dockerStorageVolCreateXMLFrom, /* 5.6.0 */
    .storageVolDelete = dockerStorageVolDelete, /* 5.6.0 */
    .storageVolGetInfo = dockerStorageVolGetInfo, /* 5.6.0 */
    .storageVolGetInfoFlags = dockerStorageVolGetInfoFlags, /* 5.6.0 */
    .storageVolGetPath = dockerStorageVolGetPath, /* 5.6.0 */
    .storageVolResize = dockerStorageVolResize, /* 5.6.0 */
};


static virConnectDriver dockerConnectDriver = {
    .remoteOnly = true,
    .uriSchemes = (const char *[]){"docker", NULL},
    .hypervisorDriver = &dockerHypervisorDriver,
    .networkDriver = &dockerNetworkDriver,
    .storageDriver = &dockerStorageDriver,
};


/*
 * dockerRegister:
 *
 * Registers the Docker driver
 */
int
dockerRegister(void)
{
    return virRegisterConnectDriver(&dockerConnectDriver, false);
}
